/* spnic.c: Phobos NIC ethernet driver for Linux. */
/*
        Originally Written 1994-1998 by Donald Becker.
        Modified April 4, 2000 by Nathan Crapo of Phobos Corp.
        
        This software may be used and distributed according to the terms
        of the GNU Public License, incorporated herein by reference.

        This driver is for the the following cards:
        Phobos P5, 3M VOL - N10VF, Phobos P100TX, 
        Phobos P100FX, 3M VOL - N100VF, Phobos P430, 
        and 3M VOL - N400VF

        Support and updates available at
        http://support.aei-it.com/
*/


#define SMP_CHECK
static const char version[] = "spnic.c:v1.8 4/24/2003 support@aei-it.com\n";


/* A few user - configurable values. */

/* Maximum events( Rx packets, etc. ) to handle at each interrupt. */
static int max_interrupt_work = 25;


#define MAX_UNITS 8
/* Used to pass the full - duplex flag, etc. */
static int full_duplex[MAX_UNITS] = {0, };
static int options[MAX_UNITS] = {0, };
static int mtu[MAX_UNITS] = {0, };                      /* Jumbo MTU for interfaces. */

/*  The possible media types that can be set in options[] are: */
static const char* const medianame[] = {
        "10baseT", "10base2", "AUI", "100baseTx", 
        "10baseT - FD", "100baseTx - FD", "100baseT4", "100baseFx", 
        "100baseFx - FD", "MII 10baseT", "MII 10baseT - FD", "MII", 
        "10baseT( forced )", "MII 100baseTx", "MII 100baseTx - FD", "MII 100baseT4", 
};


/* Set if the PCI BIOS detects the chips on a multiport board backwards. */
#ifdef REVERSE_PROBE_ORDER
static int reverse_probe = 1;
#else
static int reverse_probe = 0;
#endif

/* Keep the ring sizes a power of two for efficiency.
   Making the Tx ring too large decreases the effectiveness of channel
   bonding and packet priority.
   There are no ill effects from too - large receive rings. */
#define TX_RING_SIZE    16
#define RX_RING_SIZE    32

/* Set the copy breakpoint for the copy - only - tiny - buffer Rx structure. */
#ifdef __alpha__
static int rx_copybreak = 1518;
#else
static int rx_copybreak = 100;
#endif

/* Operational parameters that usually are not changed. */
/* Time in jiffies before concluding the transmitter is hung. */
#define TX_TIMEOUT ( 4*HZ )

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/string.h>
#include <linux/timer.h>
#include <linux/ptrace.h>
#include <linux/errno.h>
#include <linux/ioport.h>
#include <linux/slab.h>
#include <linux/interrupt.h>
#include <linux/pci.h>

#include <asm/processor.h>              /* Processor type for cache alignment. */
#include <asm/bitops.h>
#include <asm/io.h>
#include <asm/dma.h>

#include <linux/netdevice.h>
#include <linux/etherdevice.h>
#include <linux/skbuff.h>

/* Kernel compatibility defines, common to David Hind's PCMCIA package.
   This is only in the support - all - kernels source code. */
#include <linux/version.h>              /* Evil, but neccessary */
#include "spnic.h"

#ifndef KERNEL_VERSION
#define KERNEL_VERSION(a,b,c) ((a << 16) + (b << 8) + (c))
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,2,0)
#include <linux/bios32.h>
#endif

#if LINUX_VERSION_CODE > KERNEL_VERSION(2,3,13)
#define device net_device
#endif

#if LINUX_VERSION_CODE > KERNEL_VERSION(2,1,39)
#define enet_statistics net_device_stats
#endif

/* Linux kernels prior to 2.0.37 had a bug with PCI reads from
 * the CSID register
 */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,0,37)
#ifdef PCI_SUBSYSTEM_VENDOR_ID
#undef PCI_SUBSYSTEM_VENDOR_ID
#define PCI_SUBSYSTEM_VENDOR_ID 0x2c
#endif /* PCI_SUBSYSTEM_VENDOR_ID */
#ifdef PCI_SUBSYSTEM_ID
#undef PCI_SUBSYSTEM_ID
#define PCI_SUBSYSTEM_ID	0x2e  
#endif /* PCI_SUBSYSTEM_ID */
#endif /* CSID_FIX */

#ifdef NAME_IF_PQFE
#define NAMELEN                   8
#define MAX_PHOBOS_CARDS          64

static struct device              phobos_dev[MAX_PHOBOS_CARDS];
static char                       namelist[MAX_PHOBOS_CARDS * NAMELEN];
#endif

#if defined( LINUX_VERSION_CODE ) && LINUX_VERSION_CODE < 0x10300
#define RUN_AT( x ) ( x )                   /* What to put in timer->expires.  */
#define DEV_ALLOC_SKB( len ) alloc_skb( len, GFP_ATOMIC )
#define virt_to_bus( addr )  ( (unsigned long )addr )
#define bus_to_virt( addr ) ( (void* )addr )

#else  /* 1.3.0 and later */
#define RUN_AT( x ) ( jiffies + ( x ))
#define DEV_ALLOC_SKB( len ) dev_alloc_skb( len + 2 )
#endif

#if ( LINUX_VERSION_CODE >= 0x10344 )
#define NEW_MULTICAST
#include <linux/delay.h>
#endif
#ifdef SA_SHIRQ
#define IRQ( irq, dev_id, pt_regs ) ( irq, dev_id, pt_regs )
#else
#define IRQ( irq, dev_id, pt_regs ) ( irq, pt_regs )
#endif

#if ( LINUX_VERSION_CODE < 0x20123 )
#define hard_smp_processor_id() smp_processor_id()
#define test_and_set_bit( val, addr ) set_bit( val, addr )
#endif

/* This my implementation of shared IRQs, now only used for 1.2.13. */
#ifdef HAVE_SHARED_IRQ
#define USE_SHARED_IRQ
#include <linux/shared_irq.h>
#endif

/* The total size is unusually large: The 21040 aligns each of its 16
   longword - wide registers on a quadword boundary. */
#define SPNIC_TOTAL_SIZE 0x80

#ifdef HAVE_DEVLIST
struct netdev_entry spnic_drv =
{"spnic", spnic_pci_probe, SPNIC_TOTAL_SIZE, NULL};
#endif

#ifdef SPNIC_DEBUG
int spnic_debug = SPNIC_DEBUG;
#else
int spnic_debug = 1;
#endif

/*
                                Theory of Operation

I. Board Compatibility

This device driver is designed for the DECchip "Tulip", Digital's
single - chip ethernet controllers for PCI.  Supported members of the family
are the 21040, 21041, 21140, 21140A, 21142, and 21143.  These chips are used on
many PCI boards including the SMC EtherPower series.


II. Board - specific settings

PCI bus devices are configured by the system at boot time, so no jumpers
need to be set on the board.  The system BIOS preferably should assign the
PCI INTA signal to an otherwise unused system IRQ line.
Note: Kernel versions earlier than 1.3.73 do not support shared PCI
interrupt lines.

III. Driver operation

IIIa. Ring buffers

The Tulip can use either ring buffers or lists of Tx and Rx descriptors.
This driver uses statically allocated rings of Rx and Tx descriptors, set at
compile time by RX / TX_RING_SIZE.  This version of the driver allocates skbuffs
for the Rx ring buffers at open() time and passes the skb->data field to the
Tulip as receive data buffers.  When an incoming frame is less than
RX_COPYBREAK bytes long, a fresh skbuff is allocated and the frame is
copied to the new skbuff.  When the incoming frame is larger, the skbuff is
passed directly up the protocol stack and replaced by a newly allocated
skbuff.

The RX_COPYBREAK value is chosen to trade - off the memory wasted by
using a full - sized skbuff for small frames vs. the copying costs of larger
frames.  For small frames the copying cost is negligible( esp. considering
that we are pre - loading the cache with immediately useful header
information ).  For large frames the copying cost is non - trivial, and the
larger copy might flush the cache of useful data.  A subtle aspect of this
choice is that the Tulip only receives into longword aligned buffers, thus
the IP header at offset 14 isn't longword aligned for further processing.
Copied frames are put into the new skbuff at an offset of "+2", thus copying
has the beneficial effect of aligning the IP header and preloading the
cache.

IIIC. Synchronization
The driver runs as two independent, single - threaded flows of control.  One
is the send - packet routine, which enforces single - threaded use by the
dev->tbusy flag.  The other thread is the interrupt handler, which is single
threaded by the hardware and other software.

The send packet thread has partial control over the Tx ring and device spinlock.
It sets the spinlock flag whenever it's queuing a Tx packet. If the next
queue slot is empty, it clears the tbusy flag when finished otherwise it sets
the 'tp->tx_full' flag.

The interrupt handler has exclusive control over the Rx ring and records stats
from the Tx ring.  ( The Tx - done interrupt can't be selectively turned off, so
we can't avoid the interrupt overhead by having the Tx routine reap the Tx
stats. )  After reaping the stats, it marks the queue entry as empty by setting
the 'base' to zero.      Iff the 'tp->tx_full' flag is set, it clears both the
tx_full and tbusy flags.

IV. Notes

Thanks to Duke Kamstra of SMC for providing an EtherPower board.

IVb. References

http://cesdis.gsfc.nasa.gov / linux / misc / NWay.html
http://www.digital.com( search for current 21* 4* datasheets and "21X4 SROM" )
http://www.national.com / pf / DP / DP83840.html

IVc. Errata

The DEC databook doesn't document which Rx filter settings accept broadcast
packets.  Nor does it document how to configure the part to configure the
serial subsystem for normal( vs. loopback ) operation or how to have it
autoswitch between internal 10baseT, SIA and AUI transceivers.

The 21040 databook claims that CSR13, CSR14, and CSR15 should each be the last
register of the set CSR12 - 15 written.  Hmmm, now how is that possible?  */


/* A few values that may be tweaked. */
#define PKT_BUF_SZ              1536                    /* Size of each temporary Rx buffer.*/

/* This is a mysterious value that can be written to CSR11 in the 21040 ( only )
   to support a pre - NWay full - duplex signaling mechanism using short frames.
   No one knows what it should be, but if left at its default value some
   10base2( ! ) packets trigger a full - duplex - request interrupt. */
#define FULL_DUPLEX_MAGIC       0x6969



#ifndef PCI_VENDOR_ID_DEC               /* Now defined in linux / pci.h */
#define PCI_VENDOR_ID_DEC                       0x1011
#define PCI_DEVICE_ID_TULIP                     0x0002          /* 21040. */
#define PCI_DEVICE_ID_TULIP_FAST        0x0009          /* 21140. */
#endif

#ifndef PCI_DEVICE_ID_DEC_TULIP_PLUS
#define PCI_DEVICE_ID_DEC_TULIP_PLUS    0x0014          /* 21041. */
#endif
#ifndef PCI_DEVICE_ID_DEC_TULIP_21142
#define PCI_DEVICE_ID_DEC_TULIP_21142   0x0019
#endif

#ifndef PCI_VENDOR_ID_LITEON
#define PCI_VENDOR_ID_LITEON    0x11AD
#endif

#ifndef PCI_VENDOR_ID_MXIC
#define PCI_VENDOR_ID_MXIC              0x10d9
#define PCI_DEVICE_ID_MX98713   0x0512
#define PCI_DEVICE_ID_MX98715   0x0531
#define PCI_DEVICE_ID_MX98725   0x0531
#endif


/* Phy identification values
 * Phobos cards use the "Level One" LXT 970AD and LXT971 mii Phys
 * Data sheets can be found at http://www.level1.com
 */
#define PHY_LXT970AD            0x78100000
#define PHY_LXT971              0x001378e0
#define FREQ                    .5

/* The rest of these values should never change. */

/* A table describing the chip types. */
enum tbl_flag { HAS_MII=1, HAS_MEDIA_TABLE = 2, CSR12_IN_SROM = 4, IS_PHOBOS_CARD = 8};
static struct spnic_chip_table {
        int vendor_id, device_id;
        int subvendor_id, subsystem_id;    /* zeroes are don't cares */
        char* chip_name;
        int io_size;
        int valid_intrs;                        /* CSR7 interrupt enable settings */
        int flags;
        void( *media_timer )( unsigned long data );
} spnic_tbl[] = {
  { PCI_VENDOR_ID_DEC, PCI_DEVICE_ID_DEC_TULIP_21142, 
    0x13d8, 0x5, 
    "Phobos P5", 128, 0x0801fbff, IS_PHOBOS_CARD, 
    NULL }, 
  { PCI_VENDOR_ID_DEC, PCI_DEVICE_ID_DEC_TULIP_21142, 
    0x13d8, 0x2010, 
    "3M VOL - N10VF", 128, 0x0801fbff, IS_PHOBOS_CARD, 
    NULL }, 
  { PCI_VENDOR_ID_DEC, PCI_DEVICE_ID_DEC_TULIP_21142, 
    0x13d8, 0x102, 
    "Phobos P100TX", 128, 0x0401fbff, IS_PHOBOS_CARD | HAS_MII, 
    NULL }, 
  { PCI_VENDOR_ID_DEC, PCI_DEVICE_ID_DEC_TULIP_21142, 
    0x13d8, 0x101, 
    "Phobos P100FX", 128, 0x0801fbff, HAS_MII | IS_PHOBOS_CARD, 
    NULL }, 
  { PCI_VENDOR_ID_DEC, PCI_DEVICE_ID_DEC_TULIP_21142, 
    0x13d8, 0x2101, 
    "3M VOL - N100VF", 128, 0x0801fbff, HAS_MII | IS_PHOBOS_CARD, 
    NULL }, 
  { PCI_VENDOR_ID_DEC, PCI_DEVICE_ID_DEC_TULIP_21142, 
    0x13d8, 0x430, 
    "Phobos P430", 128, 0x0401fbff, HAS_MII | IS_PHOBOS_CARD, 
        NULL }, 
  { PCI_VENDOR_ID_DEC, PCI_DEVICE_ID_DEC_TULIP_21142, 
    0x13d8, 0x2432, 
    "3M VOL - N400VF", 128, 0x0801fbff, HAS_MII | IS_PHOBOS_CARD, 
        NULL }, 
  {0, 0, 0, 0}, 
};
/* This matches the table above. */
enum chips { P5, N10VF, P100TX, P100FX, N100VF, P430, N400VF };


/* A full - duplex map for media types. */
enum MediaIs {MediaIsFD = 1, MediaAlwaysFD=2, MediaIsMII=4, MediaIsFx=8, 
                  MediaIs100=16};
static const char media_cap[] =
{0, 0, 0, 16, 3, 19, 16, 24, 27, 4, 7, 5, 0, 20, 23, 20 };

/* Offsets to the Command and Status Registers, "CSRs".  All accesses
   must be longword instructions and quadword aligned. */
enum spnic_offsets {
        CSR0=0, CSR1=0x08, CSR2=0x10, CSR3=0x18, CSR4=0x20, CSR5=0x28, 
        CSR6=0x30, CSR7=0x38, CSR8=0x40, CSR9=0x48, CSR10=0x50, CSR11=0x58, 
        CSR12=0x60, CSR13=0x68, CSR14=0x70, CSR15=0x78 };

/* The bits in the CSR5 status registers, mostly interrupt sources. */
enum status_bits {
        TimerInt=0x800, TPLnkFail=0x1000, TPLnkPass=0x10, 
        NormalIntr=0x10000, AbnormalIntr=0x8000, 
        RxJabber=0x200, RxDied=0x100, RxNoBuf=0x80, RxIntr=0x40, 
        TxFIFOUnderflow=0x20, TxJabber=0x08, TxNoBuf=0x04, TxDied=0x02, TxIntr=0x01, 
};

enum desc_status_bits {
        DescOwned=0x80000000, RxDescFatalErr=0x8000, RxWholePkt=0x0300, 
};
        
/* The Spnic Rx and Tx buffer descriptors. */
struct spnic_rx_desc {
        s32 status;
        s32 length;
        u32 buffer1, buffer2;
};

struct spnic_tx_desc {
        s32 status;
        s32 length;
        u32 buffer1, buffer2;                           /* We use only buffer 1.  */
};

struct medialeaf {
        u8 type;
        u8 media;
        unsigned char* leafdata;
};

struct mediatable {
        u16 defaultmedia;
        u8 leafcount, csr12dir;                         /* General purpose pin directions. */
        unsigned has_mii:1, has_nonmii:1;
        struct medialeaf mleaf[0];
};

struct mediainfo {
        struct mediainfo* next;
        int info_type;
        int index;
        unsigned char* info;
};

struct spnic_private {
        char devname[8];                        /* Used only for kernel debugging. */
        const char* product_name;
        struct device* next_module;
        struct spnic_rx_desc rx_ring[RX_RING_SIZE];
        struct spnic_tx_desc tx_ring[TX_RING_SIZE];
        /* The saved address of a sent - in - place packet / buffer, for skfree(). */
        struct sk_buff* tx_skbuff[TX_RING_SIZE];
        /* The addresses of receive - in - place skbuffs. */
        struct sk_buff* rx_skbuff[RX_RING_SIZE];
        char* rx_buffs;                         /* Address of temporary Rx buffers. */
        u32 setup_frame[48];            /* Pseudo - Tx frame to init address table. */
        int chip_id;
        int revision;
#if LINUX_VERSION_CODE > 0x20139
        struct net_device_stats stats;
#else
        struct enet_statistics stats;
#endif
        struct timer_list timer;        /* Media selection timer. */
        int interrupt;                          /* In - interrupt flag. */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
        spinlock_t lock;
#endif
#ifdef SMP_CHECK
        int smp_proc_id;                        /* Which processor in IRQ handler. */
#endif
        unsigned int cur_rx, cur_tx;            /* The next free ring entry */
        unsigned int dirty_rx, dirty_tx;        /* The ring entries to be free()ed. */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0)
        unsigned int tx_full:1;                         /* The Tx queue is full. */
#endif
        unsigned int full_duplex:1;                     /* Full - duplex operation requested. */
        unsigned int full_duplex_lock:1;
        unsigned int fake_addr:1;                       /* Multiport board faked address. */
        unsigned int media2:4;                          /* Secondary monitored media port. */
        unsigned int medialock:1;                       /* Don't sense media type. */
        unsigned int mediasense:1;                      /* Media sensing in progress. */
        unsigned int csr6;                                      /* Current CSR6 control settings. */
        unsigned char eeprom[128];                      /* Serial EEPROM contents. */
        u16 to_advertise;                                       /* NWay capabilities advertised.  */
        signed char phys[4], mii_cnt;           /* MII device addresses. */
        int phy_ouid[4];                    /* Phy Organizationally Unique Identifier */
        struct mediatable* mtable;
        int cur_index;                                          /* Current media index. */
        unsigned char pci_bus, pci_dev_fn;
        int running, pad1;                                              /* Used for 8 - byte alignment */
};

static struct device* spnic_probe1( int pci_bus, int pci_devfn, 
                                    struct device* dev, 
                                    int chip_id, int options );
static void LinkStatusChanged( struct device* dev );
static void SetPhyANAR( struct device* dev );
static void ResetPhy( struct device* dev );
static void EnableMiiInterrupts( struct device* dev );
static void SetupPhyLeds( struct device* dev );
static void ClearMiiInterrupt( struct device* dev );
static void write_csr6( struct device* dev );
static void SendSetupFrame( struct device* dev );
static int read_eeprom( long ioaddr, int location );
static int mdio_read( struct device* dev, int phy_id, int location );
static void mdio_write( struct device* dev, int phy_id, int location, int value );
static void select_media( struct device* dev, int startup );
static int spnic_open( struct device* dev );
static void spnic_tx_timeout( struct device* dev );
static void spnic_init_ring( struct device* dev );
static int spnic_start_xmit( struct sk_buff* skb, struct device* dev );
static int spnic_rx( struct device* dev );
static void spnic_interrupt IRQ( int irq, void* dev_instance, struct pt_regs* regs );
static int spnic_close( struct device* dev );
static struct enet_statistics* spnic_get_stats( struct device* dev );
#ifdef HAVE_PRIVATE_IOCTL
static int private_ioctl( struct device* dev, struct ifreq* rq, int cmd );
#endif
#ifdef NEW_MULTICAST
static void set_rx_mode( struct device* dev );
#else
static void set_rx_mode( struct device* dev, int num_addrs, void* addrs );
#endif
static void FindMiiPhys( struct device* dev );

#ifdef EXTRA_DEBUG
static void dumpPhyRegs( struct device* dev, const char* string );
static void dumpMacRegs( struct device* dev, const char* string );
static void dumpMacRegsSubset( struct device* dev, const char* string );
#endif /* EXTRA_DEBUG */

/* A list of all installed Tulip devices, for removing the driver module. */
static struct device* root_spnic_dev = NULL;

/* This 21040 probe no longer uses a large fixed contiguous Rx buffer region, 
   but now receives directly into full - sized skbuffs that are allocated
   at open() time.
   This allows the probe routine to use the old driver initialization
   interface. */

int spnic_probe( struct device* dev )
{
    int              cards_found   = 0;
    static int       pci_index     = 0;      /* Static, for multiple probe calls. */
    unsigned char    pci_bus;
    unsigned char pci_device_fn;


    /* Ideally we would detect all network cards in slot order.  That would
       be best done a central PCI probe dispatch, which wouldn't work
       well with the current structure.  So instead we detect just the
       Tulip cards in slot order. */

#if LINUX_VERSION_CODE >= 0x20155
    if ( ! pci_present() )
    {
        return( -ENODEV );
    }
#else
    if ( ! pcibios_present() )
    {
        return( -ENODEV );
    }
#endif

    for ( ;pci_index < 0xff; pci_index++ )
    {
        u16 vendor, device, pci_command, new_command;
        u16 subsystem, subsystemvendor;
        unsigned long pci_ioaddr = 0;
        int chip_idx = 0;
        
        if ( pcibios_find_class( PCI_CLASS_NETWORK_ETHERNET << 8, 
                                 reverse_probe ? 0xfe - pci_index : pci_index, 
                                 &pci_bus, &pci_device_fn ) != PCIBIOS_SUCCESSFUL )
        {
            if ( reverse_probe )
            {
                continue;
            }
            else
            {
                break;
            }
        }

        pcibios_read_config_word( pci_bus, pci_device_fn, 
                                  PCI_VENDOR_ID, &vendor );
        pcibios_read_config_word( pci_bus, pci_device_fn, 
                                  PCI_DEVICE_ID, &device );

        pcibios_read_config_word( pci_bus, pci_device_fn, 
                                  PCI_SUBSYSTEM_VENDOR_ID, &subsystemvendor );
        pcibios_read_config_word( pci_bus, pci_device_fn, 
                                  PCI_SUBSYSTEM_ID, &subsystem );

        for ( chip_idx = 0; spnic_tbl[chip_idx].chip_name; chip_idx++ )
        {
            /* Check the vendor and device ids first */
            if ( vendor == spnic_tbl[chip_idx].vendor_id &&
                 device == spnic_tbl[chip_idx].device_id )
            {

                /* Then check the subvendor and subsystem ids if they exist */
                /* Zeroes in the table mean we don't check.  Cards with subvendor */
                /* ID must be checked before a card without due to the algorithm */
                if ( ( subsystemvendor == spnic_tbl[chip_idx].subvendor_id &&
                       subsystem == spnic_tbl[chip_idx].subsystem_id ) ||
                     ( spnic_tbl[chip_idx].subvendor_id == 0 &&
                       spnic_tbl[chip_idx].subsystem_id == 0 ) )
                {
                    break;
                }
            }
        }

        if ( spnic_tbl[chip_idx].chip_name == 0 )
        {
            if ( vendor == PCI_VENDOR_ID_DEC  ||
                 vendor == PCI_VENDOR_ID_LITEON )
            {
                printk( KERN_INFO "Unknown Tulip - style PCI ethernet chip type"
                        " %4.4x %4.4x"" detected: not configured.\n", 
                        vendor, device );
            }
            continue;
        }
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
        pci_ioaddr = pci_resource_start( pci_find_slot( pci_bus, pci_device_fn ), 0 );
#elif (LINUX_VERSION_CODE >= 0x20155) && (LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0))
        pci_ioaddr = pci_find_slot( pci_bus, pci_device_fn )->base_address[0];
#elif LINUX_VERSION_CODE < 0x20155
        pcibios_read_config_dword( pci_bus, pci_device_fn, PCI_BASE_ADDRESS_0, 
                                   ( unsigned int* ) &pci_ioaddr );
#endif
        /* Remove I / O space marker in bit 0. */
        pci_ioaddr &= ~3;

        if ( spnic_debug > 2 )
        {
            printk( KERN_DEBUG "Found %s at I / O %#lx\n", 
                    spnic_tbl[chip_idx].chip_name, pci_ioaddr );
        }

        if ( check_region( pci_ioaddr, spnic_tbl[chip_idx].io_size ) )
        {
            continue;
        }

        pcibios_read_config_word( pci_bus, pci_device_fn, 
                                  PCI_COMMAND, &pci_command );
        new_command = pci_command | PCI_COMMAND_MASTER | PCI_COMMAND_IO;

        if ( pci_command != new_command )
        {
            printk( KERN_INFO "  The PCI BIOS has not enabled this"
                    " device!  Updating PCI command %4.4x->%4.4x.\n", 
                    pci_command, new_command );
            pcibios_write_config_word( pci_bus, pci_device_fn, 
                                       PCI_COMMAND, new_command );
        }

        dev = spnic_probe1( pci_bus, pci_device_fn, dev, chip_idx, cards_found );

        /* Get and check the bus - master and latency values. */
        if ( dev )
        {
            unsigned char pci_latency;
            pcibios_read_config_byte( pci_bus, pci_device_fn, 
                                      PCI_LATENCY_TIMER, &pci_latency );

            if ( pci_latency < 10 )
            {
                printk( KERN_INFO "  PCI latency timer( CFLT ) is "
                        "unreasonably low at %d.  Setting to 64 clocks.\n", 
                        pci_latency );
                pcibios_write_config_byte( pci_bus, pci_device_fn, 
                                           PCI_LATENCY_TIMER, 64 );
            }
            else if ( spnic_debug > 1 )
            {
                printk( KERN_INFO "  PCI latency timer( CFLT ) is %#x, "
                        " PCI command is %4.4x.\n", 
                        pci_latency, new_command );
            }
            
            /* Bring the card out of power-down mode. */
            pcibios_write_config_dword( pci_bus, pci_device_fn, 
                                            0x40, 0x0 );
            dev = 0;
            cards_found++;
        }
    }

    return( cards_found ? 0 : -ENODEV );
}


static struct device* spnic_probe1( int pci_bus, int pci_device_fn, 
                                    struct device* dev, 
                                    int chip_id, int board_idx )
{
    static int               did_version = 0;         /* Already printed version info. */
    struct spnic_private*    tp;
    long                     ioaddr;
    int                      irq;
    int                      i;
    unsigned short           sum;
    

    if ( spnic_debug > 0  &&  did_version++ == 0 )
    {
        printk( KERN_INFO "%s", version );
    }
    
#ifdef NAME_IF_PQFE
    if ( board_idx > MAX_PHOBOS_CARDS )
    {
        printk( KERN_ERR "Too many Phobos cards found!\n" );
        return( NULL );
    }
    
    dev = &phobos_dev[board_idx];
    dev->name = ( char* ) &namelist[NAMELEN * board_idx];
    sprintf( dev->name, "pqfe%d", board_idx );
    
    ether_setup( dev );
    
    if ( register_netdev( dev ) != 0 )
    {
        printk( KERN_ERR "Unable to register the interface: %s\n", dev->name );
    }
#else
    dev = init_etherdev( dev, 0 );
#endif
    
#if LINUX_VERSION_CODE >= 0x20155
    irq = pci_find_slot( pci_bus, pci_device_fn )->irq;
 #if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
    ioaddr = pci_resource_start( pci_find_slot( pci_bus, pci_device_fn ), 0 );
 #else
    ioaddr = pci_find_slot( pci_bus, pci_device_fn )->base_address[0];
 #endif
#else
    {
        u8       pci_irq_line;
        u32      pci_ioaddr;


        pcibios_read_config_byte( pci_bus, pci_device_fn, 
                                  PCI_INTERRUPT_LINE, &pci_irq_line );
        pcibios_read_config_dword( pci_bus, pci_device_fn, PCI_BASE_ADDRESS_0, 
                                   &pci_ioaddr );
        irq    = pci_irq_line;
        ioaddr = pci_ioaddr;
    }
#endif
    /* Remove I / O space marker in bit 0. */
    ioaddr &= ~3;

    printk( KERN_INFO "%s: %s at %#3lx\n", 
            dev->name, spnic_tbl[chip_id].chip_name, ioaddr );

    /* Stop the chip's Tx and Rx processes. */
    outl( inl( ioaddr + CSR6 ) & ~0x2002, ioaddr + CSR6 );

    /* Clear the missed - packet counter. */
    ( volatile int ) inl( ioaddr + CSR8 );

    /* The station address ROM is read byte serially.  The register must
       be polled, waiting for the value to be read bit serially from the
       EEPROM.
    */
    sum = 0;

    {
        /* Must be a new chip, with a serial EEPROM interface. */
        /* We read the whole EEPROM, and sort it out later.  DEC has a
           specification _Digital Semiconductor 21X4 Serial ROM Format_
           but early vendor boards just put the address in the first six
           EEPROM locations. */
        unsigned char ee_data[128];
        int sa_offset = 0;

        for ( i = 0; i < sizeof( ee_data )/2; i++ )
        {
            ( (u16* )ee_data )[i] = read_eeprom( ioaddr, i );
        }


        /* Phobos cards put their ethernet address here */
        sa_offset = 84;

        for ( i = 0; i < 6; i ++ ) {
            dev->dev_addr[i] = ee_data[i + sa_offset];
            sum += ee_data[i + sa_offset];
        }
    }

    /* Phobos boards have the address byte - swapped. */
    for ( i = 0; i < 6; i+=2 )
    {
        char tmp             = dev->dev_addr[i];
        dev->dev_addr[i]     = dev->dev_addr[i + 1];
        dev->dev_addr[i + 1] = tmp;
    }

    /* We do a request_region() only to register /proc / ioports info. */
    /* Note that proper size is spnic_tbl[chip_id].chip_name, but... */
    request_region( ioaddr, SPNIC_TOTAL_SIZE, dev->name );

    dev->base_addr = ioaddr;
    dev->irq       = irq;

    /* Make certain the data structures are quadword aligned. */
    tp = ( void* )( ( ( long ) kmalloc( sizeof( *tp ), GFP_KERNEL | GFP_DMA ) + 7 ) & ~7 );
    memset( tp, 0, sizeof( *tp ) );
    dev->priv = tp;

    tp->next_module = root_spnic_dev;
    root_spnic_dev  = dev;

    tp->pci_bus    = pci_bus;
    tp->pci_dev_fn = pci_device_fn;
    tp->chip_id    = chip_id;

#ifdef SPNIC_FULL_DUPLEX
    tp->full_duplex      = 1;
    tp->full_duplex_lock = 1;
 #endif

#ifdef SPNIC_NO_MEDIA_SWITCH
    tp->medialock = 1;
#endif

    tp->running = 1;
    
    /* The lower four bits are the media type. */
    if ( board_idx >= 0  &&  board_idx < MAX_UNITS )
    {
        if ( (options[board_idx] & 0x90 ) || full_duplex[board_idx] > 0 )
        {
            tp->full_duplex = 1;
        }
        if ( mtu[board_idx] > 0 )
        {
            dev->mtu = mtu[board_idx];
        }
    }

    if ( tp->full_duplex )
    {
        tp->full_duplex_lock = 1;
    }
    
    switch( chip_id )
    {
    case P100TX:
    case P430:
        tp->to_advertise = 0x01e1;
        break;

    case P100FX:
    case N100VF:
    case N400VF:
        tp->to_advertise = 0x0181;
        break;

    default:
        tp->to_advertise = 0x01e1;
        break;
    }

    if ( spnic_tbl[tp->chip_id].flags & HAS_MII )
    {
        FindMiiPhys( dev );
        SetPhyANAR( dev );
    }
    
    /* The Spnic - specific entries in the device structure. */
    dev->open = &spnic_open;
    dev->hard_start_xmit = &spnic_start_xmit;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
    dev->tx_timeout = spnic_tx_timeout;
#endif
    dev->stop = &spnic_close;
    dev->get_stats = &spnic_get_stats;
#ifdef HAVE_PRIVATE_IOCTL
    dev->do_ioctl = &private_ioctl;
#endif
#ifdef HAVE_MULTICAST
    dev->set_multicast_list = &set_rx_mode;
#endif

    /* Reset the xcvr interface and turn on heartbeat. */
    switch ( chip_id )
    {
    case N100VF:
    case P100FX:
    case P100TX:
    case P430:
        break;

    case P5:
        outl( 0x00000000, ioaddr + CSR13 );
        outl( 0x00000705, ioaddr + CSR14 );
        outl( 0x081e000e, ioaddr + CSR15 );
        outl( 0x000e000e, ioaddr + CSR15 );
        outl( 0x00000009, ioaddr + CSR13 );
        break;

    case N10VF:
        outl( 0x82420200, ioaddr + CSR6 );
        outl( 0x00000000, ioaddr + CSR13 );
        outl( 0x00007F3D, ioaddr + CSR14 );
        outl( 0x081e0008, ioaddr + CSR15 );
        outl( 0x000e0008, ioaddr + CSR15 );
        outl( 0x00000001, ioaddr + CSR13 );
        break;
    }

    return( dev );
}

/* Serial EEPROM section. */
/* The main routine to parse the very complicated SROM structure.
   Search www.digital.com for "21X4 SROM" to get details.
   This code is very complex, and will require changes to support
   additional cards, so I'll be verbose about what is going on.
   */

#define EEPROM_SIZE 128
#if defined( __i386__ )
#define get_u16( ptr ) ( *( u16* ) ( ptr ) )
#else
#define get_u16( ptr ) ( ( ( u8* ) ( ptr ) )[0] + ( ( ( u8* ) ( ptr ) )[1] << 8 ) )
#endif


/* Reading a serial EEPROM is a "bit" grungy, but we work our way through:->.*/

/*  EEPROM_Ctrl bits. */
#define EE_SHIFT_CLK            0x02    /* EEPROM shift clock. */
#define EE_CS                   0x01    /* EEPROM chip select. */
#define EE_DATA_WRITE           0x04    /* EEPROM chip data in. */
#define EE_WRITE_0              0x01
#define EE_WRITE_1              0x05
#define EE_DATA_READ            0x08    /* EEPROM chip data out. */
#define EE_ENB                  ( 0x4800 | EE_CS )

/* Delay between EEPROM clock transitions.
   The 1.2 code is a "nasty" timing loop, but PC compatible machines are
   *supposed* to delay an ISA - compatible period for the SLOW_DOWN_IO macro.  */
#ifdef _LINUX_DELAY_H
#define eeprom_delay( nanosec )   udelay( (nanosec + 999 )/1000 )
#else
#define eeprom_delay( nanosec )   do { int _i = 3; while ( --_i > 0 ) { __SLOW_DOWN_IO; } } while ( 0 )
#endif


/* The EEPROM commands include the alway - set leading bit. */
#define EE_WRITE_CMD            ( 5 << 6 )
#define EE_READ_CMD             ( 6 << 6 )
#define EE_ERASE_CMD            ( 7 << 6 )


static int read_eeprom( long ioaddr, int location )
{
    int i;
    unsigned short retval = 0;
    long ee_addr = ioaddr + CSR9;
    int read_cmd = location | EE_READ_CMD;
        
    outl( EE_ENB & ~EE_CS, ee_addr );
    outl( EE_ENB, ee_addr );
        
    /* Shift the read command bits out. */
    for ( i = 10; i >= 0; i-- )
    {
        short dataval = ( read_cmd & ( 1 << i ) ) ? EE_DATA_WRITE : 0;
        outl( EE_ENB | dataval, ee_addr );
        eeprom_delay( 100 );
        outl( EE_ENB | dataval | EE_SHIFT_CLK, ee_addr );
        eeprom_delay( 150 );
        outl( EE_ENB | dataval, ee_addr );        /* Finish EEPROM a clock tick. */
        eeprom_delay( 250 );
    }
    outl( EE_ENB, ee_addr );
        
    for ( i = 16; i > 0; i-- )
    {
        outl( EE_ENB | EE_SHIFT_CLK, ee_addr );
        eeprom_delay( 100 );
        retval = ( retval << 1 ) | ( (inl( ee_addr ) & EE_DATA_READ ) ? 1 : 0 );
        outl( EE_ENB, ee_addr );
        eeprom_delay( 100 );
    }

    /* Terminate the EEPROM access. */
    outl( EE_ENB & ~EE_CS, ee_addr );
    return( retval );
}

/* MII transceiver control section.
   Read and write the MII registers using software - generated serial
   MDIO protocol.  See the MII specifications or DP83840A data sheet
   for details. */

/* The maximum data clock rate is 2.5 Mhz.  The minimum timing is usually
   met by back - to - back PCI I / O cycles, but we insert a delay to avoid
   "overclocking" issues or future 66Mhz PCI. */
#define mdio_delay() inl( mdio_addr )

/* Read and write the MII registers using software - generated serial
   MDIO protocol.  It is just different enough from the EEPROM protocol
   to not share code.  The maxium data clock rate is 2.5 Mhz. */
#define MDIO_SHIFT_CLK          0x10000
#define MDIO_DATA_WRITE0        0x00000
#define MDIO_DATA_WRITE1        0x20000
#define MDIO_ENB                0x00000         /* Ignore the 0x02000 databook setting. */
#define MDIO_ENB_IN             0x40000
#define MDIO_DATA_READ          0x80000


static int mdio_read( struct device* dev, int phy_id, int location )
{
    int     i;
    int     read_cmd      = ( 0xf6 << 10 ) | ( phy_id << 5 ) | location;
    int     retval        = 0;
    long    mdio_addr     = dev->base_addr + CSR9;


    /* Establish sync by sending at least 32 logic ones. */ 
    for ( i = 32; i >= 0; i-- )
    {
        outl( MDIO_ENB | MDIO_DATA_WRITE1, mdio_addr );
        mdio_delay();
        outl( MDIO_ENB | MDIO_DATA_WRITE1 | MDIO_SHIFT_CLK, mdio_addr );
        mdio_delay();
    }

    /* Shift the read command bits out. */
    for ( i = 15; i >= 0; i-- )
    {
        int dataval = ( read_cmd & ( 1 << i ) ) ? MDIO_DATA_WRITE1 : 0;

        outl( MDIO_ENB | dataval, mdio_addr );
        mdio_delay();
        outl( MDIO_ENB | dataval | MDIO_SHIFT_CLK, mdio_addr );
        mdio_delay();
    }
    /* Read the two transition, 16 data, and wire - idle bits. */
    for ( i = 19; i > 0; i-- )
    {
        outl( MDIO_ENB_IN, mdio_addr );
        mdio_delay();
        retval = ( retval << 1 ) | ( ( inl( mdio_addr ) & MDIO_DATA_READ) ? 1 : 0 );
        outl( MDIO_ENB_IN | MDIO_SHIFT_CLK, mdio_addr );
        mdio_delay();
    }
    return( ( retval >> 1 ) & 0xffff );
}

static void mdio_write( struct device* dev, int phy_id, int location, int value )
{
    int i;
    int cmd = ( 0x5002 << 16) | ( phy_id << 23) | ( location<<18) | value;
    long mdio_addr = dev->base_addr + CSR9;


    /* Establish sync by sending 32 logic ones. */ 
    for ( i = 32; i >= 0; i--)
    {
        outl( MDIO_ENB | MDIO_DATA_WRITE1, mdio_addr );
        mdio_delay();
        outl( MDIO_ENB | MDIO_DATA_WRITE1 | MDIO_SHIFT_CLK, mdio_addr );
        mdio_delay();
    }

    /* Shift the command bits out. */
    for ( i = 31; i >= 0; i--)
    {
        int dataval = ( cmd & ( 1 << i)) ? MDIO_DATA_WRITE1 : 0;
        outl( MDIO_ENB | dataval, mdio_addr );
        mdio_delay();
        outl( MDIO_ENB | dataval | MDIO_SHIFT_CLK, mdio_addr );
        mdio_delay();
    }
    /* Clear out extra bits. */
    for ( i = 2; i > 0; i--)
    {
        outl( MDIO_ENB_IN, mdio_addr );
        mdio_delay();
        outl( MDIO_ENB_IN | MDIO_SHIFT_CLK, mdio_addr );
        mdio_delay();
    }
    return;
}


static int spnic_open( struct device* dev )
{
    struct spnic_private*     tp         = ( struct spnic_private*)dev->priv;
    long                      ioaddr     = dev->base_addr;

    
    /* On some chip revs we must set the MII / SYM port before the reset!? */
    if ( tp->mii_cnt )
    {
        outl( 0x02040000, ioaddr + CSR6 );
    }
    
    /* Reset the chip, holding bit 0 set at least 50 PCI cycles. */
    outl( 0x00000001, ioaddr + CSR0 );

    udelay( 100 );

    /* Deassert reset.
       486: Set 8 longword cache alignment, 8 longword burst.
       586: Set 16 longword cache alignment, no burst limit.
       Cache alignment bits 15:14        Burst length 13:8
            0000    No alignment  0x00000000 unlimited              0800 8 longwords
            4000    8  longwords            0100 1 longword         1000 16 longwords
            8000    16 longwords            0200 2 longwords        2000 32 longwords
            C000    32  longwords           0400 4 longwords
       Wait the specified 50 PCI cycles after a reset by initializing
       Tx and Rx queues and the address filter list. */

#if defined( __alpha__)
    /* ToDo: Alpha setting could be better. */
    outl( 0x01A00000 | 0xE000, ioaddr + CSR0 );
#elif defined( __powerpc__ )
    outl( 0x01A00080 | 0x8000, ioaddr + CSR0 );
#elif defined( __i386__ )
#if defined( MODULE )
    /* When a module we don't have 'x86' to check. */
    outl( 0x01A00000 | 0x4800, ioaddr + CSR0 );
#else
#if ( LINUX_VERSION_CODE > 0x2014c )
#define x86 boot_cpu_data.x86
#endif
    outl( 0x01A00000 | ( x86 <= 4 ? 0x4800 : 0x8000 ), ioaddr + CSR0 );
    if ( x86 <= 4)
    {
        printk( KERN_INFO "%s: This is a 386 / 486 PCI system, setting cache "
                "alignment to %x.\n", dev->name, 
                0x01A00000 | ( x86 <= 4 ? 0x4800 : 0x8000 ) );
    }
#endif
#else
    outl( 0x01A00000 | 0x4800, ioaddr + CSR0 );
#warning Processor architecture undefined!
#endif

#ifdef SA_SHIRQ
    if ( request_irq( dev->irq, &spnic_interrupt, SA_SHIRQ, dev->name, dev ) )
    {
        return( -EAGAIN );
    }
#else
    if ( irq2dev_map[dev->irq] != NULL ||
         ( irq2dev_map[dev->irq] = dev ) == NULL ||
         dev->irq == 0 ||
         request_irq( dev->irq, &spnic_interrupt, 0, 
                         spnic_tbl[tp->chip_id].chip_name ) )
    {
        return( -EAGAIN );
    }
#endif

    if ( spnic_debug > 1 )
    {
        printk( KERN_DEBUG "%s: spnic_open() irq %d.\n", dev->name, dev->irq );
    }
    
    MOD_INC_USE_COUNT;

    spnic_init_ring( dev );

    outl( virt_to_bus( tp->rx_ring ), ioaddr + CSR3 );
    outl( virt_to_bus( tp->tx_ring ), ioaddr + CSR4 );

    select_media( dev, 1 );
    
    /* reset the phy */
    if ( spnic_tbl[tp->chip_id].flags & HAS_MII )
    {
        ResetPhy( dev );
        SetPhyANAR( dev );
        EnableMiiInterrupts( dev );
        SetupPhyLeds( dev );
    }
    
    tp->cur_index = 0;

    SendSetupFrame( dev );

    /* Start the chip's Tx to process setup frame. */
    tp->csr6 |= OPMODE_ST;
    write_csr6( dev );

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0)
    /* Setup state variables */
    dev->tbusy     = 0;
    tp->interrupt  = 0;
    dev->start     = 1;
#endif
    
    /* Check to see if setup frame was processed? */

    /* Start the Receive Process */
    tp->csr6 |= 0x2002;
    write_csr6( dev );
    outl( 0, ioaddr + CSR2 );           /* Rx poll demand */

    /* Clear pending interrupts */
    outl( spnic_tbl[tp->chip_id].valid_intrs, ioaddr + CSR5 );    

    /* Enable interrupts by setting the interrupt mask. */
    outl( spnic_tbl[tp->chip_id].valid_intrs, ioaddr + CSR7 );

    if ( spnic_debug > 2)
    {
        printk( KERN_DEBUG "%s: Done spnic_open(), CSR0 %8.8x, CSR5 %8.8x CSR6 %8.8x.\n",
                dev->name, inl( ioaddr + CSR0 ), inl( ioaddr + CSR5 ),
                inl( ioaddr + CSR6 ) );
    }

    LinkStatusChanged( dev );
    
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
    netif_start_queue( dev );
#endif

    
    return( 0 );
}


/* Set up the transceiver control registers for the selected media type. */
static void select_media( struct device* dev, int startup )
{
    long                     ioaddr     = dev->base_addr;
    struct spnic_private*    tp         = ( struct spnic_private* ) dev->priv;
    u32                      new_csr6   = 0;


    switch ( tp->chip_id )
    {
    case P5:
        /* This card is hardwired to 10MB HD */
        dev->if_port = 2;
        outl( 0x00000000, ioaddr + CSR13 );
        outl( 0x00000705, ioaddr + CSR14 );
        outl( 0x081e000e, ioaddr + CSR15 );
        outl( 0x000e000e, ioaddr + CSR15 );
        outl( 0x00000009, ioaddr + CSR13 );
        new_csr6 = 0x02000000;
        break;
        
    case N10VF:
        /* This card is hardwired to 10MB FD */
        dev->if_port = 0;
        outl( 0x00000000, ioaddr + CSR13 );
        outl( 0x00007F3D, ioaddr + CSR14 );
        outl( 0x081e0008, ioaddr + CSR15 );
        outl( 0x000e0008, ioaddr + CSR15 );
        outl( 0x00000001, ioaddr + CSR13 );
        new_csr6 = 0x02000200;
        break;
        
    case N100VF:
    case P100FX:
    case P430:
    case N400VF:
    case P100TX:
        dev->if_port = 11;
        outl( 0x00000000, ioaddr + CSR13 );
        outl( 0x00000000, ioaddr + CSR14 );

        /* gepl stuff */
        outl( 0x090e0000, ioaddr + CSR15 );
        outl( 0x000e0000, ioaddr + CSR15 );

        new_csr6 = OPMODE_PS | OPMODE_HBD | OPMODE_PB;

        /* clear mii interrupt */
        ClearMiiInterrupt( dev );
        break;

    default:
        printk( KERN_DEBUG "%s: Unknown card. . .\n", dev->name );
        break;
    }
    
    tp->csr6 = new_csr6 | tp->csr6;
    return;
}


static void ClearMiiInterrupt( struct device* dev )
{
    struct spnic_private*               tp = dev->priv;

    
    mdio_read( dev, tp->phys[0], 1 );
    mdio_read( dev, tp->phys[0], 18 );
}

static void LinkStatusChanged( struct device* dev )
{
    struct spnic_private*        tp          = ( struct spnic_private*)dev->priv;
    long                         ioaddr      = dev->base_addr;
    static int                   status      = 0;
    int                          new_csr6;
    int                          new_status  = 0;
    int                          timeout;
    int                          csr6;
    

    if ( tp->phy_ouid[0] == PHY_LXT970AD )
    {
        new_status = mdio_read( dev, tp->phys[0], 0x14 );
        new_status &= 0x3b04;
    }
    else if ( tp->phy_ouid[0] == PHY_LXT971 )
    {
        new_status = mdio_read( dev, tp->phys[0], 0x11 );
    }
        
    /* Determine if anything has changed */
    if ( new_status == status )
    {
        return;
    }

    status = new_status;

#if 0
    dumpPhyRegs( dev, "LinkStatusChanged" );
#endif
    
    new_csr6 = tp->csr6;
    
    if ( tp->phy_ouid[0] == PHY_LXT970AD )
    {
        /* check duplex */
        if ( status & 0x1000 )
        {
            new_csr6 |= OPMODE_FD;
        }
        else
        {
            new_csr6 &= ~OPMODE_FD;
        }
        
        /* check speed */
        if ( status & 0x800 )
        {
            new_csr6 &= ~OPMODE_TTM;
        }
        else
        {
            new_csr6 |=  OPMODE_TTM;
        }
    }
    else if ( tp->phy_ouid[0] == PHY_LXT971 )
    {
        /* check duplex */
        if ( status & 0x0200 )
        {
            new_csr6 |= OPMODE_FD;
        }
        else
        {
            new_csr6 &= ~OPMODE_FD;
        }
        
        /* check speed */
        if ( status & 0x4000 )
        {
            new_csr6 &= ~OPMODE_TTM;
        }
        else
        {
            new_csr6 |=  OPMODE_TTM;
        }
    }
    else
    {
        printk( KERN_INFO "Unknown PHY during LinkStatusChanged\n" );
    }
    
    /* stop tx / rx process */
    tp->running = 0;

    timeout = 10000;
    while ( timeout )
    {
        tp->csr6 &= ~ ( OPMODE_ST | OPMODE_SR );
        write_csr6( dev );

        csr6 = inl( ioaddr + CSR6 );

        if ( ( csr6 & ( OPMODE_ST | OPMODE_SR ) ) == 0 )
        {
            break;
        }
        timeout--;
    }

    if ( timeout == 0 )
    {
        printk( KERN_INFO "%s: LinkStatusChanged to\n", dev->name );
    }

    /* Engage new settings */
    new_csr6 &= ~ ( OPMODE_ST | OPMODE_SR );
    tp->csr6 = new_csr6;
    write_csr6( dev );

    /* turn on rx and tx process */
    tp->csr6 |= ( OPMODE_ST | OPMODE_SR );
    write_csr6( dev );

    tp->running = 1;
}


static void spnic_tx_timeout( struct device* dev )
{
    struct spnic_private*     tp      = ( struct spnic_private* )dev->priv;
    long                      ioaddr  = dev->base_addr;


    if ( tp->chip_id == N10VF )
    {
        /* Stop the transmit process. */
        outl( tp->csr6 | 0x0002, ioaddr + CSR6 );
        printk( KERN_WARNING "%s: 21140 transmit timed out, status %8.8x, "
                "SIA %8.8x %8.8x %8.8x %8.8x, resetting...\n", 
                dev->name, inl( ioaddr + CSR5 ), inl( ioaddr + CSR12 ), 
                inl( ioaddr + CSR13), inl( ioaddr + CSR14 ),
                inl( ioaddr + CSR15 ) );

        outl( tp->csr6 | 0x2002, ioaddr + CSR6 );
        tp->stats.tx_errors++;
        dev->trans_start = jiffies;
    }
    else
    {
        /* Do nothing -- the media monitor should handle this. */
        if ( spnic_debug > 1 )
        {
            printk( KERN_WARNING "%s: Transmit timeout using MII device.\n", 
                    dev->name );
        }
        dev->trans_start = jiffies;
    }

    return;
}


/* Initialize the Rx and Tx rings, along with various 'dev' bits. */
static void spnic_init_ring( struct device* dev )
{
    struct spnic_private*              tp  = ( struct spnic_private* ) dev->priv;
    int                                i;



#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0)
    tp->tx_full    = 0;
#endif
    tp->cur_rx     = tp->cur_tx     = 0;
    tp->dirty_rx   = tp->dirty_tx   = 0;

    for ( i = 0; i < RX_RING_SIZE; i++)
    {
        tp->rx_ring[i].status = 0x80000000;     /* Owned by Spnic chip */
        tp->rx_ring[i].length = PKT_BUF_SZ;

        {
            /* Note the receive buffer must be longword aligned.
               dev_alloc_skb() provides 16 byte alignment.  But do* not*
               use skb_reserve() to align the IP header! */
            struct sk_buff* skb;
            skb = DEV_ALLOC_SKB( PKT_BUF_SZ );
            tp->rx_skbuff[i] = skb;
            if ( skb == NULL )
            {
                break;                  /* Bad news!  */
            }
            skb->dev = dev;                 /* Mark as being used by this device. */
#if LINUX_VERSION_CODE > 0x10300
            tp->rx_ring[i].buffer1 = virt_to_bus( skb->tail );
#else
            tp->rx_ring[i].buffer1 = virt_to_bus( skb->data );
#endif
        }
        tp->rx_ring[i].buffer2 = virt_to_bus( &tp->rx_ring[i + 1] );
    }
    /* Mark the last entry as wrapping the ring. */
    tp->rx_ring[i - 1].length = PKT_BUF_SZ | 0x02000000;
    tp->rx_ring[i - 1].buffer2 = virt_to_bus( &tp->rx_ring[0] );

    /* The Tx buffer descriptor is filled in as needed, but we
     * do need to clear the ownership bit.
     */
    for ( i = 0; i < TX_RING_SIZE; i++ )
    {
        tp->tx_skbuff[i]              = 0;
        tp->tx_ring[i].status         = 0x00000000;
        tp->tx_ring[i].buffer2        = virt_to_bus( &tp->tx_ring[i + 1] );
    }
    tp->tx_ring[i - 1].buffer2 = virt_to_bus( &tp->tx_ring[0] );
}


static int spnic_start_xmit( struct sk_buff* skb, struct device* dev )
{
    struct spnic_private*            tp            = ( struct spnic_private* )dev->priv;
    int                              entry;
    u32                              flag;

    
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
    spin_lock_irq( &tp->lock );
#else
    /* Block a timer - based transmit from overlapping.  This could better be
       done with atomic_swap( 1, dev->tbusy), but set_bit() works as well. */
    if ( test_and_set_bit( 0, ( void* ) &dev->tbusy ) != 0 )
    {
        if ( jiffies - dev->trans_start < TX_TIMEOUT )
        {
            return( 1 );
        }
        spnic_tx_timeout( dev );
        return( 1 );
    }
#endif
    
    /* Caution: the write order is important here, set the base address
       with the "ownership" bits last. */

#if 0
    dumpMacRegs( dev, "xmit" );
#endif
    
#if 0
    dumpPhyRegs( dev, "xmit:" );
#endif

    /* Calculate the next Tx descriptor entry. */
    entry = tp->cur_tx % TX_RING_SIZE;

    tp->tx_skbuff[entry]       = skb;
    tp->tx_ring[entry].buffer1 = virt_to_bus( skb->data );

    if ( tp->cur_tx - tp->dirty_tx < TX_RING_SIZE / 2 )
    {
        /* Typical path */
        flag       = 0x60000000; /* No interrupt */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0)
        dev->tbusy = 0;
#endif        
    }
    else if ( tp->cur_tx - tp->dirty_tx == TX_RING_SIZE / 2 )
    {
        flag       = 0xe0000000; /* Tx - done intr. */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0)        
        dev->tbusy = 0;
#endif
    }
    else if ( tp->cur_tx - tp->dirty_tx < TX_RING_SIZE - 2 )
    {
        flag       = 0x60000000; /* No Tx - done intr. */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0)        
        dev->tbusy = 0;
#endif
    }
    else
    {
        /* Leave room for set_rx_mode() to fill entries. */
        flag        = 0xe0000000; /* Tx - done intr. */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)        
        netif_stop_queue( dev );
#else
        tp->tx_full = 1;
#endif
    }
    if ( entry == TX_RING_SIZE - 1 )
    {
        flag |= 0xe2000000;
    }
    
    tp->tx_ring[entry].length = skb->len | flag;
    tp->tx_ring[entry].status = 0x80000000; /* Pass ownership to the chip. */
    tp->cur_tx++;
    /* Trigger an immediate transmit demand. */
    outl( 0, dev->base_addr + CSR1 );

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
    spin_unlock_irq( &tp->lock );
#endif

    dev->trans_start = jiffies;

    return( 0 );
}


/* The interrupt handler does all of the Rx thread work and cleans up
   after the Tx thread. */
static void spnic_interrupt IRQ( int irq, void* dev_instance, struct pt_regs* regs )
{
#ifdef SA_SHIRQ         /* Use the now - standard shared IRQ implementation. */
    struct device* dev = ( struct device* ) dev_instance;
#else
    struct device* dev = ( struct device* ) ( irq2dev_map[irq] );
#endif

    struct spnic_private*     tp;
    long                      ioaddr;
    int                       csr5;
    int                       work_budget = max_interrupt_work;

    if ( dev == NULL )
    {
        printk( KERN_ERR" spnic_interrupt(): irq %d for unknown device.\n", 
                irq );
        return;
    }

#if 0
    dumpMacRegs( dev, "interrupt" );
#endif
    
    ioaddr = dev->base_addr;
    tp     = ( struct spnic_private* ) dev->priv;

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0)
    if ( test_and_set_bit( 0, ( void* ) &tp->interrupt ) )
    {
#ifdef SMP_CHECK
        printk( KERN_ERR "%s: Re - entering the interrupt handler with proc %d, "
                " proc %d already handling.\n", dev->name, 
                tp->smp_proc_id, hard_smp_processor_id() );
#else
        printk( KERN_ERR "%s: Re - entering the interrupt handler.\n", dev->name );
#endif /* SMP_CHECK */
        return;
    }
#endif /* LINUX_VERSION_CODE... */

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0)
    dev->interrupt = 1;
#endif
    
#ifdef SMP_CHECK
    tp->smp_proc_id = hard_smp_processor_id();
#endif

    do {
        csr5 = inl( ioaddr + CSR5 );

        /* Acknowledge all of the current interrupt sources ASAP. */
        outl( csr5 & 0x0401ffff, ioaddr + CSR5 );
        
        if ( spnic_debug > 4)
        {
            printk( KERN_DEBUG "%s: interrupt  csr5=%#8.8x new csr5=%#8.8x.\n", 
                    dev->name, csr5, inl( dev->base_addr + CSR5 ) );
        }
        
        if ( csr5 & 0x04000000 )
        {
            /* clear mii interrupt */
            ClearMiiInterrupt( dev );

            /* do link status change stuff */
            LinkStatusChanged( dev );
            break;
        }

        if ( ( csr5 & ( NormalIntr | AbnormalIntr ) ) == 0 )
        {
            break;
        }
        
        if ( csr5 & ( RxIntr | RxNoBuf ) )
        {
            work_budget -= spnic_rx( dev );
        }

        if ( csr5 & ( TxNoBuf | TxDied | TxIntr ) )
        {
            unsigned int dirty_tx;

            for ( dirty_tx = tp->dirty_tx; tp->cur_tx - dirty_tx > 0;
                  dirty_tx++ )
            {
                int entry   = dirty_tx % TX_RING_SIZE;
                int status  = tp->tx_ring[entry].status;

                if ( status < 0 )
                {
                    break;                  /* It still hasn't been Txed */
                }
                
                /* Check for Rx filter setup frames. */
                if ( tp->tx_skbuff[entry] == NULL )
                {
                    continue;
                }

                if ( status & 0x8000 )
                {
                    /* There was an major error, log it. */
#ifndef final_version
                    if ( spnic_debug > 1 )
                    {
                        printk( KERN_DEBUG "%s: Transmit error, Tx status %8.8x.\n", 
                                dev->name, status);
                    }
#endif
                    tp->stats.tx_errors++;
                    if ( status & 0x4104 )
                    {
                        tp->stats.tx_aborted_errors++;
                    }
                    
                    if ( status & 0x0C00 )
                    {
                        tp->stats.tx_carrier_errors++;
                    }
                    
                    if ( status & 0x0200 )
                    {
                        tp->stats.tx_window_errors++;
                    }
                    
                    if ( status & 0x0002 )
                    {
                        tp->stats.tx_fifo_errors++;
                    }
                    
                    if ( ( status & 0x0080 ) && tp->full_duplex == 0 )
                    {
                        tp->stats.tx_heartbeat_errors++;
                    }
                    
#ifdef ETHER_STATS
                    if ( status & 0x0100 )
                    {
                        tp->stats.collisions16++;
                    }
#endif
                }
                else
                {

#ifdef ETHER_STATS
                    if ( status & 0x0001 )
                    {
                        tp->stats.tx_deferred++;
                    }
#endif

#if LINUX_VERSION_CODE > 0x20127
                    tp->stats.tx_bytes += tp->tx_ring[entry].length & 0x7ff;
#endif

                    tp->stats.collisions += ( status >> 3 ) & 15;
                    tp->stats.tx_packets++;
                }

                                /* Free the original skb. */
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,4,0)
                dev_kfree_skb_irq( tp->tx_skbuff[entry] );
#elif ( LINUX_VERSION_CODE > 0x20155 ) && ( LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0) )
                dev_kfree_skb( tp->tx_skbuff[entry] );
#else
                dev_kfree_skb( tp->tx_skbuff[entry], FREE_WRITE );
#endif
                tp->tx_skbuff[entry] = 0;
            }

#ifndef final_version
            if ( tp->cur_tx - dirty_tx > TX_RING_SIZE )
            {
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0)
                printk( KERN_ERR "%s: Out - of - sync dirty pointer, %d vs. %d, full=%d.\n", 
                        dev->name, dirty_tx, tp->cur_tx, tp->tx_full );
#else
                printk( KERN_ERR "%s: Out - of - sync dirty pointer, %d vs. %d.\n", 
                        dev->name, dirty_tx, tp->cur_tx );
#endif                
                dirty_tx += TX_RING_SIZE;
            }
#endif


#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
            if ( ( tp->cur_tx - dirty_tx )  < TX_RING_SIZE - 2 )
#else
            if ( ( tp->tx_full ) &&
                 ( dev->tbusy ) &&
                 ( tp->cur_tx - dirty_tx  < TX_RING_SIZE - 2 ) )
#endif                
            {
                /* The ring is no longer full, clear tbusy. */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
                netif_wake_queue( dev );
#else                
                tp->tx_full   = 0;
                dev->tbusy    = 0;
                mark_bh( NET_BH );
#endif
            }

            tp->dirty_tx = dirty_tx;
        }

        /* Log errors. */
        if ( csr5 & AbnormalIntr )
        {
            /* Abnormal error summary bit. */
            if ( csr5 & TxJabber )
            {
                tp->stats.tx_errors++;
            }

            if ( csr5 & TxFIFOUnderflow )
            {
                if ( ( tp->csr6 & 0xC000 ) != 0xC000 )
                {
                    printk( KERN_INFO "bumping the tx threshold\n" );
                    tp->csr6 += 0x4000;     /* Bump up the Tx threshold */
                }
                else
                {
                    printk( KERN_INFO "store-n-forward\n" );
                    tp->csr6 |= 0x00200000;  /* Store - n-forward. */
                }

                /* Restart the transmit process. */
                tp->csr6 |= OPMODE_ST;
                write_csr6( dev );

                tp->csr6 |= OPMODE_SR;
                write_csr6( dev );
            }

            if ( csr5 & RxDied )
            {
                /* Missed a Rx frame. */
                tp->stats.rx_errors++;
                tp->stats.rx_missed_errors += inl( ioaddr + CSR8 ) & 0xffff;
            }

            if ( csr5 & TimerInt )
            {
                printk( KERN_ERR "%s: Something Wicked happened! %8.8x.\n", 
                        dev->name, csr5 );
                /* Hmmmmm, it's not clear what to do here. */
            }

            /* Clear all error sources, included undocumented ones! */
            outl( 0x0800f7ba, ioaddr + CSR5 );
        }

        if ( --work_budget < 0 )
        {
            if ( spnic_debug > 1 )
            {
                printk( KERN_WARNING "%s: Too much work at interrupt, "
                        "csr5=0x%8.8x.\n", dev->name, csr5 );
            }
            
            /* Acknowledge all interrupt sources. */
            outl( 0x8001ffff, ioaddr + CSR5 );
            break;
        }
    } while ( 1 );

    if ( spnic_debug > 3 )
    {
        printk( KERN_DEBUG "%s: exiting interrupt, csr5=%#4.4x.\n", 
                dev->name, inl( ioaddr + CSR5 ) );
    }
    
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0)
    dev->interrupt = 0;
    clear_bit( 0, ( void* ) &tp->interrupt );
#endif

    return;
}


static int spnic_rx( struct device* dev )
{
    struct spnic_private*             tp              = ( struct spnic_private* ) dev->priv;
    int                               entry           = tp->cur_rx % RX_RING_SIZE;
    int                               rx_work_limit   = tp->dirty_rx + RX_RING_SIZE - tp->cur_rx;
    int                               work_done       = 0;

    
    if ( spnic_debug > 4 )
    {        
        printk( KERN_DEBUG " In spnic_rx(), entry %d %8.8x.\n", entry, 
                tp->rx_ring[entry].status );
    }

#if 0
    dumpMacRegsSubset( dev, "spnic_rx" );
#endif
    
    /* If we own the next entry, it's a new packet. Send it up. */
    while ( tp->rx_ring[entry].status >= 0 )
    {
        s32 status = tp->rx_ring[entry].status;

        if ( --rx_work_limit < 0 )
        {
            break;
        }
        
        if ( ( status & 0x0300 ) != 0x0300 )
        {
            if ( ( status & 0xffff ) != 0x7fff )
            {
                /* Ingore earlier buffers. */
                if ( spnic_debug > 1 )
                {
                    printk( KERN_WARNING "%s: Oversized Ethernet frame spanned "
                            "multiple buffers, status %8.8x!\n", 
                            dev->name, status );
                }
                tp->stats.rx_length_errors++;
            }
        }
        else if ( status & 0x8000 )
        {
            /* There was a fatal error. */
            if ( spnic_debug > 2 )
            {
                printk( KERN_DEBUG "%s: Receive error, Rx status %8.8x.\n", 
                        dev->name, status );
            }
            
            tp->stats.rx_errors++; /* end of a packet.*/
            if ( status & 0x0890 )
            {
                tp->stats.rx_length_errors++;
            }
            
            if ( status & 0x0004 )
            {
                tp->stats.rx_frame_errors++;
            }
            
            if ( status & 0x0002 )
            {
                tp->stats.rx_crc_errors++;
            }

            if ( status & 0x0001 )
            {
                tp->stats.rx_fifo_errors++;
            }
        }
        else
        {
            /* Omit the four octet CRC from the length. */
            short pkt_len = ( status >> 16 ) - 4;
            struct sk_buff* skb;

            /* Check if the packet is long enough to just accept without
               copying to a properly sized skbuff. */
            if ( pkt_len < rx_copybreak &&
                 ( skb = DEV_ALLOC_SKB( pkt_len + 2 ) ) != NULL)
            {
                skb->dev = dev;
                skb_reserve( skb, 2 );    /* 16 byte align the IP header */

#if LINUX_VERSION_CODE < 0x10300
                memcpy( skb->data, tp->rx_ring[entry].buffer1, pkt_len );
#elif LINUX_VERSION_CODE < 0x20200  || defined( __alpha__ )
                memcpy( skb_put( skb, pkt_len ), 
                        bus_to_virt( tp->rx_ring[entry].buffer1 ), pkt_len );
#else
                eth_copy_and_sum( skb, bus_to_virt( tp->rx_ring[entry].buffer1 ), 
                                  pkt_len, 0 );
                skb_put( skb, pkt_len );
#endif
                work_done++;
            }
            else
            {
                /* Pass up the skb already on the Rx ring. */
                skb = tp->rx_skbuff[entry];
                tp->rx_skbuff[entry] = NULL;
#ifndef final_version
                {
                    void* temp = skb_put( skb, pkt_len );
                    if ( bus_to_virt( tp->rx_ring[entry].buffer1) != temp )
                        printk( KERN_ERR "%s: Internal consistency error! The "
                                "skbuff addresses do not match in spnic_rx:"
                                " %p vs. %p / %p.\n", dev->name, 
                                bus_to_virt( tp->rx_ring[entry].buffer1 ), 
                                skb->head, temp );
                }
#else
                skb_put( skb, pkt_len );
#endif
            }
#if LINUX_VERSION_CODE > 0x10300
            skb->protocol = eth_type_trans( skb, dev );
#else
            skb->len = pkt_len;
#endif
            netif_rx( skb );
            dev->last_rx = jiffies;
            tp->stats.rx_packets++;
#if LINUX_VERSION_CODE > 0x20127
            tp->stats.rx_bytes += pkt_len;
#endif
        }
        entry = ( ++tp->cur_rx ) % RX_RING_SIZE;
    }

    /* Refill the Rx ring buffers. */
    for ( ; tp->cur_rx - tp->dirty_rx > 0; tp->dirty_rx++ )
    {
        entry = tp->dirty_rx % RX_RING_SIZE;
        if ( tp->rx_skbuff[entry] == NULL )
        {
            struct sk_buff* skb;
            skb = tp->rx_skbuff[entry] = DEV_ALLOC_SKB( PKT_BUF_SZ );
            if ( skb == NULL )
            {
                break;
            }
            
            skb->dev = dev;                 /* Mark as being used by this device. */
#if LINUX_VERSION_CODE > 0x10300
            tp->rx_ring[entry].buffer1 = virt_to_bus( skb->tail );
#else
            tp->rx_ring[entry].buffer1 = virt_to_bus( skb->data );
#endif
            work_done++;
        }
        tp->rx_ring[entry].status = 0x80000000;
    }

    return( work_done );
}


static int spnic_close( struct device* dev )
{
    long                       ioaddr   = dev->base_addr;
    struct spnic_private*      tp       = ( struct spnic_private* ) dev->priv;
    int                        i;


#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
    netif_stop_queue( dev );
#else
    dev->start = 0;
    dev->tbusy = 1;
#endif

    if ( spnic_debug > 1 )
    {
        printk( KERN_DEBUG "%s: Shutting down ethercard, status was %2.2x.\n", 
                dev->name, inl( ioaddr + CSR5 ) );
    }
    
    /* Disable interrupts by clearing the interrupt mask. */
    outl( 0x00000000, ioaddr + CSR7 );

    /* Stop the chip's Tx and Rx processes. */
    tp->csr6 &= ~( OPMODE_ST | OPMODE_SR );
    write_csr6( dev );

    tp->stats.rx_missed_errors += inl( ioaddr + CSR8 ) & 0xffff;

#ifdef SA_SHIRQ
    free_irq( dev->irq, dev );
#else
    free_irq( dev->irq );
    irq2dev_map[dev->irq] = 0;
#endif

    /* Free all the skbuffs in the Rx queue. */
    for ( i = 0; i < RX_RING_SIZE; i++ )
    {
        struct sk_buff* skb       = tp->rx_skbuff[i];
        tp->rx_skbuff[i]          = 0;
        tp->rx_ring[i].status     = 0;              /* Not owned by Tulip chip. */
        tp->rx_ring[i].length     = 0;
        tp->rx_ring[i].buffer1    = 0xBADF00D0; /* An invalid address. */

        if ( skb )
        {
#if LINUX_VERSION_CODE < 0x20100
            skb->free = 1;
#endif
#if ( LINUX_VERSION_CODE > 0x20155)
            dev_kfree_skb( skb );
#else
            dev_kfree_skb( skb, FREE_WRITE );
#endif
        }
    }

    for ( i = 0; i < TX_RING_SIZE; i++ )
    {
        if ( tp->tx_skbuff[i] )
        {
#if ( LINUX_VERSION_CODE > 0x20155 )
            dev_kfree_skb( tp->tx_skbuff[i] );
#else
            dev_kfree_skb( tp->tx_skbuff[i], FREE_WRITE );
#endif
        }
        
        tp->tx_skbuff[i] = 0;
    }


    MOD_DEC_USE_COUNT;

    return( 0 );
}


static struct enet_statistics* spnic_get_stats( struct device* dev )
{
    struct spnic_private*       tp        = ( struct spnic_private* ) dev->priv;
    long                        ioaddr    = dev->base_addr;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
    unsigned long               flags;
#endif
    
    
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
    if ( netif_running( dev ) )
#else
    if ( dev->start )
#endif        
    {
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
        spin_lock_irqsave( &tp->lock, flags );
#endif
        tp->stats.rx_missed_errors += inl( ioaddr + CSR8 ) & 0xffff;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
        spin_unlock_irqrestore( &tp->lock, flags );
#endif
    }

    return( &tp->stats );
}


#ifdef HAVE_PRIVATE_IOCTL
/* Provide ioctl() calls to examine the MII xcvr state. */
static int private_ioctl( struct device* dev, struct ifreq* rq, int cmd )
{
    struct spnic_private*       tp                = ( struct spnic_private* ) dev->priv;
    long                        ioaddr            = dev->base_addr;
    u16*                        data16            = ( u16* ) &rq->ifr_data;
    u32*                        data32            = ( u32* ) &rq->ifr_data;
    int                         phy               = tp->phys[0] & 0x1f;
    long                        flags;
    int                         mac_reg_offset;
    long                        mac_reg_value;
    

    switch(cmd) {
    case SIOCDEVPRIVATE:		/* Get the address of the PHY in use. */
        data16[0] = phy;
        return( 0 );

    case SIOCDEVPRIVATE + 1:		/* Read the specified MII register. */
        save_flags( flags );
        cli();
        data16[1] = mdio_read( dev, tp->phys[0], data16[0] );
        restore_flags( flags );
        return( 0 );

    case SIOCDEVPRIVATE + 2:		/* Write the specified MII register */
        if ( !suser() )
        {
            return( -EPERM );
        }
        
        save_flags( flags );
        cli();
        mdio_write( dev, tp->phys[0], data16[0], data16[1] );
        restore_flags( flags );
        return( 0 );

    case SIOCDEVPRIVATE + 3:		/* Read the specified Mac register */
        save_flags( flags );
        cli();
        mac_reg_offset = ( data32[0] * 0x8 );
        mac_reg_value  = inl( ioaddr + mac_reg_offset );
        data32[1]      = mac_reg_value;
        restore_flags( flags );
        return( 0 );

    case SIOCDEVPRIVATE + 4:		/* Write the specified Mac register */
        if ( !suser() )
        {
            return( -EPERM );
        }
        
        save_flags( flags );
        cli();
        mac_reg_offset = ( data32[0] * 0x8 );
        mac_reg_value  = data32[1];
        outl( mac_reg_value, ioaddr + mac_reg_offset );
        restore_flags( flags );
        return( 0 );
        
    default:
        return( -EOPNOTSUPP );
    }

    return( -EOPNOTSUPP );
}
#endif  /* HAVE_PRIVATE_IOCTL */

/* Set or clear the multicast filter for this adaptor.
   Note that we only use exclusion around actually queueing the
   new frame, not around filling tp->setup_frame.  This is non - deterministic
   when re - entered but still correct. */

/* The little - endian AUTODIN32 ethernet CRC calculation.
   N.B. Do not use for bulk data, use a table - based routine instead.
   This is common code and should be moved to net / core / crc.c */


static unsigned const ethernet_polynomial_le = 0xedb88320U;


static inline unsigned ether_crc_le( int length, unsigned char* data )
{
    unsigned int        crc                = 0xffffffff;
    int                 bit;
    unsigned char       current_octet;


    while ( --length >= 0 )
    {
        current_octet = *data++;

        for ( bit = 8; --bit >= 0; current_octet >>= 1 )
        {
            if ( ( crc ^ current_octet ) & 1 )
            {
                crc >>= 1;
                crc ^= ethernet_polynomial_le;
            }
            else
            {
                crc >>= 1;
            }
        }
    }
    return( crc );
}


#ifdef NEW_MULTICAST
static void set_rx_mode( struct device* dev )
#else
static void set_rx_mode( struct device* dev, int num_addrs, void* addrs )
#endif
{
    long                     ioaddr   = dev->base_addr;
    struct spnic_private*    tp       = ( struct spnic_private* ) dev->priv;
    

    tp->csr6 &= ~( OPMODE_PM | OPMODE_PR | OPMODE_IF |
                   OPMODE_HO | OPMODE_HP );

    if ( dev->flags & IFF_PROMISC )
    {
        /* Set promiscuous. */
        tp->csr6 |= ( OPMODE_PR | OPMODE_PM );

        /* Unconditionally log net taps. */
        printk( KERN_INFO "%s: Promiscuous mode enabled.\n", dev->name );
    }
    else if ( ( dev->mc_count > 1000 )  ||  ( dev->flags & IFF_ALLMULTI ) )
    {
        /* Too many to filter perfectly -- accept all multicasts. */
        tp->csr6 |= OPMODE_PM;
    }
    else
    {
        u32* setup_frm = tp->setup_frame;
        struct dev_mc_list* mclist;
        u16* eaddrs;
        u32 tx_flags;
        int i;

        if ( dev->mc_count > 14 )
        {
            /* Must use a multicast hash table. */
            u16 hash_table[32];


            memset( hash_table, 0, sizeof( hash_table ) );
            /* This should work on big - endian machines as well. */
            for ( i = 0, mclist = dev->mc_list; mclist && i < dev->mc_count;
                  i++, mclist = mclist->next )
            {
                set_bit( ether_crc_le( ETH_ALEN, mclist->dmi_addr) & 0x1ff, 
                         hash_table);
            }
            
            /* Copy the hash table to the setup frame.
               NOTE that only the LOW SHORTWORD of setup_frame[] is valid! */
            for ( i = 0; i < 32; i++)
            {
                *setup_frm++ = hash_table[i];
            }
            
            setup_frm += 7;
            tx_flags = 0x08400000 | 192;
            /* Too clever: i > 15 for fall - though. */

            tp->csr6 |= ( OPMODE_HP | OPMODE_HO );
        }
        else
        {
            /* We have <= 15 addresses so we can use the wonderful
               16 address perfect filtering of the Tulip. */
            for ( i = 0, mclist = dev->mc_list; i < dev->mc_count;
                  i++, mclist = mclist->next )
            {
                /* Note that only the low shortword of setup_frame[] is valid!
                   This code may require tweaking for non - x86 architectures! */
                eaddrs = ( u16*)mclist->dmi_addr;

                *setup_frm++ = *eaddrs++;
                *setup_frm++ = *eaddrs++;
                *setup_frm++ = *eaddrs++;
            }
            /* Fill the rest of the table with our physical address.
               Once again, only the low shortword or setup_frame[] is valid! */
            *setup_frm++ = 0xffff;
            *setup_frm++ = 0xffff;
            *setup_frm++ = 0xffff;
            tx_flags = 0x08000000 | 192;
        }

        eaddrs = ( u16*)dev->dev_addr;

        do
        {
            *setup_frm++ = eaddrs[0];
            *setup_frm++ = eaddrs[1];
            *setup_frm++ = eaddrs[2];
        } while ( ++i < 15 );

        /* Now add this frame to the Tx list. */
        if ( tp->cur_tx - tp->dirty_tx > TX_RING_SIZE - 2 )
        {
            /* Same setup recently queued, we need not add it. */
        }
        else
        {
            unsigned long flags;
            unsigned int entry, dummy = 0;

            save_flags( flags );
            cli();
            entry = tp->cur_tx++ % TX_RING_SIZE;

            if ( entry != 0 )
            {
                /* Avoid a chip errata by prefixing a dummy entry. */
                tp->tx_skbuff[entry] = 0;
                tp->tx_ring[entry].length =
                    ( entry == TX_RING_SIZE - 1) ? 0x02000000 : 0;
                tp->tx_ring[entry].buffer1 = 0;
                /* race with chip, set DescOwned later */
                dummy = entry;
                entry = tp->cur_tx++ % TX_RING_SIZE;
            }

            tp->tx_skbuff[entry] = 0;
            /* Put the setup frame on the Tx list. */
            if ( entry == TX_RING_SIZE - 1 )
            {
                tx_flags |= 0x02000000;         /* Wrap ring. */
            }
            
            tp->tx_ring[entry].length = tx_flags;
            tp->tx_ring[entry].buffer1 = virt_to_bus( tp->setup_frame );
            tp->tx_ring[entry].status = 0x80000000;

            if ( tp->cur_tx - tp->dirty_tx >= TX_RING_SIZE - 2 )
            {
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
                netif_stop_queue( dev );
#else
                dev->tbusy  = 1;
                tp->tx_full = 1;
#endif
            }
            if ( dummy >= 0 )
            {
                tp->tx_ring[dummy].status = DescOwned;
            }
            
            restore_flags( flags );

            /* Trigger an immediate transmit demand. */
            outl( 0, ioaddr + CSR1 );
        }
        write_csr6( dev );
    }
}


static void write_csr6( struct device* dev )
{
    struct spnic_private*       tp         = ( struct spnic_private* ) dev->priv;
    int                         ioaddr     = dev->base_addr;
    

    outl( tp->csr6 | OPMODE_MB1, ioaddr + CSR6 );
}


static void SendSetupFrame( struct device* dev )
{
    u16*                     eaddrs;
    u32*                     setup_frm;
    u32                      i;
    struct spnic_private*    tp;


    tp         = ( struct spnic_private* ) dev->priv;
    eaddrs     = ( u16* ) dev->dev_addr;
    setup_frm  = tp->setup_frame;
    
    /* Fill the whole address filter table with our physical address. */
    /* You must add the broadcast address when doing perfect filtering! */
    *setup_frm++ = 0xffff;
    *setup_frm++ = 0xffff;
    *setup_frm++ = 0xffff;

    /* Fill the rest of the accept table with our physical address. */
    for ( i = 1; i < 16; i++ )
    {
        *setup_frm++ = eaddrs[0];
        *setup_frm++ = eaddrs[1];
        *setup_frm++ = eaddrs[2];
    }

    /* Put the setup frame on the Tx list. */
    tp->tx_ring[0].length   = 0x08000000 | 192;
    tp->tx_ring[0].buffer1  = virt_to_bus( tp->setup_frame );
    tp->tx_ring[0].status   = 0x80000000;
    
    tp->cur_tx++;
}


static void SetPhyANAR( struct device* dev )
{
    struct spnic_private*    tp;
    int                      anar;
    

    tp = ( struct spnic_private* ) dev->priv;

    anar = mdio_read( dev, tp->phys[0], 4 );
    printk( KERN_DEBUG "%s:  Advertising %4.4x on PHY %d, "
            " previously advertising %4.4x.\n", 
            dev->name, tp->to_advertise, tp->phys[0], anar );

    mdio_write( dev, tp->phys[0], 4, tp->to_advertise );
}


static void FindMiiPhys( struct device* dev )
{
    int                      phy;
    int                      mii_status;
    int                      mii_reg0;
    int                      phy_idx;
    unsigned int             ouid;
    unsigned int             temp;
    struct spnic_private*    tp;


    tp = ( struct spnic_private* ) dev->priv;

    /* Find the connected MII xcvrs.
       Doing this in open() would allow detecting external xcvrs later, 
       but takes much time. */

    for ( phy = 0, phy_idx = 0; phy < 32 && phy_idx < sizeof( tp->phys ); phy++ )
    {
        mii_status = mdio_read( dev, phy, 1 );
 
        if ( ( mii_status != 0xffff ) && ( mii_status != 0x0000 ) )
        {
            mii_reg0 = mdio_read( dev, phy, 0 );
            tp->phys[phy_idx] = phy;
            printk( KERN_INFO "%s:  MII transceiver found at MDIO address "
                    "%d, config %4.4x status %4.4x.\n", 
                    dev->name, phy, mii_reg0, mii_status );

            /* Get Phy Organizationally Unique Identifier of the Phy
             */
            ouid = 0;
            temp = 0;
            ouid = mdio_read( dev, phy, 3 );

            /* mask off the silicon revision number */
            ouid &= ~0xf;
            temp = mdio_read( dev, phy, 2 );
            ouid |= ( temp << 16 );
            tp->phy_ouid[phy_idx] = ouid;

            phy_idx++;
        }
    }

    tp->mii_cnt = phy_idx;
    if ( tp->mtable  &&  tp->mtable->has_mii  &&  phy_idx == 0 )
    {
        printk( KERN_INFO "%s: ***WARNING***: No MII transceiver found!\n", 
                dev->name );
        tp->phys[0] = 1;
        return;
    }

    SetupPhyLeds( dev );
}


static void ResetPhy( struct device* dev )
{
    struct spnic_private*       tp;
    int                         mii_reg0;
    

    tp = ( struct spnic_private* ) dev->priv;
    

    if ( tp->phy_ouid[0] == PHY_LXT970AD )
    {
        mdio_write( dev, tp->phys[0], 0, 0x8000 );

        /* 15us is typical, wait 50 just in case */
        udelay( 50 );
    }
    else if ( tp->phy_ouid[0] == PHY_LXT971 )
    {
        mdio_write( dev, tp->phys[0], 0, 0x8000 );

        do {
            mii_reg0 = mdio_read( dev, tp->phys[0], 0 );
        } while ( mii_reg0 & 0x8000 );
    }
    else
    {
        printk( KERN_INFO "%s: ResetPhy being called when there is not PHY\n", dev->name );
    }
}


static void EnableMiiInterrupts( struct device* dev )
{
    struct spnic_private*            tp = dev->priv;
    int                              mii_int_reg;

    
    /* Enable the Mii Interrupts */
    if ( tp->phy_ouid[0] == PHY_LXT970AD )
    {
        /* Enable interrupts.  This is the only way we were able
         * to ensure that value was taken by the phy
         */
        do
        { 
            mdio_write( dev, tp->phys[0], 0x11, 0x2 );
            mii_int_reg = mdio_read( dev, tp->phys[0], 0x11 );
        } while ( mii_int_reg != 0x2 );
    }
    else if ( tp->phy_ouid[0] == PHY_LXT971 )
    {
        /* Enable the interrupts */
        mdio_write( dev, tp->phys[0], 0x12, 0x0002 );

        /* Set the Interrupt mask */
        mdio_write( dev, tp->phys[0], 0x12, 0x00f2 );
    }
    else
    {
        printk( KERN_INFO "%s: EnableMiiInterrupts being called when there is not PHY\n", dev->name );
    }
}    


static void SetupPhyLeds( struct device* dev )
{
    struct spnic_private*            tp = dev->priv;

    
    /* Change the LED configuration if necessary */
    if ( tp->phy_ouid[0] == PHY_LXT971 )
    {
        switch( tp->chip_id )
        {
        case P430:
        case P100TX:
            mdio_write( dev, tp->phys[0], 0x14, 0x0742 );
            break;

        default:
            mdio_write( dev, tp->phys[0], 0x14, 0x2142 );
            break;
        }
    }
}


#ifdef MODULE
#if LINUX_VERSION_CODE > 0x20118
MODULE_AUTHOR( "AEI Corp. <support@aei-it.com>" );
MODULE_DESCRIPTION( "AEI NIC ethernet driver" );
MODULE_LICENSE("GPL");
MODULE_PARM( debug, "i" );
MODULE_PARM( max_interrupt_work, "i" );
MODULE_PARM( reverse_probe, "i" );
MODULE_PARM( rx_copybreak, "i" );
MODULE_PARM( options, "1-" __MODULE_STRING( MAX_UNITS) "i" );
MODULE_PARM( full_duplex, "1-" __MODULE_STRING( MAX_UNITS) "i" );
#endif


/* An additional parameter that may be passed in... */
static int debug = -1;


int init_module( void )
{
    if ( debug >= 0 )
    {
        spnic_debug = debug;
        printk( "spnic_debug = %d\n", spnic_debug );
    }

    return( spnic_probe( NULL ) );
}


void cleanup_module( void )
{
    struct device* next_dev;


    /* No need to check MOD_IN_USE, as sys_delete_module() checks. */
    while ( root_spnic_dev )
    {
        next_dev = ( ( struct spnic_private* )root_spnic_dev->priv )->next_module;
        unregister_netdev( root_spnic_dev );
        release_region( root_spnic_dev->base_addr, SPNIC_TOTAL_SIZE );
#ifndef NAME_IF_PQFE
        kfree( root_spnic_dev );
#endif
        root_spnic_dev = next_dev;
    }
}

#endif  /* MODULE */

#ifdef EXTRA_DEBUG
static void dumpPhyRegs( struct device* dev, const char* string )
{
    struct spnic_private*        tp;


    tp = ( struct spnic_private* ) dev->priv;

    
    printk( KENR_DEBUG "%s:\n"
            "mdiireg 0: %x, 1: %x, 4: %x, 5: %x, 6: %x, 17: %x, 18: %x, 19: %x, 20: %x\n", 
            string,
            mdio_read( dev, tp->phys[0], 0 ), 
            mdio_read( dev, tp->phys[0], 1 ), 
            mdio_read( dev, tp->phys[0], 4 ), 
            mdio_read( dev, tp->phys[0], 5 ), 
            mdio_read( dev, tp->phys[0], 6 ), 
            mdio_read( dev, tp->phys[0], 17 ), 
            mdio_read( dev, tp->phys[0], 18 ), 
            mdio_read( dev, tp->phys[0], 19 ), 
            ( mdio_read( dev, tp->phys[0], 20 ) & 0x3b04 ) );
}


static void dumpMacRegs( struct device* dev, const char* string )
{
    int ioaddr;


    ioaddr = dev->base_addr;
    
    printk( KERN_DEBUG "%s:\n"
            "csr0: %x, csr1: %x, csr2: %x, csr3: %x, csr4: %x, \n"
            "csr5: %x, csr6: %x, csr7: %x, csr8: %x, \n"
            "csr9:  %x, csr10: %x, csr11: %x, \n"
            "csr12: %x, csr13: %x, csr14: %x, csr15: %x\n", 
            string,
            inl( ioaddr + CSR0 ), inl( ioaddr + CSR1 ), 
            inl( ioaddr + CSR2 ), inl( ioaddr + CSR3 ), 
            inl( ioaddr + CSR4 ), inl( ioaddr + CSR5 ), 
            inl( ioaddr + CSR6 ), inl( ioaddr + CSR7 ), 
            inl( ioaddr + CSR8 ), inl( ioaddr + CSR9 ), 
            inl( ioaddr + CSR10 ), inl( ioaddr + CSR11 ), 
            inl( ioaddr + CSR12 ), inl( ioaddr + CSR13 ), 
            inl( ioaddr + CSR14 ), inl( ioaddr + CSR15 ) );
}


static void dumpMacRegsSubset( struct device* dev, const char* string )
{
    int ioaddr;


    ioaddr = dev->base_addr;
    
    printk( KERN_DEBUG "%s:\n"
            "csr5: %x, csr6: %x\n",
            string,
            inl( ioaddr + CSR5 ), inl( ioaddr + CSR6 ) );
}
#endif /* EXTRA_DEBUG */

