/*+M*************************************************************************
 * Promise FastTrak device driver for Linux.
 *
 * Copyright (c) 2002  Promise Technology, Inc.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * --------------------------------------------------------------------------  
 * Copyright (c) 1999-2002 Promise Technology, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions, and the following disclaimer,
 *    without modification, immediately at the beginning of the file.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * Where this Software is combined with software released under the terms of 
 * the GNU Public License ("GPL") and the terms of the GPL would require the 
 * combined work to also be released under the terms of the GPL, the terms
 * and conditions of this License will apply in addition to those of the
 * GPL with the exception of any terms or conditions of this License that
 * conflict with, or are expressly prohibited by, the GPL.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *   		
 *-M*************************************************************************/

 /*********************************************************************
 *   Artemis Driver for NAS of ConnectStorII
 *   If you would like to run FastTrak , please only un-define "Artemis".
 *   Author  : Linus Chen
 *   History :
 *	02/06/01	v1.20 build 3	
 *	   1.	#define ResetAllDrives(x) x 
 *		When you want to clear all drives on debugging.
 *	   2.	Ioctl supports 380A,380B,380C
 *	   3.	Set feature twice PIO/(U)DMA 
 *	   4.	Disable drive interrupt(nIEN) on AccessReservedSector()
 *	   5.	Delete partition table(MBR) when system re-defined
 *	03/21/01	v1.20 build 3.4
 *	   1.	Fix Ioctl 380C - delay some time to set new RAID mode
 *		if the channel is busy.
*********************************************************************/
   
#include "wrapper.h"
#include "global.h"
//#define Artemis 			// Carefully !!
 
#ifdef	Artemis
#define	ResetAllDrives(x)		// Carefully !!
#define	ADEB(x)			 x
         
#include <linux/malloc.h>

extern  int input_data(unchar dev,void *buffer,unsigned wcount);
extern	ulong	wReservedSectorLBA(ByTE,ByTE,ByTE);
extern  int ft_copy_info(ft_proc_t *info, char *fmt, ...);

struct defdritab {
   unchar	Mpresent;
   unchar	redefarr;
   unchar	newdri;
   unchar 	Oblraid;
   unchar	Obwidth;
   unchar	set;	// 1-(Nblraid,Nbwidth) ,2-clear Reserved Sectors
   unchar	Nblraid;
   unchar	Nbwidth;
} DefDriTab[2] = { {0},{0} };	// Master drive of x channel and Initial
struct defdritab *T0 = &DefDriTab[0] ;
struct defdritab *T1 = &DefDriTab[1] ;
struct artemisresult {
   unchar	conflict;	// arrays previously defined conflict
   unchar 	Pblraid;	// Primary/Master RAID mode
   unchar	Pbwidth;
   unchar	Sblraid;	// Secondary/Master RAID mode
   unchar	Sbwidth;
} ArtemisResult = { 0 } ;
/*********************************************************************
 * Function:
 *   char AccessReservedSector(unchar fun,void *position,void *args)
 *
 * Description:
 *   Access Reserved Sector ( 0 - 4 )
 *	fun : 	0 - read data					
 *		1 - write data	
 *	position :				
 *		Drive x : ReservedSector y ---- "x:y"		
 *		Drive x : Master Boot Record -- "x:M"	
 * Return:
 *	0 : success
 *	1 : drive not exist
 *	-1: access fail
 ********************************************************************/
char AccessReservedSector(unchar fun,void *position,void *args) {
   char		error  = 1 ;
   ulong	ul,flags,lba;
   unchar	status;
   unchar	*str1  = (unchar *) position;
   ushort	*str2  = (ushort *) args;
   unchar	dev    = str1[0] - '0';
   unchar	sector = str1[2] - '0';
   ushort	base   = ft_drive[dev].pchannel->base;
   ushort	ctrl   = ft_drive[dev].pchannel->ctrl;
   ft_drive_t 	*drive = &ft_drive[dev] ;

   ft_outb( drive->select , base+6 ) ;
   for( ul=0 ; ul <= 1000 ; ul++ ) {	
	status = ft_inb( ctrl ) ; 
//	printk("select cmd : i=%lu status=%#x\n",ul,status);
	if( ( status & 0xD0 ) == 0x50 )		// BSY=0, DRDY=1
	   break;
	ft_udelay(1000);
	if( ul >= 1000 ) {
	   printk("AccessReservedSector: Drive%d not Exist.\n",dev);
	   return(1);	// drive[] not exist
	}
   }

   lba = (str1[2] == 'M') ? 0 : wReservedSectorLBA(dev/2,dev%2,0)+ sector;
//   ( lba ) ? 1 : printk("%s drive %d MBR !!\n",((fun)?"write":"read"),dev);
//   ft_outb( 2 , ctrl );
   save_flags(flags);
   cli();	
   ft_outb( 2 , ctrl );

   ft_outb( 1 , base+2 ) ;
   ft_outb( (unchar)lba , base+3 ) ;
   ft_outb( (unchar)(lba >> 8) , base+4 ) ;
   ft_outb( (unchar)(lba >> 16) , base+5 ) ;
   ft_outb( (unchar)(lba >> 24) | drive->select | 0x40 , base+6);

   switch ( fun ) {
	case 0 :
   	   ft_outb( 0x20 , base+7 ) ;	// ATA CMD : read a sector
   	   for( ul=0 ; ul <= 1000 ; ul++ ) {	// wait response : 1 sec
		ft_udelay(1000);
		status = ft_inb(ctrl) ; 
		if( (status & 0xD0) == 0x50 )	{	// BSY=0 DRDY=1
	   	   ft_inb(base+7) ; 
   	   	   input_data(dev,str2,512/2);
	   	   error = 0 ;
	   	   break;
		}
   	   }
	   if( error )	printk("Read Reserved Sector%d Error.\n",sector);
	   break;
	case 1 :
   	   ft_outb( 0x30 , base+7 ) ;	// ATA CMD : write a sector
//	   {
//	   	for( ul=0 ; ul<5 ; ul++ )
//   	   	   printk("%4x  ",str1[ul]);
//	   	printk("\n");
//	   };
   	   for( ul=0 ; ul <= 1000 ; ul++ ) {	// wait response : 1 sec
		ft_udelay(1000);
		if( (ft_inb(ctrl) & 0xD0) == 0x50 ) {	// BSY=0 DRDY=1
		   uint	wcount=0, ui=0 ;
	   	   ft_inb(base+7) ; 
		   while( ++wcount <= 512 ) {
			ft_outw( str2[wcount-1] , base );
		   }
		   // wait response : 1 sec
   	   	   for( ui=0 ; ui <= 1000*3 ; ui++ ) {	
			if( (ft_inb(ctrl) & 0xD0) == 0x50 ) {	
//   	   	   	   printk("Write Reserved Sector(%ld)\n",ul);
	   	   	   ft_inb(base+7) ; 
	   	   	   error = 0 ;
	   	   	   break;
			}
			ft_udelay(1000);
   	   	   }	   
	   	   break;
		}
   	   }	   
	   if( error )	printk("Write Reserved Sector%d Error.\n",sector);
	   break;
   }
   if( error )	error = -1 ;		//   drive[] not exist
	
   ft_outb( 0 , ctrl );
   restore_flags(flags);
//   ft_outb( 0 , ctrl );
   return(error);
}
/*********************************************************************
 * Function:
 *   int Fill1(char dev) ;
 *
 * Description:
 *   Read Reserved Sector 4 ,
 *   to fill out fields of newdri and redefarr and return 
 *	0 : normal
 *	1 : new drive or re-define array error
 *	2 : re-define array
 ********************************************************************/
int Fill1(char dev) {
   unchar	pos[100];
   unchar 	data[512+1];
   struct defdritab *p ;
//   int		i ;

   p = ( !dev ) ? T0 : T1 ;
   sprintf(pos,"%d:4",dev); 
   if( AccessReservedSector(0,&pos,data) )
   	printk("Read dev%d Reserved Sector 4 Error\n",dev);     
   if( !strcmp(data,"Promise NAS") ) {		// normal state
	return(0);
   }
   if( strncmp(data,"Promise NAS ",12) ) {	// new drive
	p->newdri = 1 ;
	return(1);
   }
   switch ( data[12] ) {
   	case 'I' :				// re-define Arrary
	case 'S' : 				// re-define Arrary
	case 'R' : 				// for myself test
	   p->redefarr = data[12];
	   return(2);
	case 'M' : 			
	   if( T0->Mpresent && T1->Mpresent ) {
		p->redefarr = data[12];
		return(2);			// re-define Arrary
	   }
   }

   // Re-Define Array Error !!
   printk("Checking Re-Define Array Error !!\n");
   return(1);
}
/*********************************************************************
 * Function:
 *   int Fill2(char dev , DISK *p_disk) ;
 *
 * Description:
 *   Read Reserved Sector 1 ,
 *   to fill out fields of Oblraid and Obwidth
 ********************************************************************/
int Fill2(char dev , DISK *p_disk) {
   unchar	pos[100], sec;
   unchar 	*data;
   struct defdritab *p ;
//   int		i ;

   p = ( !dev ) ? T0 : T1 ;
   sprintf(pos,"%d:1",dev); 
   data = (unchar *)p_disk;
   sec = 1 ;
   if( AccessReservedSector(0,&pos,(data+512*sec)) )
   	printk("Read dev%d Reserved Sector 1 Error\n",dev);
   p->Oblraid = p_disk->links->lun.blRAID ;
   p->Obwidth = p_disk->links->lun.bWidth ;
   return(0);
}
/*********************************************************************
 * Function:
 *   int FillOutTabH(DISK *p_disk);
 *
 * Description:
 *   Fill out some head fields of DefDriTab[2] data structure ,
 *   redefarr, newdri, Oblraid and Obwidth
 ********************************************************************/
int FillOutTabH(DISK *p_disk) {
   unchar ch, dev;
   struct defdritab *p ;
   char WRS4(unchar dev);

   ch = 0 ;
   do {
	p = &DefDriTab[ch] ;
	if( !p->Mpresent )
	   continue ;
	dev = ch * 2 ;
	if( Fill1(dev) ) ;	// fill out redefarr and newdri
//	   WRS4(dev);	// reset Reserved Sector 4 if not normal state
	else		// have been defined by NAS previously	   
	   Fill2(dev,p_disk);	// fill out Oblraid and Obwidth
   } while( ++ch < 2 ) ;   
   return(0);
}
/*********************************************************************
 * Function:
 *   void set2a0(void);
 *   void set1m2(void);
 *
 * Description:
 *   define array 2Stripe(2+0) and 2Mirror(1*2)
 ********************************************************************/
void set2a0(void) {
   T0->set     = T1->set     = 1 ;
   T0->Nblraid = T1->Nblraid = 0 ;
   T0->Nbwidth = T1->Nbwidth = 2 ;
   return;
}
void set1m2(void) {
   T0->set     = T1->set     = 1 ;
   T0->Nblraid = T1->Nblraid = 1 ;
   T0->Nbwidth = T1->Nbwidth = 2 ;
   return;
}
char CheckArrayCritical(DISK *p_disk) {
   ulong	wSize ;
   ushort	hCyl;
   unchar	bHd, bSctr;
   unchar	*data, pos[100], sec ;
   data = (unchar *)p_disk;
   sec = 1 ;
   sprintf(pos,"0:1"); 
   if( AccessReservedSector(0,&pos,(data+512*sec)) )
   	printk("CheckArrayCritical: Read dev0 Reserved Sector 1 Error\n");
   wSize = p_disk->links->lun.wSize ;
   hCyl  = p_disk->links->lun.hCyl ;
   bHd   = p_disk->links->lun.bHd ;
   bSctr = p_disk->links->lun.bSctr ;
   sprintf(pos,"2:1"); 
   if( AccessReservedSector(0,&pos,(data+512*sec)) )
   	printk("CheckArrayCritical: Read dev2 Reserved Sector 1 Error\n");
   if( wSize == p_disk->links->lun.wSize )
   	if( hCyl == p_disk->links->lun.hCyl )
	   if( bHd == p_disk->links->lun.bHd )
		if( bSctr == p_disk->links->lun.bSctr ) 
		   return(0);
   return(1);
}
#define Stripe1(x)	((x->Oblraid == 0) && (x->Obwidth == 1))
#define Stripe2(x)	((x->Oblraid == 0) && (x->Obwidth == 2))
#define Mirror2(x)	((x->Oblraid == 1))
/*********************************************************************
 * Function:
 *   int SetTwoDrives(DISK *p_disk)
 *
 * Description:
 *   Primary/Master and Secondary/Master drives exist
 *   Set T0,T1 according to combination of drives
 ********************************************************************/
#define case1	( T0->newdri && T1->newdri )
#define case2	((T0->newdri && Stripe1(T1)) || (Stripe1(T0) && T1->newdri))
#define case3	((T0->newdri && Stripe2(T1)) || (Stripe2(T0) && T1->newdri))
#define case4	((T0->newdri && Mirror2(T1)) || (Mirror2(T0) && T1->newdri))
#define case5	( Stripe1(T0) && Stripe1(T1)) 
#define case6	((Stripe1(T0) && Stripe2(T1)) || (Stripe2(T0) && Stripe1(T1)))
#define case7	((Stripe1(T0) && Mirror2(T1)) || (Mirror2(T0) && Stripe1(T1)))
#define case8	( Stripe2(T0) && Stripe2(T1)) 
#define case9	((Stripe2(T0) && Mirror2(T1)) || (Mirror2(T0) && Stripe2(T1)))
#define case10	( Mirror2(T0) && Mirror2(T1)) 

int SetTwoDrives( DISK *p_disk ) {
   unchar uc;
   struct defdritab *p ;

   // checking whether to re-define array or not
   for( uc=0 ; uc<2 ; uc++ ) {
	p = &DefDriTab[uc] ;
	if( p->redefarr ) {
	   ( p->redefarr == 'S' ) ? set2a0() : set1m2() ;
	   return(0);
	}
   }

   if( case1 || case3 )
	set1m2() ;			// first time setting mode
   else if( case2 || case4 ) {
	T0->set = ( T0->newdri ) ? 1 : 0 ;
	T1->set = ( T1->newdri ) ? 1 : 0 ;
   }
   else if( case5 || case7 || case10 ) {
	T0->set = T1->set = 0 ;		// all drives pass , not to set.
	// set conflict field: case5 case7 case10(Crital)
	ArtemisResult.conflict = ( !case10 ) ? 1 :
				 ( CheckArrayCritical(p_disk) ) ? 1 : 0 ;
	if( ArtemisResult.conflict ) {
	   ArtemisResult.Pblraid = T0->Oblraid ;
	   ArtemisResult.Pbwidth = T0->Obwidth ;
	   ArtemisResult.Sblraid = T1->Oblraid ;
	   ArtemisResult.Sbwidth = T1->Obwidth ;
	}
   }
   else if( case6 || case9 ) {		// Stripe2(x) will be cleared.
	T0->set = ( Stripe2(T0) ) ? 2 : 0 ;
	T1->set = ( Stripe2(T1) ) ? 2 : 0 ;
   }
   else if( case8 ) {
	if( CheckArrayCritical(p_disk) ) set1m2();	// Array is critical.
	else 	     T0->set = T1->set = 0 ;		// Array is OK. 	   
   }
   else {
	printk("SetTwoDrives: Artemis Internal Error !!\n");
	set1m2() ;
	return(-1);
   }
   return(0);
}

#undef 	case1
#undef 	case2
#undef 	case3
#undef 	case4
#undef 	case5
#undef 	case6
#undef 	case7
#undef 	case8
#undef 	case9
#undef 	case10
/*********************************************************************
 * Function:
 *   int FillOutTabT(DISK *p_disk);
 *
 * Description:
 *   Fill out some tail fields of DefDriTab[2] data structure ,
 *   set, Nblraid and Nbwidth and return 
 *   	-1: error
 *    	 0: success
 ********************************************************************/
int FillOutTabT(DISK *p_disk) {
   char	totaldrives = T0->Mpresent + T1->Mpresent ;
   struct defdritab *p ;

   if( !totaldrives ) 
	return(-1);

   if( totaldrives > 2 ) { 
	printk("FillOutTabT: Artemis Internal Error !!\n");
	return(-1);
   }

   // myself testing to reset all drivers.
   if((T0->redefarr == 'R') || (T1->redefarr == 'R')) {
 	T0->set = T1->set = 2 ;	
	return(1);
   }

   if( totaldrives == 1 ) {
	p = ( T0->Mpresent ) ? T0 : T1 ;
	if( !(p->redefarr || p->newdri) && (Stripe1(p) || Mirror2(p)) )
	   p->set = 0 ;
	else {
	   p->set = 1 ;
	   p->Nblraid = 0 ;
	   p->Nbwidth = 1 ;
	}
	return(0);
   }

   SetTwoDrives( p_disk );
   return(0);
}
#undef 	Stripe1(x)
#undef 	Stripe2(x)
#undef 	Mirror2(x)
// -------------------------------------------------------------------
void InitDiskSec(DISK *p_disk,unchar sec) {
   unchar 	*data;
   int		i ;
   int		start = sec*512 ;
   int		sec_size = 512 ;

   data = (unchar *)p_disk ;
   data[start] = i = 0 ;
   while ( ++i < sec_size )
	data[start+i] = data[start+i-1] - 1 ;
   return;
}
/*********************************************************************
 * Function:
 *   char CheckLBA(unchar,unchar,ulong *,ushort *,unchar *,unchar *);
 *
 * Description:
 *   normal status , return 0 .
 ********************************************************************/
char	CheckLBA(unchar RAID,unchar TotalDrives,
   	ulong *wSize,ushort *hCyl,unchar *bHd,unchar *bSctr) {
   ulong wReservedSectorLBA(ByTE,ByTE,ByTE);
   ulong size1, size2, size, tmp;

   if( TotalDrives > 1 ) {
   	if( !(ft_drive[0].present && ft_drive[2].present) ) {
   	   printk("FastTrak: CheckLBA() Error.\n");
   	   return(1) ;
   	}
   	size1 = wReservedSectorLBA(0,0,0);
   	size2 = wReservedSectorLBA(1,0,0);
   	*wSize = (size1 > size2) ? size2 : size1 ;
   	*wSize = ( RAID ) ? *wSize : *wSize * 2 ;	// Sectors
   }
   else	// one drive(dev=0) stripe
   	*wSize = wReservedSectorLBA(((T1->Mpresent)?1:0),0,0);

   size  = *wSize / 2 / 1024 ; 			// MByte
   if     ( size <= 528  )	*bHd = 16 ;
   else if( size <= 1000 )	*bHd = 32 ;
   else if( size <= 2100 )	*bHd = 64 ;
   else if( size <= 4200 )	*bHd = 128 ;
   else 			*bHd = 255 ;
   *bSctr = 63 ;
   tmp = *wSize / *bHd / *bSctr ;
   *bHd  = *bHd - 1 ;
   *hCyl = (ushort) tmp - 1 ;
   return(0);
}
/*********************************************************************
 * Function:
 *   int WRS0(unchar dev,DISK *p_disk)
 *   int WRS1(unchar dev,DISK *p_disk)
 *   int WRS2(unchar dev,DISK *p_disk)
 *   int WRS3(unchar dev,DISK *p_disk)
 *   int WRS4(unchar dev)
 *
 * Description:
 *   Write Reserved Sectors from 0 ~ 4 
 *	Reserved Sectors (0~3) 	: According to Engine Spec.
 *	Reserved Sectors (4)	: Defined by Linus Chen for NAS
 ********************************************************************/
int WRS0(unchar dev,DISK *p_disk) {
   unchar	pos[100];
   unchar 	*data, uc;
   unchar	sec=0 ;

   InitDiskSec(p_disk,sec);
   sprintf(pos,"%d:0",dev); 
   data = (unchar *)p_disk ; 	   	

   strcpy(p_disk->id.bSignature,"Promise Technology, Inc.");
   p_disk->id.wFormat = 0x20000 ;
   for( uc=0 ; uc<8 ; uc++ ) 
   	p_disk->id.bSerial[uc] = 0;	
   for( uc=0 ; uc<6 ; uc++ ) 
   	p_disk->id.bDate[uc]   = 0;	

   if( AccessReservedSector(1,&pos,(data+512*sec)) )
   	printk("Write dev%d Reserved Sector 0 Error\n",dev);
   return(0);
}
// -------------------------------------------------------------------
int WRS1(unchar dev,DISK *p_disk) {
   unchar	pos[100];
   unchar 	*data, uc;
   unchar	sec=1 ;
   // default Array0 -> RAID0 // default block = 64 KB
   unchar	RAID = DefDriTab[dev/2].Nblraid ;
   unchar	TotalDrives = DefDriTab[dev/2].Nbwidth;
   unchar	bStripeBlockPower = 7 ;

   InitDiskSec(p_disk,sec);
   sprintf(pos,"%d:1",dev);   	   	
   data = (unchar *)p_disk ; 	   	
   p_disk->links->blFlag = 0x80 ;
   p_disk->links->bReserved[0] = 0 ;
   p_disk->links->bReserved[1] = 0 ;
   p_disk->links->bReserved[2] = 0 ;

   p_disk->links->blk.blFlag = 0x07;
   p_disk->links->blk.bStream = ( TotalDrives == 1 ) ? 0 : ((dev)?1:0) ;
   p_disk->links->blk.bCh     = ( dev ) ? 1 : 0 ;
   p_disk->links->blk.bID     = 0 ;
   for( uc=0 ; uc<8 ; uc++ ) 
	p_disk->links->blk.bSerial[uc] = 0;	
   p_disk->links->blk.wSize = wReservedSectorLBA((( dev ) ? 1 : 0),0,0);
   p_disk->links->blk.wWatermark = 0xFFFFFFFF;

   p_disk->links->lun.hGeneration = 0x01;
   p_disk->links->lun.blFlag = 0x0f;
   p_disk->links->lun.blRAID = RAID ; 				
   p_disk->links->lun.bWidth = TotalDrives ;
   p_disk->links->lun.bStripeBlockPower = bStripeBlockPower ;	
   p_disk->links->lun.bLogicalWidth = RAID ? 1 : TotalDrives;
   p_disk->links->lun.bID = 0;	// Array0

   {		// Get LBA size & C/H/S
	ulong	wSize;
	ushort	hCyl;
	unchar	bHd, bSctr;
	if(CheckLBA(RAID,TotalDrives,&wSize,&hCyl,&bHd,&bSctr)){
	   printk("FastTrak : CheckLBA Error.\n");
	   return(0);
	}
//	printk("wSize/C/H/S/ = %#lx/%#x/%x/%#x\n",wSize,hCyl,bHd,bSctr);
	p_disk->links->lun.wSize = wSize ;
	p_disk->links->lun.hCyl  = hCyl  ;
	p_disk->links->lun.bHd   = bHd   ;
	p_disk->links->lun.bSctr = bSctr ;
   }
   for( uc=0 ; uc<8 ; uc++ ) 
	p_disk->links->lun.bSerial[uc] = 0;
   for( uc=0 ; uc<8 ; uc++ ) {
   	unchar	uc1;
	p_disk->links->lun.strm[uc].blFlag = ( uc >= TotalDrives ) ? 0 : 0x07 ;
	p_disk->links->lun.strm[uc].bIndex = 0;
	p_disk->links->lun.strm[uc].bCh = 
	   	   	(uc >= TotalDrives) ? 0 : ((dev) ? 1 : 0 ) ;
	p_disk->links->lun.strm[uc].bID = 0;
	for( uc1=0 ; uc1<8 ; uc1++ )
	 		p_disk->links->lun.strm[uc].bSerial[uc1] = 0 ;
   }
   if( AccessReservedSector(1,&pos,(data+512*sec)) )
   	printk("Write dev%d Reserved Sector 1 Error\n",dev);
   return(0);
}
// -------------------------------------------------------------------
int WRS2(unchar dev,DISK *p_disk) {
   unchar	pos[100];
   unchar 	*data;
   unchar	sec=2 ;

   InitDiskSec(p_disk,sec);
   sprintf(pos,"%d:2",dev); 
   data = (unchar *)p_disk ; 	   	
   if( AccessReservedSector(1,&pos,(data+512*sec)) )
   	printk("Write dev%d Reserved Sector 2 Error\n",dev);
   return(0);
}	
// -------------------------------------------------------------------
int WRS3(unchar dev,DISK *p_disk,unchar reset) {
   unchar	pos[100];
   unchar 	*data;
   unchar	sec=3 ;

   InitDiskSec(p_disk,sec);
   sprintf(pos,"%d:3",dev); 
   data = (unchar *)p_disk ; 	   	
   if( !reset ) 
   	p_disk->wChecksum = wGetChecksum((bpwBUF)p_disk,sizeof(DISK)-4);
   if( AccessReservedSector(1,&pos,(data+512*sec)) )
   	printk("Write dev%d Reserved Sector 3 Error\n",dev);     	
   return(0);
}
// -------------------------------------------------------------------
int WRS4(unchar dev,unchar reset) {
   unchar	pos[100];
   unchar 	data[512+1] = { 0 } ;
//   int		i ;

   sprintf(pos,"%d:4",dev); 
   if( !reset ) 
   	sprintf(data,"Promise NAS");
   if( AccessReservedSector(1,&pos,data) )
   	printk("Write dev%d Reserved Sector 4 Error\n",dev);     	
   return(0);
}
/*********************************************************************
 * Function:
 *   int ClearMBR(void)
 *
 * Description:
 *   Clear original disk partitions and confirm success.
 ********************************************************************/
int ClearMBR(unchar dev) {
   unchar	pos[100];
   unchar 	data[512+1] = { 0 } ;
   uint		ui ;
   char		retry, error;

   sprintf(pos,"%d:M",dev); 
   retry = 3 ;
   while( retry-- > 0 ) {
   	if( AccessReservedSector(1,&pos,data) ) {
   	   printk("Write dev%d MBR Error\n",dev);     	
	   continue;
	}
   	if( AccessReservedSector(0,&pos,data) ) {
   	   printk("Read dev%d MBR Error.\n",dev);     	
   	   if( AccessReservedSector(0,&pos,data) ) {
   	   	printk("Write dev%d MBR Error..\n",dev);     	
	   	continue;
	   }
	}
	error = 0 ;
	for( ui=0 ; ui < 512 ; ui++ ) {
	   if( data[ui] ) {
		error = 1 ;
		break ;
	   }
	}
	if( !error )	return(0) ;
   	printk("Retry to Write dev%d MBR .\n",dev);     	
   }
   return(-1);
}
/*********************************************************************
 * Function:
 *   char SetATConf(void)
 *
 * Description:
 *   Artemis Main Program 
 ********************************************************************/
char SetATConf(void) {
   unchar 	uc, dev;
   DISK		*p_disk;
   struct defdritab *p ;

//   ADEB(printk("Set Artemis Configuration !! \n"););
   MaintenanceLength = 0;
   p_disk = kmalloc(sizeof(DISK),GFP_ATOMIC);

///*
   // fill out 'Mpresent' field of global data structure DefDriTab
   T0->Mpresent = ft_channel[0].pdrive[0]->present ;
   T1->Mpresent = ft_channel[1].pdrive[0]->present ;

   // fill out fields : redefarr, newdri, Oblraid, Obwidth
//   printk("Fill out FillOutTabH() !! \n");
   FillOutTabH(p_disk); 

   // fill out fields : pass, Nblraid, Nbwidth
//   printk("Fill out FillOutTabT() !! \n");
   FillOutTabT(p_disk); 

   // for myself debugging to clear Master drives
   ResetAllDrives( T0->set = T1->set = 2 ; ) ;

   ADEB(
   	for( uc=0 ; uc<2 ; uc++ ) {
	   p = &DefDriTab[uc];
   	   printk("%d/",p->Mpresent);
	   (p->redefarr) ? printk("%c/",p->redefarr) : printk("0/");
   	   printk("%d   %d/%d   ",p->newdri,p->Oblraid,p->Obwidth);
   	   printk("%d/%d/%d\n",p->set,p->Nblraid,p->Nbwidth);
   	}  
//   	printk("ArtemisResult: %d/",ArtemisResult.conflict);
   	printk("%d/",ArtemisResult.conflict);
   	printk("%d/%d/"  ,ArtemisResult.Pblraid,ArtemisResult.Pbwidth);
   	printk("%d/%d/\n",ArtemisResult.Sblraid,ArtemisResult.Sbwidth);
   ) ;

   uc = 0 ;
   do {
	dev = uc * 2 ;
	p = &DefDriTab[uc];
	if( p->Mpresent && p->set ) {
	   unchar   reset = ( p->set == 2 ) ? 1 : 0 ;
	   if( p->redefarr )
	   	printk("ReDef-");
//	   	printk("Re-Define Array from drive%d.\n",dev);
 	   WRS0(dev,p_disk);
 	   WRS1(dev,p_disk);
   	   WRS2(dev,p_disk);
 	   WRS3(dev,p_disk,reset);
 	   WRS4(dev,reset);
	   ClearMBR(dev);
//	   printk("Write drive%d Reserved Sectors 0,1,2,3,4/MBR.\n",dev);
	   printk("RS(%d).\n",dev);
	}
   } while( ++uc < 2 ) ;
//*/

   kfree(p_disk);
   return(0);
}
/*********************************************************************
 * Function:
 *   int Ioctl380A(unchar *args)
 *   int Ioctl380B(unchar *args)
 *   int Ioctl380C(unchar *args)
 *
 * Description:
 *   380A: Get Artemis Arrays information about arrays conflict.
 *   380B: Read Reserved Sector x or MBR
 *   380C: Re-Define Array ( I, S, M, R )
 ********************************************************************/
int Ioctl380A(unchar *args) {
   unchar	*p, i;
   p = ( unchar * ) &ArtemisResult ;
   for( i=0 ; i < sizeof(struct artemisresult) ; i++ )
	args[i] = p[i] ;
   return(0);
}
// -------------------------------------------------------------------
int Ioctl380B(unchar *args) {
   unchar pos[100];
   if(  (args[0] < '0') || (args[0] >= '8') )
	return(-1);
   if( ((args[2] < '0') || (args[2] >= '5')) && (args[2] != 'M') )
	return(-2);
   sprintf(pos,args);
   if( AccessReservedSector(0,pos,args) < 0 )
   	printk("Ioctl380B Access Fail !!\n");     	
   return(0);
}
// -------------------------------------------------------------------
int Ioctl380C(unchar *args) {
   unchar pos[100], totaldrives, data[512+1]={0}, error;
   uint	  count ;

   switch( args[0] ) {
	case 'I' :		// initial(default)
	case 'S' :
	case 'M' :	
	case 'R' :	break ;	// reset drive
	default  :	return(-1) ;	
   }
   totaldrives = T0->Mpresent + T1->Mpresent ;
   if( (args[0] == 'M') && ( totaldrives != 2 ) )
	return(-2);
   sprintf(data,"Promise NAS %c",args[0]);
   error = 0 ;
   if( T0->Mpresent ) {
	count = 5*1000 ;	// max waitting time = 5 sec
	while( ft_drive[0].busy ) {
	   ft_udelay(1000);
	   printk("0");
	   if( --count == 0 )
		break;
	}
	if( !count )	return(-3);
   	sprintf(pos,"0:4");
   	if( AccessReservedSector(1,pos,data) ) 
	   error |= 0x01 ;
   }
   if( T1->Mpresent ) {
	count = 5*1000 ;	// max waitting time = 5 sec
	while( ft_drive[2].busy ) {
	   ft_udelay(1000);
	   printk("2");
	   if( --count == 0 )
		break;
	}
	if( !count )	return(-4);
   	sprintf(pos,"2:4");
   	if( AccessReservedSector(1,pos,data) )
	   error |= 0x04 ;
   }
   if( error ) {
   	   printk("Ioctl380C Access Fail, error %2x !!\n",error);    
   	   printk("--> %s !!\n",data);    
	   return(error);
   }
   printk("\n");    
   return(0);
}
// -------------------------------------------------------------------
int  ATIoctl(int cmd,unsigned char *args)	{ 
   switch( cmd ) {
	case 0x380A :	return(Ioctl380A(args)) ;
	case 0x380B :	return(Ioctl380B(args)) ;
	case 0x380C :	return(Ioctl380C(args)) ;
	default :	printk("FastTrak Internal Error !!") ;
   			return(-1); 
   }
}
/*********************************************************************
 * Function:
 *   int	GetATResult(ft_proc_t **buf) 
 *	
 * Description:
 *   Return global variable "ArtemisResult" for ConnectStor II
 *   when catting /proc/scsi/FastTrak/x 
*********************************************************************/
int	GetATResult(ft_proc_t **buf) {
   ft_copy_info(*buf,"%d",ArtemisResult.conflict);
   ft_copy_info(*buf,"%d",ArtemisResult.Pblraid);
   ft_copy_info(*buf,"%d",ArtemisResult.Pbwidth);
   ft_copy_info(*buf,"%d",ArtemisResult.Sblraid);
   ft_copy_info(*buf,"%d\n",ArtemisResult.Sbwidth);
   return(0);
}
/*********************************************************************
 * Function:
 *   int  ProductName(char *p) ; 
 *   int  ATConflict(char dev) ; 
 *	
 * Description:
 *   1. Get product's name of PROMISE (FastTrak/Artemis)
 *   2. If Artemis conflicts , drive 2 return zero.
*********************************************************************/
int  ProductName(char *p) { 
   return(ft_sprintf(p,"Artemis")); 
}
int  ATConflict(char dev) { 
   if( ArtemisResult.conflict && (dev == 2) )	return(0);   
   return(1); 
} 
#else
/*********************************************************************
 * Function:
 *   char SetArtemisConf(void)
 *   int  ArtemisIoctl(int cmd,unsigned char *args)
 *   int  ProductName(char *p)   
 *
 * Description:
 *   If "Artemis" does not be defined ,
 *   do nothing for FastTrak and return -1 or "FastTrak".
 ********************************************************************/
char SetATConf(void)          		{ return(-1); }
int  ATIoctl(int cmd,unsigned char *args)	{ return(-1); }
int  GetATResult( ft_proc_t **buf) { return(-1); } 
int  ProductName(char *p)   { return(ft_sprintf(p,"FastTrak")); }
int  ATConflict(char dev) 			{ return(1); } 
#endif

