/*+M*************************************************************************
 * Promise FastTrak device driver for Linux.
 *
 * Copyright (c) 2002  Promise Technology, Inc.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * --------------------------------------------------------------------------  
 * Copyright (c) 1999-2002 Promise Technology, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions, and the following disclaimer,
 *    without modification, immediately at the beginning of the file.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * Where this Software is combined with software released under the terms of 
 * the GNU Public License ("GPL") and the terms of the GPL would require the 
 * combined work to also be released under the terms of the GPL, the terms
 * and conditions of this License will apply in addition to those of the
 * GPL with the exception of any terms or conditions of this License that
 * conflict with, or are expressly prohibited by, the GPL.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *   		
 *-M*************************************************************************/

/*********************************************************************
* channel.c
* 08/23/00 Frank Tiernan
*       Moved all the functions exported by the channel driver
*       to this seperate file. This enables us to compile the
*       channel driver and distribute it as a binary with the 
*       engine. 
* 10/30/00 Linus Chen
*	(1) Set Double Clock (set_DClock) in channel driver. 
*	(2) To solve SMP unstable , Adding spin_lock_irqsave
*		(spin_unlock_irqrestore) on ft_handler() .
*	(3) do_ft_handler: using bus master status instead of 
*		reading fifo status in order to understand which
*		drives issue interrupts.
* 11/24/00 Linus Chen
*	Add PDC20270 option.And save flags in front of start bus 
*	master in SubmitChannel().
* 12/12/00 Linus Chen
*	Remove ide.h and scsi.h to compile as independent object 
* 02/06/01 Linus Chen : v1.20 build 3
*	When drive was offline , light RED led for GigaByte.
*	Legend - Avoid writing Primary/Secondary drives concurrently 
* 04/07/02 Myst Lin : v1.02.0.20
*       don't reset soft s/g bit when doing hardware reset (index0:1 bit9) 
* 05/23/02 Myst Lin : v1.02.0.21
*       set swap box led depend on disk status when InitChannel
* 05/27/02 Myst Lin : v1.02.0.21
*       avoid accessing two channels at the same time on 26x card
* 05/27/02 Myst Lin : v1.02.0.21
*       use index0 and data0 register of each card
*       when doing hardware reset (dual controller)
*********************************************************************/

#include "global.h"
#include "wrapper.h"


#include "promise_raid_engine/public_globals.h"
#include "promise_raid_engine/ata-3.h"

bpDTASK disktasks[FT_Max_Dev]; //use this instead of task field in 
                               //ft_drive to hide the bpDTASK structure
                               //from reverse engineering 
extern struct ft_timer_list error_timer_ch0;
extern struct ft_timer_list error_timer_ch1;
extern struct ft_timer_list error_timer_ch2;
extern struct ft_timer_list error_timer_ch3;

/*********************************************************************
 * Function:
 *   int set_DClock(unchar dev,unchar fun)
 *
 * Description:
 *   Set Double Clock(internal clock) before starting bus master when
 *   UDMA mode 3,4,5 and disable Double Clock after stoping bus master
 *   fun - 0 - disable Double Clock
 *	- 1 - enable  Double Clock
*********************************************************************/
int set_DClock(unchar dev,unchar fun)
{
   unchar	xfermode = ft_drive[dev].xfermode ;
   unchar	mode = xfermode & 0x07 ;
   unchar	doubleclock = 0 ;
   unchar 	CLOCK_STA ;
   ushort 	CLOCK_PORT ;

   if( fun ) {		// disable double clock , fun = 0 
   	if( (xfermode >> 3) == 8 ) {	// UDMA
	   if( mode > 2 ) {		// UDMA3,4,5
		doubleclock = 1 ;
	   }
	}
   }
   CLOCK_PORT = ft_drive[dev].pchannel->padapter->pchannel[0]->bm + 0x11;
   CLOCK_STA = ft_inb( CLOCK_PORT ) ;
   if( doubleclock ) {		// set DClock & enable ATAPI port
	ft_outb( CLOCK_STA | ((dev%4 >= 2)?8:2) , CLOCK_PORT );
   }
   else 			// free DClock & disable ATAPI port
	ft_outb( CLOCK_STA & ((dev%4 >= 2)?(~0x08):(~0x02)) , CLOCK_PORT );
   return(0);
}
/*********************************************************************
 * Function:
 *   int LBA48_26x(unchar dev,bpDTASK task) 
 *
 * Description:
 *   Open/Close ATAPI port for 48bits LBA address on PDC2026x
*********************************************************************/
int LBA48_26x(unchar dev,bpDTASK task) {

   ft_drive_t	*p_dev = &ft_drive[dev] ;
   ft_channel_t	*p_ch  = ft_drive[dev].pchannel ;
   ushort	BA4    = p_ch->padapter->pchannel[0]->bm ;
   ushort 	port   = (p_ch->id%2) ? BA4+0x24 : BA4+0x20 ;
   ulong	value ;

   if ( !p_dev->LBA_48bit )  {	// for 26x 48bit LBA
	return(0) ;
   }
   if( task ) { // Open ATAPI port by exporting word count
   	unchar	bCmd = task->ata_reg.cmd.bCmd ;
	ulong	WordCount = task->ata_reg.cmd.bCount << 8 ;
   	ulong	enable = ( bCmd == READ_DMA ) ? (0x05<<24) : (0x06<<24) ;
	value = enable | WordCount ;
   }	
   else  	// Close ATAPI port by disabling Ultra DMA bit
	value = 0 ;
   ft_outl( value , port );
   return(1) ;
}
/*********************************************************************
 * Function:
 *   error_handler
 *
 * Description:
 *   If FastTrak interrupts time-out, the routine will be woken up.
*********************************************************************/
void error_handler(ulong dev)
{
//   unchar	ch   = dev/2 ;
   ushort	bm   = ft_drive[dev].pchannel->bm;
   ushort	base = ft_drive[dev].pchannel->base;
   ushort	deviceid = ft_drive[dev].pchannel->padapter->deviceid ;
   bpDTASK	task = disktasks[dev];

   ft_printk("FastTrak : Drive Interrupt Time Out.(%ld)\n",dev+1);
   ft_inb(bm+2) ;
   ft_outb( ft_inb(bm)&~1 , bm ) ;	// stop bus master
   if( DID26x(deviceid) ) {
   	set_DClock(dev,0);	
	LBA48_26x( dev,NULL ) ;
   }
   ft_inb(base+7) ;
   		
   task->bStatus = tsTIMED_OUT ;	// tsTIMED_OUT = 0x80
   bResetController = 1 ;
   ft_drive[dev].lbusy = 0 ;
   return(ChannelReturn(task));
}
/*********************************************************************
*  Functions:
*    BoOL ChannelTimedout(bpDTASK p_tk)
* 
*  Description:
*    It is error handler of Channel Driver for Engine 
*    It was replaced with my function of "error_handler"
*********************************************************************/
BoOL ChannelTimedout(bpDTASK p_tk)
{
   unchar	ch   = p_tk->bCh ;
   unchar	dev  = p_tk->bCh*2 + p_tk->bID ;
   ushort	bm   = ft_drive[dev].pchannel->bm;
   ushort	base = ft_drive[dev].pchannel->base;
   ushort	deviceid = ft_drive[dev].pchannel->padapter->deviceid ;
   bpDTASK	task = disktasks[dev];

   DebEngFun(ft_printk("DebEngFun : ChannelTimedout()\n"););

   if( task != p_tk ) {
   	ft_printk("ChannelTimedout: not DEV%d task\n",dev+1);
	return(0);
   }

   if(ch == 0)	ft_del_timer(&error_timer_ch0);
   else if(ch == 1)	ft_del_timer(&error_timer_ch1);
   else if(ch == 2)	ft_del_timer(&error_timer_ch2);
   else if(ch == 3)	ft_del_timer(&error_timer_ch3);

   ft_inb(bm+2) ;
   ft_outb( ft_inb(bm)&~1 , bm ) ;	// stop bus master
   if( DID26x(deviceid) ) {
   	set_DClock(dev,0);	
	LBA48_26x( dev,NULL ) ;
   }
   ft_inb(base+7) ;
   		
   task->bStatus = tsTIMED_OUT ;	// tsTIMED_OUT = 0x80
   bResetController = 1 ;
   ft_drive[dev].lbusy = 0 ;
   ChannelReturn(task);
   return(TRUE);	
}
// -------------------------------------------------------------------
ByTE bReadErrorRegister(ByTE bCh, ByTE bID)
{
   unchar	status;
   DebEngFun(ft_printk("DebEngFun : bReadErrorRegister()\n"););
   status = ft_inb(ft_channel[bCh].base+1) ;
   ft_printk("DebEngFun : bReadErrorRegister(dev%d) = ",bCh*2+bID+1);
   ft_printk("%#x \n",status);
   return( status ); 
}
/*********************************************************************
*  Functions:
*     	WORD wReadLBA_Register(ByTE bCh,ByTE bID)
* 
*  Description:
*     	If encounter Disk Error(4133) event , 
*	the function may be called.
*********************************************************************/
WORD _call wReadLBA_Register(ByTE bCh,ByTE bID)
{
   ulong 	LBA=0 ;
   ushort	base = ft_channel[bCh].base ;
   ft_printk("wReadLBA_Register - bCh(%#x) ",bCh);
   ft_printk("bID(%#x) ",bID);
   LBA |=  ft_inb( base+3 )       ;
   LBA |=  ft_inb( base+4 ) <<  8 ;
   LBA |=  ft_inb( base+5 ) << 16 ;
   LBA |= (ft_inb( base+6 ) & 0x0F) << 24 ;
   ft_printk("LBA(%#x)\n",LBA);
   return(LBA);
}
/*********************************************************************
 * Function:
 *   ft_handler
 *
 * Description:
 *   Handler FastTrak interrupts .
*********************************************************************/
void ft_handler(int irq,void *dev_id) 
{
   unchar	bm_stat, cmd_stat, err_stat, dev, ch ;
   ushort	base ,bm ;
   ushort	deviceid ;
// unchar 	*ptr = (unchar *) dev_id;
   bpDTASK	task = NULL ;
   ft_adapter_t	*ft_ada0 = (ft_adapter_t *) dev_id ;
   uint		myirq = ft_adapter[0].irq;
  
   DebFtFun(printk("In FastTrak ISR :\n"););
   if( (!ft_ada0) || (ft_ada0->irq != myirq) ) {
   	DEB(printk("Interrupts not from my devices"););
	return;
   }
   DebFtFun(printk("In FastTrak ISR : Our Interrupt\n"););

   for( ch=0 ; ch<FT_Max_Ch ; ch++ ) {
	if( !ft_channel[ch].padapter->present )
	   break;
	base = ft_channel[ch].base;
   	bm   = ft_channel[ch].bm;
   	bm_stat = ft_inb( bm+2 );		// bus master status
   	if( (bm_stat & 0x06) != 0x04 )		
	   continue;	// interrupt happened but not from the channel
	dev = ch*2 ;
	if( ft_drive[dev].present && ft_drive[dev].lbusy ) 
	   ;
	else if( ft_drive[dev+1].present && ft_drive[dev+1].lbusy )
	   dev = dev+1 ;
	else {
	   ft_inb( ft_channel[ch].base+7 ) ;
	   continue;	// interrupt happened but not from the drives
	}
   	if( bm_stat & 0x01 ) {	
	   DEB(ft_printk("FastTrak: bus master still active.\n"););
	   continue ;	// interrupt happened but bus master active
	}
	if(ch == 0)		ft_del_timer(&error_timer_ch0);
	else if(ch == 1)	ft_del_timer(&error_timer_ch1);
	else if(ch == 2)	ft_del_timer(&error_timer_ch2);
	else if(ch == 3)	ft_del_timer(&error_timer_ch3);
   	task = disktasks[dev];
	// tsNO_ERROR = 0x04
	// ataDRIVE_READY = 0x40 ;
   	task->bStatus = tsNO_ERROR;  
   	task->ata_reg.cmd.bStatus = ataDRIVE_READY;
   
   	ft_outb( ft_inb(bm+2) | 0x06 , bm+2 ) ;	// clear BM status
//   	ft_outb( ft_inb(bm)&~1 , bm ) ;	// stop bus master
	deviceid = ft_channel[ch].padapter->deviceid ;
   	if( DID26x(deviceid) ) {
   	   set_DClock( dev,0 ) ;	
	   LBA48_26x( dev,NULL ) ;
	}
   	cmd_stat = ft_inb( base+7 );		// 1F7 status
   	err_stat = ft_inb( base+1 );		// 1F1 status
   	DEB(ft_printk("bm/base=%#x/%#x\n",bm,base));
   	if( (bm_stat  &  0x07) != 0x04 ) {	// Normal outputs == 0x04
   	   ft_printk("FastTrak : Bad bm_status = %#x\n",bm_stat);
   	   task->bStatus = tsTIMED_OUT ;
   	}
   	if( (cmd_stat & ~0x16) != 0x40 ) { 	// Normal outputs == 0x40
	   ulong	wLBA = task->ata_reg.cmd.wLBA ;
	   unchar	bCmd = task->ata_reg.cmd.bCmd ;
   	   ft_printk("FastTrak : Register Status - BM(%#x)",bm_stat);
   	   ft_printk(" CMD(%#x)",cmd_stat);
   	   ft_printk(" Error(%#x)\n",err_stat);
	   if( (ft_drive[dev].downmode = err_stat & (1<<7)) ) {	// ECC 
           	log_event(4119,ft_drive[dev].arr,dev,wLBA,0,bCmd,cmd_stat);
   	   	task->bStatus = tsTIMED_OUT ;
	   }
	   else {		// non disk ECC error occurs
           	log_event(4133,ft_drive[dev].arr,dev,wLBA,0,bCmd,cmd_stat);
   	   	task->bStatus = tsXFER_ERROR ;	// encounter BAD sectors
	   }
   	   task->ata_reg.cmd.bStatus = cmd_stat;	
   	}
   	if( task->bStatus == tsTIMED_OUT ) 
	   bResetController = 0;
   	ft_drive[dev].lbusy = 0 ;
   	ChannelReturn(task);
   }
   return;
}
/*********************************************************************
*  Functions:
*	unchar outcmd16(bpDTASK p_tk);   	   
* 
*  Description:
*    	Subroutines for Channel & OS drivers
*********************************************************************/
unchar outcmd16(bpDTASK p_tk)
{
   unchar 	dev = p_tk->bCh*2 + p_tk->bID ;
   ft_drive_t 	*drive=&ft_drive[dev];
   ushort 	base = ft_channel[p_tk->bCh].base;
//   ushort 	ctrl = ft_channel[p_tk->bCh].ctrl;

   if( Check1F7(dev,1000,0,BSY) )	
	return(1);

   ft_outb( drive->select , base+6 );
   ft_outb( drive->select , base+6 );
   if( Check1F7(dev,1,DRDY|SEEK,BSY|DF) )	
	return(2);

   if( CheckChannelBus(p_tk->bCh) )
	return(3);
	
   if ( drive->LBA_48bit ) {
	// 48 bits-previous , set HOB to 1
//	ft_outb( 0x80 , ctrl ) ;	
	// engine only support ByTe bCount;
   	ft_outb( 0 , base+2 ) ;		
	// engine only support 32 bit LBA
   	ft_outb( p_tk->ata_reg.b1Fx[3] , base+3 ) ;
	ft_outb( 0 , base+4 ) ;
	ft_outb( 0 , base+5 ) ;

	// 48 bits-current , set HOB to 0
//	ft_outb( 0x00 , ctrl ) ;	
   	ft_outb( p_tk->ata_reg.b1Fx[5] , base+2 ) ;
   	ft_outb( p_tk->ata_reg.b1Fx[0] , base+3 ) ;
   	ft_outb( p_tk->ata_reg.b1Fx[1] , base+4 ) ;
   	ft_outb( p_tk->ata_reg.b1Fx[2] , base+5 ) ;
	ft_outb( drive->select | 0x40 , base+6 ) ;
	ft_outb( drive->select | 0x40 , base+6 ) ;
   }
   else {
    	ft_outb( p_tk->ata_reg.b1Fx[4] , base+1 ) ;
   	ft_outb( p_tk->ata_reg.b1Fx[5] , base+2 ) ;
   	ft_outb( p_tk->ata_reg.b1Fx[0] , base+3 ) ;
   	ft_outb( p_tk->ata_reg.b1Fx[1] , base+4 ) ;
   	ft_outb( p_tk->ata_reg.b1Fx[2] , base+5 ) ;
   	ft_outb( p_tk->ata_reg.b1Fx[3] | drive->select | 0x40 , base+6 ) ;
   	ft_outb( p_tk->ata_reg.b1Fx[3] | drive->select | 0x40 , base+6 ) ;
   }
   return(0);
}
/*********************************************************************
*  Functions:
*	int HS_Reset( ByTE bType,ByTE bCh )
* 
*  Description:	Hard/Soft reset
*	reture value -
*	0  : reset complete
*	-1 : argument not correct
*	-2 : channel bus abnormal
*********************************************************************/
int HS_Reset( ByTE bType,ByTE bCh ) {
   ft_channel_t	*channel = &ft_channel[bCh] ;
   unchar	status ,dev ;
   unchar	pio_mode, dma_mode, udma_mode;
   int		i , error;

   if( !channel->padapter )
	return(-1) ;			// adapter not exist
   ft_printk("FastTrak : RESET Channel%d ",bCh+1);
   switch( bType ) {
	case bHARD_RESET :		// 0x01
	   if( DID27x(channel->padapter->deviceid) ) {
		/* use index0 and data0 register of each card */
		ushort	index0 = ft_channel[(bCh/2)*2].bm + 1 ;
		ushort	data0  = ft_channel[(bCh/2)*2].bm + 3 ;
		/* don't touch bit9: soft s/g engine reset bit */
		unchar	hrst_bit = ( bCh%2 ) ? 0xFB : 0xF7 ;
		ft_outb( 1 , index0 );
		status = ft_inb( data0 );
		ft_outb(  (status & hrst_bit) , data0 );
		ft_udelay(10);		// 10us
		ft_outb( status , data0 );
	   	ft_printk("* ");
		break;
	   }
	case bSOFT_RESET :		// 0x02
	   ft_outb( 4 , ft_channel[bCh].ctrl);
	   ft_udelay(10);		// 10us
	   ft_outb( 0 , ft_channel[bCh].ctrl);
	   ft_printk("- ");
	   break;
	default :
	   return(-1);
   }

   for(i=0;i<200;i++) 		// wait reset 200ms
	ft_udelay(1000);

   // Initialize Hard Drives of the channel
   for( dev = bCh*2 ; dev < bCh*2+2 ; dev++ ) {
   	pio_mode  = dma_mode  = udma_mode = 0xFF ;
	error = 0 ;
	if( !ft_drive[dev].present )
	   continue;
	if( channel->abnormality )
	   error = channel->abnormality ;		
	else if( !(error=Check1F7(dev,3*1000,0,BSY)) ) {
   	   ft_outb( ft_drive[dev].select , ft_channel[bCh].base+6 ) ;
   	   ft_outb( ft_drive[dev].select , ft_channel[bCh].base+6 ) ; // for 277
	   if( !(error=Check1F7(dev,3,DRDY|SEEK,BSY|DF)) )
	      	if( !(error=set_feature(dev,&pio_mode,&dma_mode,&udma_mode)) )
	   	   set_speed( dev, pio_mode, dma_mode, udma_mode);
	}
	if( error ) {
	   ft_printk("DEV%d(Fail) ",dev+1);
	   ft_drive[dev].lpending = 1 ;
	   continue;	// drive[] not exist
   	}	   
	ft_printk("DEV%d(OK) ",dev+1);
	ft_drive[dev].lpending = 0 ;
   }	
   ft_printk("\n");
   return(0);
}
/*********************************************************************
*  Functions:
*	void InitChannel(ByTE bType,ByTE bCh,bpDTASK p_tk)
* 
*  Description:
*    	Channel Driver
*********************************************************************/
void InitChannel(ByTE bType,ByTE bCh,bpDTASK p_tk)
{
   unchar	dev, ch, uc ;
   ulong	flags;
   int		ATConflict(char dev);

   DebEngFun(ft_printk("DebEngFun : InitChannel-bType(%d)",bType););
   DebEngFun(ft_printk("-bCh(%d)\n",bCh););
   switch( bType ) {		// 0x00
	case bPOWER_UP :
	   for( dev = 0 ; dev < FT_Max_Dev ; dev++ ) 
	   	blDiskFlag[dev/2][dev%2] = (ATConflict(dev)) ? 
                                   ft_drive[dev].present : 0 ;
	   break;
	case bHARD_RESET :			// 0x01
	case bSOFT_RESET :			// 0x02
	   if( DID27x(ft_channel[bCh].padapter->deviceid) )
		bType = bHARD_RESET ;
	   if( bCh == bALL_CHANNELS ) 
		for( ch=0 ; ch < FT_Max_Ch ; ch++ )
		   HS_Reset(bType,ch);
	   else
		   HS_Reset(bType,bCh);

	   save_flags(flags);
   	   cli();
	   /* set swap box led depend on disk status */
	   for( uc=0 ; uc<2 ; uc++ ) {
		unchar 		dev = bCh*2 + uc ;
		ft_drive_t	*drive = &ft_drive[dev] ;
	   	if( drive->light ) {
		   CtrlLED(dev, drive->light);
		}
		else if (!drive->present) {
		   CtrlLED(dev, LED_Red);
		}
		else if (drive->rebuilding == 1) {
		   CtrlLED(dev, LED_Amber);
		}
		else {
		   CtrlLED(dev, LED_Green);
		}
	   }
   	   restore_flags(flags);
	   break;
	default :
	   ft_printk("FastTrak : InitChannel argument error\n");
   }
   return;
} 
/*********************************************************************
*  Functions:
*	void SubmitChannel(bpDTASK task)
* 
*  Description:
*    	Channel Driver
*********************************************************************/
void SubmitChannel(bpDTASK task)
{
   unchar 	dev = task->bCh*2 + task->bID ;
   unchar	reading = 0 ; 
   ft_drive_t	*drive	 = &ft_drive[dev] ;
   ft_channel_t	*channel = drive->pchannel ;
   ushort	deviceid = channel->padapter->deviceid ;
   ushort 	base = channel->base;
   ushort 	bm   = channel->bm;
   unchar	bCmd = task->ata_reg.cmd.bCmd ;
   ulong	flags;

   DebEngFun(ft_printk("DebEngFun : SubmitChannel(dev%d) .\n",dev););
   if( drive->lbusy ) 
	ft_printk("SubmitChannel Error , drive%d is busy.\n",dev);
   disktasks[dev] = task ;
   task->ata_reg.cmd.bStatus = ataTASK_NOT_COMPLETED ;

   if( (bCmd == READ_PIO) || (bCmd == WRITE_PIO) ) 
	if(task->blCmdFlag & 1) 
	   bCmd = (bCmd == READ_PIO) ? READ_DMA : WRITE_DMA ;
   task->ata_reg.cmd.bCmd = bCmd ;

   if( (bCmd != READ_DMA) && (bCmd != WRITE_DMA) ) {
	printk("SubmitChannel: Unknown CMD (%#X/%#x)\n",bCmd,task->blCmdFlag);
	return ;
   }
   if ( drive->LBA_48bit ) 
	bCmd = (bCmd == READ_DMA) ? READ_DMA_EXT : WRITE_DMA_EXT ;	
   save_flags(flags);
   cli();
   switch(bCmd) {
	case READ_DMA      :				// 0xC8
	case READ_DMA_EXT  : 	reading = (1<<3) ; 	// 0x25  
	case WRITE_DMA     :				// 0xCA
	case WRITE_DMA_EXT :				// 0x35
	   if( outcmd16(task) ) {	// outcmd 1F1-1F6
		task->bStatus = tsSELECTION_TIMEOUT ;
		bResetController = 1;
		if( drive->lpending ) {
//	   	   blDiskFlag[dev/2][dev%2] = 0 ;
		   ft_drive[dev].present = ft_drive[dev].xfermode=0 ;
		   if( channel->abnormality )
		      ft_printk("FastTrak : Bus Error(%d)\n",task->bCh+1);
       		   log_event(4123,ft_drive[dev].arr,dev,0,0,0,0);
		   CtrlLED(dev, LED_Red);
		}
		restore_flags(flags);
		ChannelReturn(task);	
		return;
	   }
	   drive->lpending = 0 ;
	   switch (task->bCh)  { 
		case 0:
   	   	   ft_init_timer(&error_timer_ch0);
	   	   error_timer_ch0.expires = ft_get_jiffies() + 5 * HZ ;
	   	   error_timer_ch0.data = dev ;
	   	   error_timer_ch0.function = error_handler ;
	   	   ft_add_timer(&error_timer_ch0);
		   break;
		case 1:
   	   	   ft_init_timer(&error_timer_ch1);
	   	   error_timer_ch1.expires = ft_get_jiffies() + 5 * HZ ;
	   	   error_timer_ch1.data = dev ;
	   	   error_timer_ch1.function = error_handler ;
	   	   ft_add_timer(&error_timer_ch1);
		   break;
		case 2:
   	   	   ft_init_timer(&error_timer_ch2);
	   	   error_timer_ch2.expires = ft_get_jiffies() + 5 * HZ ;
	   	   error_timer_ch2.data = dev ;
	   	   error_timer_ch2.function = error_handler ;
	   	   ft_add_timer(&error_timer_ch2);
		   break;
		case 3:
   	   	   ft_init_timer(&error_timer_ch3);
	   	   error_timer_ch3.expires = ft_get_jiffies() + 5 * HZ ;
	   	   error_timer_ch3.data = dev ;
	   	   error_timer_ch3.function = error_handler ;
	   	   ft_add_timer(&error_timer_ch3);
		   break;
		default :
		   ft_printk("SubmitChannel: channel %d error\n",task->bCh);
	   }	// switch(task->bCh) end !!
	   drive->lbusy = 1 ;
	   ft_outl( task->wPRD , bm+4 );		// PRDTable address
	   DEB(ft_printk("wPRD : %lx\n",task->wPRD));
	   ft_outb( reading , bm );		// read/write
	   ft_outb( ft_inb(bm+2)|0x06 , bm+2 ); 	// clear status bits
//   	   ft_outb( task->ata_reg.b1Fx[6] , base+7 );	// cmd 1F7
   	   ft_outb( bCmd , base+7 );	// cmd 1F7
   	   if( DID26x(deviceid) ) {
	      	set_DClock( dev,1 ) ;	
		LBA48_26x( dev,task ) ;
		if( (bCmd == 0x30) || (bCmd == 0xCA) ) {
		   uint		ui, CH;
		   unchar	ret;
		   if( (task->bCh)%2 )	CH=task->bCh-1;
		   else			CH=task->bCh+1;
		   /* check the other channel bm status */
		   /* don't access two channels at the same time */
		   for ( ui=0 ; ui<10000000 ; ui++ ) {
		   	ret = ft_inb(ft_channel[CH].bm) ;
		   	if( ret != 1 ) 
			   break;
			ft_udelay(1);
		   }
		   if( ui >= 10000000 )
			   ft_printk("BM busy(drive %d)",dev);

	        }
	   }
	   ft_outb( ft_inb(bm)|0x01 , bm );	// start bus master
	   break;
	default   :	
	   ft_printk("Unknown command : %#x in SubmitChannel().\n",bCmd);
	   break;

   }	// switch(task->ata_reg.cmd.bCmd) end !!
   restore_flags(flags);
   return;
} 
// -----------------------------------------------
