/*+M*************************************************************************
 * Promise FastTrak device driver for Linux.
 *
 * Copyright (c) 2002  Promise Technology, Inc.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * --------------------------------------------------------------------------  
 * Copyright (c) 1999-2002 Promise Technology, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions, and the following disclaimer,
 *    without modification, immediately at the beginning of the file.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * Where this Software is combined with software released under the terms of 
 * the GNU Public License ("GPL") and the terms of the GPL would require the 
 * combined work to also be released under the terms of the GPL, the terms
 * and conditions of this License will apply in addition to those of the
 * GPL with the exception of any terms or conditions of this License that
 * conflict with, or are expressly prohibited by, the GPL.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *   		
 *-M*************************************************************************/
/*********************************************************************
* ftlog.c
* 09/01/00 Frank Tiernan
* 	System Log for FastTrak RAID Adapter
* 	Based on Event Log for FastCheck Application
* 10/30/00 Linus Chen
* 	Not complete all error cases.
* 11/24/00 Linus Chen
* 	Message reported extend to support four channels(two adapters)
* 12/12/00 Linus Chen
*	Remove ide.h 
* 02/06/01 Linus Chen : v1.20 build 3
*	Add log_snmp_event() to support Ioctl(0x3806) Interface
*********************************************************************/

#include <linux/kernel.h>
#include <linux/string.h>

#include "global.h"

// -------------------------------------------------------------------
void modelname(char *str,int size) {
   char	  *p;
   unchar uc;

   for( uc=0 ; uc< (size/2) ; uc++ ) {
	char	tmp;
	tmp = str[uc*2] ;
	str[uc*2]   = str[uc*2+1] ;
	str[uc*2+1] = tmp ;
   }
   p = str ;
   do {
	if( (*p == ' ') && (*(p+1) == ' ') ) {
	   *p = '\0';
	   break;
	}
   } while( ++p < (str+size-1) );
   return;
}
// -------------------------------------------------------------------       
void log_snmp_event (int sequence, int eventid, int arrayid, int diskid) {
   //don't log more than 50 events at a time to prevent running out of
   //space if ioctl is not called. ioctl resets snmp_log_index
   //once all events have been sent.
   if (snmp_log_index < 49) {
    	snmp_log[snmp_log_index].sequence  = sequence;
    	snmp_log[snmp_log_index].eventid   = eventid;
    	snmp_log[snmp_log_index].arrayid   = arrayid;
    	snmp_log[snmp_log_index++].diskid  = diskid;
   }    
}  
// -------------------------------------------------------------------
void log_event(int eventid, int ArrayID, int dev, unsigned long LBA, int percent, short command, short status) {
   static int num = 0; 
   struct ft_driveid 	*id ; 
   char diskname[100], str[100];

   num++;
   ArrayID++;
   id = ft_drive[dev].identify ;
   strcpy(diskname, id->model);
   modelname((char *)&diskname,sizeof(id->model));
   diskname[sizeof(id->model)-1] = '\0' ; 
   sprintf( str, " (Channel %d/" , ft_drive[dev].pchannel->id+1 );
   strcat(diskname, str );
   strcat(diskname,(ft_drive[dev].select == 0xA0)?"Master)":"Slave)");
  
   //log the event for retreiva thru ioctl
   log_snmp_event(num,eventid, ArrayID, dev+1);

   switch (eventid) {
      case 4103: printk(KERN_ERR "FastTrak|%3d|4103|Warning|Disk Rebuild Started: Array %d, Disk %s\n", num, ArrayID, diskname);
                 break;
      case 4104: printk(KERN_ERR "FastTrak|%3d|4104|Information|Disk Synchronize Started: Array %d\n", num, ArrayID);
                 break;     
      case 4119: printk(KERN_ERR "FastTrak|%3d|4119|Error|ECC Error:Disk %s at LBA 0x%lx\n", num, diskname, LBA);
                 break; 
      case 4120: printk(KERN_ERR "FastTrak|%3d|4120|Error|Disk Set Down: %s\n", num, diskname);
                 break;     
      case 4121: printk(KERN_ERR "FastTrak|%3d|4121|Information|Disk Plugged in: %s\n", num, diskname);
                 break;     
      case 4122: printk(KERN_ERR "FastTrak|%3d|4122|Information|Reserved Sector Error: Disk %s at LBA 0x%lx\n", num, diskname, LBA);
                 break;     
      case 4123: printk(KERN_ERR "FastTrak|%3d|4123|Error|Disk Offline: %s\n", num, diskname);
                 break;     
      case 4124: printk(KERN_ERR "FastTrak|%3d|4124|Warning|Array %d has bad sector on Disk %s at LBA 0x%lx\n", num, ArrayID, diskname, LBA);
                 break;     
      case 4125: printk(KERN_ERR "FastTrak|%3d|4125|Warning|Retry Disk Access : Disk %s LBA 0x%lx\n", num, diskname, LBA);
                 break;     
      case 4126: printk(KERN_ERR "FastTrak|%3d|4126|Information|Array %d Rebuild Completed on Disk %s\n", num, ArrayID, diskname);
                 break;
      case 4127: printk(KERN_ERR "FastTrak|%3d|4127|Information|Array %d, Synchronize Completed\n", num, ArrayID);
                 break;     
      case 4128: printk(KERN_ERR "FastTrak|%3d|4128|Warning|Rebuild aborted on array %d, disk %s, at %d percent complete\n", num, ArrayID, diskname, percent);
                 break;
      case 4129: printk(KERN_ERR "FastTrak|%3d|4129|Warning|Synchronization aborted on array %d, at %d complete\n", num, ArrayID, percent);
                 break;     
      case 4130: printk(KERN_ERR "FastTrak|%3d|4130|Warning|Retry Disk Access : Disk %s LBA 0x%lx\n", num, diskname, LBA);
                 break;     
      case 4131: printk(KERN_ERR "FastTrak|%3d|4131|Error|Array %d, Status Offline\n", num, ArrayID);
                 break;     
      case 4132: printk(KERN_ERR "FastTrak|%3d|4132|Warning|Array %d, Status Critical\n", num, ArrayID);
                 break;     
      case 4133: printk(KERN_ERR "FastTrak|%3d|4133|Information|Array %d, Disk %s Disk Error at LBA 0x%lx cmd=0x%x status=0x%x\n", num, ArrayID, diskname, LBA, command, status);
                 break;     
      case 4134: printk(KERN_ERR "FastTrak|%3d|4134|Information|Synchronization Comparison Error on Array %d a LBA 0x%lx\n", num, ArrayID, LBA);
                 break;     
      case 4136: printk(KERN_ERR "FastTrak|%3d|4136|Warning|Disk %s has Error while startup\n", num, diskname);
                 break;     
      case 4137: printk(KERN_ERR "FastTrak|%3d|4137|Warning|Array %d is Critical while startup\n", num, ArrayID);
                 break;     
      case 4138: printk(KERN_ERR "FastTrak|%3d|4138|Error|ECC Error Fix Failed - Uncorrectable Error\n", num);
                 break;     


   }
}
       
