/*+M*************************************************************************
 * Promise FastTrak device driver for Linux.
 *
 * Copyright (c) 2002  Promise Technology, Inc.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * --------------------------------------------------------------------------  
 * Copyright (c) 1999-2002 Promise Technology, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions, and the following disclaimer,
 *    without modification, immediately at the beginning of the file.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * Where this Software is combined with software released under the terms of 
 * the GNU Public License ("GPL") and the terms of the GPL would require the 
 * combined work to also be released under the terms of the GPL, the terms
 * and conditions of this License will apply in addition to those of the
 * GPL with the exception of any terms or conditions of this License that
 * conflict with, or are expressly prohibited by, the GPL.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *   		
 *-M*************************************************************************//*********************************************************************
* os.c
* 10/30/00 Linus Chen
*       Moved all the functions exported by the os driver to this 
*	file, and declare array to save SCSI commands instead of 
*	allocating memory dynamically. 
* 11/08/00 Linus Chen
*	Fix SwapBox power control 
* 11/24/00 Frank Tiernan
*	Add reset_controller() to fix bonnie problem called by
*	fasttrak_reset().And Adding 'pending' member to struct REQ_head.
* 12/12/00 Linus Chen
*	Remove ide.h and scsi.h to compile as independent object 
* 02/06/01 Linus Chen	v1.20 build 3
*	Add verification of Promise controller on "SCSI INQUIRY" 
*	to prevent from the partial source released to be used on
*	other IDE controller.
* 02/01/02 Myst Lin     v1.20 build 16
*       Turn off slave drive hot-plug function in bCheckDriveStatus()
*       for smp+133 issue
* 05/29/02 Myst Lin     v1.20 build 21
*       rewrite SubmitTimer and add a new timer to handle event.
*********************************************************************/
 
#include <linux/blkdev.h>
#include "global.h"
#include "wrapper.h"

#include "promise_raid_engine/public_globals.h"
#include "promise_raid_engine/ata-3.h"
#include "promise_raid_engine/scsi-3.h"

#define	DID_OK		0
#define DID_ABORT	5

unchar	INestcli=0 ; 
ulong	SaveFlags;
extern struct timer_list	submit_timer;	// for submit_timer_handler
ulong	StartRebJiffies = 0 ;	// start rebuild jiffies
unchar	SCSIRequestFlag = 0 ;
struct	REQ_head {
   REQ			req;
   SG			prdt[Max_SG_Table];
   void			*cmd;
   void			(*fn)(void *) ;
   struct REQ_head	*next;
   unsigned char        pending;
} ; 
typedef	struct	REQ_head REQ_Head_t ;
REQ_Head_t	Queue[Can_Queue+1], *PoolH, *PoolF ;	// Head/Foot
int 	fasttrak_internal_cmd(REQ_Head_t *myqueue);
int 	fasttrak_rw_cmd(REQ_Head_t *myqueue);
bpEVENT CurrentEvent=0;
/*********************************************************************
 * Function:
 *   submit_timer_handler sub-routines :
 *	char PreReb(unchar array)
 *	char PreNotReb(unchar array)
 *
 * Description:
 *   	PreReb    - Array Rebuilding on last submit timer handler
 *	PreNotReb - Array Not Rebuilding on last submit timer handler
 *	   return - represent the Array is rebuilding or not
 *	        0 - Not Rebuilding
 *		1 - Rebuilding
*********************************************************************/
char PreReb(unchar array) {

   static int	percent ;
   unchar 	dev;

   // find previous rebuilding drive : dev
   for( dev=0 ; dev < FT_Max_Dev ; dev++ ) {
	ft_drive_t	*dp = &ft_drive[dev] ;
	if( (dp->arr == array) && (dp->rebuilding) ) 
	   break;
   }		

   if( p_evMaintenanceEvent[array] ) {
	unsigned char arr     = ft_drive[dev].arr ;	
	unsigned char strm    = ft_drive[dev].strm ;
	unsigned char block   = aryDef[arr].strm[strm].bIndex; 
	unsigned long wSize   = blkDisk[block].wSize / 2048 ;
	unsigned long wWater  = blkDisk[block].wWatermark / 2048;
	percent = wWater * 100 / wSize ;
	return(1);
   }

   // Rebuilding is stopped
   if( aryDef[array].blFlag & blRECOVERING_LUN ) {
	unchar	abnormality = ft_drive[dev].pchannel->abnormality ;
   	if( !abnormality && (aryDef[array].blFlag & blMARKED_LUN) ) {
	   CtrlLED(dev, LED_Green);
	   if (ft_drive[dev].rebuilding == 2) 
	      log_event(4127, array, dev, 0,0,0,0); 
	   else if (ft_drive[dev].rebuilding == 1)
	      log_event(4126, array, dev, 0,0,0,0); 
	}
	else {
	   //	*** Rebuild aborted ***
	   //	The source disk maybe has been powered off
	   //            or encountered a bad block
	   CtrlLED(dev, LED_Red);
	   if (ft_drive[dev].rebuilding == 2) 
	      log_event(4129, array, dev, 0,percent,0,0); 	
	   else if (ft_drive[dev].rebuilding == 1)
	      log_event(4128, array, dev, 0,percent,0,0); 	
	}
   }
   else { 	// complete : 0x0F , fail : 0x1F
	// 	*** Rebuild successful ***
	CtrlLED(dev, LED_Green);
	if (ft_drive[dev].rebuilding == 2) 
	   log_event(4127, array, dev, 0,0,0,0); 
	else if (ft_drive[dev].rebuilding == 1)
	   log_event(4126, array, dev, 0,0,0,0); 
   }
   ft_drive[dev].rebuilding = 0;
   StartRebJiffies = 0 ;
   return(0);	
}
// -----------------------------------------------
char PreNotReb(unchar array) {

   unchar dev, uc;
   unchar NormalStrm = blVALID_BLK | blWORKING_BLK | blASSIGNED_BLK ;
   unchar RecoveringStrm = NormalStrm | blREDIRECT_BLK ;
   
   if( p_evMaintenanceEvent[array] == NULL ) 
   	return(0);
   			
   //	*** Start Rebuilding ***
   StartRebJiffies = ft_get_jiffies() ;
   // Define new field value of ft_drive[dev] .arr & .strm
   if( DefDriveStruct() ) {	
   	ft_printk("FastTrak: Define Drive Structure Error..\n");
   	return(0);
   }
   		
   // find rebuilding drive on aryDef[array]
   for( uc=0 ; uc < FT_Max_Dev ; uc++ ) {
   	dev = 0 ;
	if( aryDef[array].strm[uc].blFlag == RecoveringStrm ) {
	   do {
	   	ft_drive_t	*dp = &ft_drive[dev] ;
	   	if((dp->present)&&(dp->arr == array)&&(dp->strm == uc)) {
		   if (dp->rebuilding == 2)
		      log_event(4104, array, dev,0,0,0,0);
		   else {
		      log_event(4103, array, dev,0,0,0,0);
		      dp->rebuilding = 1 ;
		   }
	   	   break;
	   	}
	   } while( dev++ < FT_Max_Dev ) ;
	   break;
	}
   }		
   CtrlLED(dev, LED_Amber);
   return(1);	
}		
/*********************************************************************
 * Function:
 *   Timer Handler :
 *	void submit_timer_handler(ulong data)
 *	void error_handler(ulong dev)
 *
 * Description:
 *   	submit_timer_handler : initialize in SubmitTimer(bpEVENT)
 *	error_handler        : initialize in SumitChannel(bpDTASK)
*********************************************************************/
void submit_timer_handler(ulong data) {

   unchar		array ;	
   bpEVENT		p_ev ;
   static unchar 	reb_arys[FT_Max_Arr] = { 0 } ;
   ulong		flags;

   save_flags(flags);
   cli();
   DebFtFun(ft_printk("submit_timer_handler(%ld).\n",ft_get_jiffies()););

   init_timer(&submit_timer);
   submit_timer.expires = jiffies + 200;
   submit_timer.data = 0;
   submit_timer.function = submit_timer_handler;
   add_timer(&submit_timer);

   // Put information to system log if Array status is 
   // different with last submit timer.
   for( array=0 ; array < FT_Max_Arr ; array++ ) 
	reb_arys[array] = (reb_arys[array])?PreReb(array):PreNotReb(array);
   	   
   if ( CurrentEvent != 0) {
	   hTicksTillEvent -= 36;

	   if ( (short)hTicksTillEvent <=0) {
		   p_ev  = CurrentEvent ;
		   hTicksTillEvent = 0 ; 
		   CurrentEvent = 0;
		   TimerReturn(p_ev);
	   }
   }
   restore_flags(flags);
   return ;
} 
/***********************************************************
 * Function:
 *   void InitQueueBuffer(void)
***********************************************************/
void InitQueueBuffer (void) {
   unchar	queue;
   for( queue=0 ; queue <= Can_Queue ; queue++ ) {
	if( queue == Can_Queue )
	   Queue[queue].next = NULL ; 
	else
	   Queue[queue].next = &Queue[queue+1] ; 
   }
   PoolH = &Queue[0] ;
   PoolF = &Queue[Can_Queue] ;
   return;
}
/***********************************************************
 * Function:
 *   do_queuecommand
 *
 * Description:
 *   cmd - SCSI command.
 *   fn  - function called on completion.
***********************************************************/
extern char EndRequest(void *cmd,void *fn,int result) ;
int do_queuecommand(ScsiEngReq_p CurScsiReq) 
{
   unchar		max_cdb_len = 12;
   ushort		us;
   REQ_Head_t		*myqueue;

   DebScsiFun(ft_printk("DebScsiFun : do_queuecommand(). \n");); 
   myqueue 	= PoolH ;
   PoolH	= myqueue->next;
   memset(myqueue,0,sizeof(REQ_Head_t));

   myqueue->cmd                 = CurScsiReq->cmd ;
   myqueue->fn                  = CurScsiReq->fn  ;
   myqueue->req.bAdapterStatus 	= 1 ;
   myqueue->req.bTargetStatus  	= 0 ;
   myqueue->req.bID     	= CurScsiReq->target ;
   myqueue->req.bLUN    	= 0 ;
   // a number of PRD Table entries
   myqueue->req.bSG_Count      	= CurScsiReq->use_sg ;
   myqueue->req.bSenseSize     	= 16 ;
   myqueue->req.p_buf          	= CurScsiReq->buffer ;
   myqueue->req.p_sns          	= CurScsiReq->sense ;
   for( us=0 ; us < max_cdb_len ; us++ ) 
   	myqueue->req.bCDB[us] = CurScsiReq->cmnd[us] ;
   if(CurScsiReq->use_sg) {	
	for( us=0 ; us < CurScsiReq->use_sg ; us++ ) {
	   myqueue->prdt[us].wStartAddr = CurScsiReq->WStartAddr[us] ;
	   myqueue->prdt[us].wCount     = CurScsiReq->WCount[us] ;
	}   
	myqueue->prdt[CurScsiReq->use_sg-1].wCount |= (1<<31) ;	// EOT=1
   	myqueue->req.p_sg = &myqueue->prdt[0] ;
   }
   SCSIRequestFlag ++ ;
   switch (myqueue->req.bCDB[0]) {
     	case scsiTEST_UNIT_READY :
	case scsiINQUIRY :
     	case scsiREAD_CAPACITY:
		return(fasttrak_internal_cmd(myqueue));
	case scsiREAD_6:
     	case scsiREAD_10:
     	case scsiWRITE_6:
     	case scsiWRITE_10:
  	   	return(fasttrak_rw_cmd(myqueue)); 		
     	case scsiMODE_SENSE_6:
   	   	SubmitRAID((REQ *)myqueue) ;   
   		return 0;
     	default:
	   	break;
   }
   ft_printk("FastTrak : Unknown SCSI command , %#x\n",myqueue->req.bCDB[0]);
   EndRequest(myqueue->cmd,myqueue->fn,DID_ABORT << 16);
   return 0;
}
// -----------------------------------------------
int fasttrak_internal_cmd(REQ_Head_t *myqueue)
{
   DebScsiFun(ft_printk("DebScsiFun : fasttrak_internal_cmd(). \n"););
   switch (myqueue->req.bCDB[0]) {
     	case scsiTEST_UNIT_READY :
   	   SubmitRAID((REQ *)myqueue) ;   
	   break;
	case scsiINQUIRY :
	   myqueue->req.wSize = myqueue->req.bCDB[4] ;
   	   SubmitRAID((REQ *)myqueue) ;   
	   break;
     	case scsiREAD_CAPACITY:	
	   // Fujitsu - To Avoid system hanged while submit 
	   // 		scsiREAD_CAPACITY during rebuilding.
	   {
   		unchar	ID   = (unchar  )myqueue->req.bID;
		unchar	*buf = (unchar *)myqueue->req.p_buf;
   		ulong	wSize  ;
   	   	wSize = aryDef[ID].wSize - 1 ;
	   	buf[0] = (unchar) (wSize >> 24) ;
	   	buf[1] = (unchar) (wSize >> 16) ;
	   	buf[2] = (unchar) (wSize >>  8) ;
	   	buf[3] = (unchar)  wSize ;
	   	buf[4] = 0 ;
	   	buf[5] = 0 ;
	   	buf[6] = 0x02 ;
	   	buf[7] = 0 ;
	   }
   	   SCSIRequestFlag -- ;
   	   PoolF->next = myqueue ;
   	   PoolF = myqueue ;
   	   PoolF->next = NULL ;
   	   EndRequest(myqueue->cmd,myqueue->fn,(DID_OK << 16));
	   break ;
   }
   return 0;
}
// -----------------------------------------------
int fasttrak_rw_cmd(REQ_Head_t *myqueue)
{
   unchar cmnd4 = myqueue->req.bCDB[4];
   unchar cmnd7 = myqueue->req.bCDB[7];
   unchar cmnd8 = myqueue->req.bCDB[8];

   DebScsiFun(ft_printk("DebScsiFun : fasttrak_rw_cmd(). \n"););
   switch( myqueue->req.bCDB[0] ) {
	case scsiREAD_6  :		//	0x08
	case scsiWRITE_6 :		//	0x0A
   	   myqueue->req.wSize = ( cmnd4 ? (cmnd4 * 512) : 0x20000 );
	   if(!myqueue->req.bCDB[4]) ft_printk("fasttrak_rw_cmd: cmd->cmnd[4] = 0\n ");
	   break;						// 256*512
	case scsiREAD_10  :		//	0x28
	case scsiWRITE_10 :		//	0x2A
   	   myqueue->req.wSize = ( (cmnd7 << 8) | cmnd8 ) * 512 ;
	   break;
   }
   SubmitRAID((REQ *)myqueue);
   return(0);
}
/***********************************************************
*  Functions: (Frank Tiernan)
*    int reset_controller(void) 
*
*  Description:
*    Reset the Engine. Called when Scsi_Cmnd times out at the
*  OS level by fasttrak_reset().And Set MaintenanceLength = 0
*  if reset_controller is called to slow down rebuild rate so
*  scsi command can complete.
***********************************************************/
int reset_controller(void) {
   int index;
   
   // If rebuild is in progress, the command may never complete in time
   // unless rebuild stops... so modify the rebuild rate of the engine
   bResetController = TRUE;  
   MaintenanceLength = 0;

   for (index = 0; index <= Can_Queue; index++) {  
    	// Flush out all the pending requests. It is assumed one 
    	// is the one that timed out.
    	if (Queue[index].pending) {
      	   if (!RequestTimedOut((REQ *)&Queue[index])) {
           // we might have to do something here to wait
           // for pending commands that have not timed out
      	   }
    	}
  }
  return TRUE;
}
/***********************************************************
*  Functions:
*    exported by the OS Driver  
*
*  Description:
*   
***********************************************************/
void RAID_Return(bpREQ request)
{
   REQ_Head_t 	*myqueue = (REQ_Head_t *) request ;
   void		*cmd  = myqueue -> cmd ;
   void		(*fn)(void *) = myqueue -> fn ;
   int		result ; 

   DebEngFun(ft_printk("DebEngFun : RAID_Return() .\n"););
   if ( request->bAdapterStatus != 1 ) 		// normal : bAdapterStatus = 1 
   	DEB(ft_printk("RAID_Return : Error(%d).\n",request->bAdapterStatus););
   if( request->bAdapterStatus == 1 ) 
	result = DID_OK << 16 ;
   else if( request->bAdapterStatus == 0x84 ) {		// HDDs power off
   	DEB(printk("*** bAdapterStatus(0x84) - HDD Power Off ****\n"););
   	result = ( aryDef[myqueue->req.bID].blRAID ) ? 
	   ( DID_OK << 16 ) : ( DID_ABORT << 16 ) ; // Mirror : blRAID = 1
   }
   else 
	result = DID_ABORT << 16 ;

   SCSIRequestFlag -- ;
   PoolF->next = myqueue ;
   PoolF = myqueue ;
   PoolF->next = NULL ;
   EndRequest(cmd,fn,result);
   DEB(ft_printk("End RAID_Return(bpREQ request) .\n"));
   return ;
}
// -----------------------------------------------
void SubmitTimer(bpEVENT p_ev)
{
   DebEngFun(ft_printk("DebEngFun : SubmitTimer(%d).\n",p_ev->hETA););

   if (p_ev) {
	   CurrentEvent = p_ev;
	   hTicksTillEvent = p_ev->hETA;
	   return;
   }
}
// -----------------------------------------------
ulong wGet55msecTicksSinceMidnight(void)
{
   DebEngFun(ft_printk("DebEngFun : wGet55msecTicksSinceMidnight.\n"););
   return(ft_get_jiffies());
}
// -----------------------------------------------
HALF hGetDaysSince1980(void)
{
   struct timeval	tv ;
   ulong		days ,seconds ;

   DebEngFun(ft_printk("DebEngFun : hGetDaysSince1980(void)\n"););
   ft_do_gettimeofday(&tv);				// from 1970
   seconds = (ulong)tv.tv_sec - (10*365+2)*24*60*60 ; 	// from 1980
   days = seconds / (24*60*60) ;
//   ft_printk("hGetDaysSince1980(void) : %ld\n",days);
   return(days);
}
// -----------------------------------------------
void DisableInt(void)
{
   if(!INestcli++) { 
	save_flags(SaveFlags);
	cli();
   }
   return;
}
// -----------------------------------------------
void EnableInt(void)
{
//   if( (!INestcli) || (!--INestcli) ) {
   if( (!--INestcli) ) {
	restore_flags(SaveFlags);
   }
   return;
} 
// -----------------------------------------------
int GetCHS(unchar dev,uint *pcyls,uint *pheads,uint *psectors) {

   struct ft_driveid	*id=ft_drive[dev].identify;
   uint		cyls    = id->cyls ;
   uint		heads   = id->heads ;
   uint		sectors = id->sectors ;
   ushort	CapH = id->max_48LBA[1] ;
   ushort	CapL = id->max_48LBA[0] ;
   
   DebEngFun(ft_printk("DebEngFun : wReservedSectorLBA()\n"););
   if ( ft_drive[dev].LBA_48bit ) {
	if( cyls == 16383 ) {
	   cyls = ( ((CapH << 16)+CapL) / heads ) / sectors ;
	   // Don't transfer CHS again when cyls larger then 65535, 
	   // in order to match BIOS setting on Reserved Sector LBA
//	   if ( cyls >= 65535 ) {
//	   	cyls = 65535;
//	   	heads = 255;
//	   	sectors = ( ((CapH << 16)+CapL) / cyls ) / heads;
//	   }
	}
   }	
   else {	   
	if( cyls == 16383 )
	   cyls = ( id->lba_capacity / heads ) / sectors ;
   }
   *pcyls    = cyls ;
   *pheads   = heads ;
   *psectors = sectors ;
   return(0);
}
// -----------------------------------------------
ulong wReservedSectorLBA(ByTE bCh,ByTE bID,ByTE flag) {

   unchar	dev = bCh*2 + bID ;
   uint		cyls, heads, sectors ;
   
   DebEngFun(ft_printk("DebEngFun : wReservedSectorLBA()\n"););
   GetCHS(dev,&cyls,&heads,&sectors) ;

   if(flag)
	return (( (cyls-1)*heads - 1 ) * sectors) ;
   else 		// default flag = 0
	return (( cyls * heads - 1 ) * sectors) ;
}
// -------------------------------------------------------------------
ulong wGetGeometry(ByTE bCh,ByTE bID,ByTE flag) {

   unchar	dev = bCh*2 + bID ;
   uint		cyls, heads, sectors ;

   DebEngFun(ft_printk("DebEngFun : wGetGeometry().\n"););
   GetCHS(dev,&cyls,&heads,&sectors) ;
   return( (((ushort)cyls)<<16)+(heads<<8)+sectors);
}
// -----------------------------------------------
void wGetDriveName(ByTE bCh,ByTE bID,ByTE *pName)
{
// In fact : Engine does not call the fun.
   DebEngFun(ft_printk("DebEngFun : wGetDriveName().\n"););
   return;
}
// -----------------------------------------------
ulong wLBAnext(ByTE bCh,ByTE bID)
{
// In fact : Engine does not call the fun.
   DebEngFun(ft_printk("DebEngFun : wLBAnext()\n"););
   return(0);
}
/*********************************************************************
 * Function:
 *   BoOL bCheckDriveStatus(ByTE bCh,ByTE bID)
 *
 * Description:	
 *   To make a sure the specific drive is present.
 *	
 * Return Value:
 *   present   - 1
 *   not exist - 0
 *
*********************************************************************/		
BoOL bCheckDriveStatus(ByTE bCh,ByTE bID)
{
   struct ft_driveid	*id ;
   unchar 		dev , result ;
   ushort		base2;
   ulong		flags;
   int			ATConflict(char dev);
   ft_channel_t		*p_ch = &ft_channel[bCh];
   
   dev = bCh * 2 + bID ;
   base2 = ft_drive[dev].pchannel->base+2 ;	
   DEB(ft_printk("DebEngFun : bCheckDriveStatus(DEV%d) \n",dev+1););
   // check channel exist or not
   if(bID%2) {
	//turn off slave drive hot-plug function for smp+133 issue
	return(0);
   }
   if( (bCh >= FT_Max_Ch) || (!ft_channel[bCh].padapter->present) )	
   	return(0);
   if( p_ch->pdrive[0]->lbusy || p_ch->pdrive[1]->lbusy )	
	return(0);
   if( !ATConflict(dev) )	// when Artemis Conflict , drive 2 disable
	return(0);
   if( p_ch->abnormality && CheckChannelBus(bCh) )
	return(0);
   save_flags(flags);
   cli();
   restore_flags(flags);
   id = ft_drive[dev].identify ;
   result = identify_device(dev,id) ;
   switch(result) {
	case 0	: break;
	case 3  : ft_printk("FastTrak : (Error) Irq has been disabled !!\n");
	case 1  : 
	case 2  : return(0);		// Device not Exist 	
	default : ft_printk("FastTrak : identify_device() fail !!\n");
   }
// blDiskFlag[dev/2][dev%2] = 1 ;	// Engine automatically set it .
   ft_drive[dev].present = 1 ;
   ft_drive[dev].light = LED_Amber ;	// Orange Light
   DEB(ft_printk("FastTrak : FIND NEW DRIVE !! Device %d\n",dev););
   MaintenanceLength = 4;
   log_event(4121, 0, dev, 0,0,0,0); 	// New Disk plugged in
   return(1);
}
// -----------------------------------------------
void OutputToSerialPort(apBUF p_b )
{
// In fact : Engine does not call the fun.
   DebEngFun(ft_printk("DebEngFun : OutputToSerialPort()\n"););
   return;
}   
// --------------------------------- ( for NEW ENGINE )
void Led_Turnon(ByTE bCh, ByTE bID)
{
   DEB(ft_printk("Led_Turnon() (WIN/NT) : rainmker.c\n"););
   return;
}
// -----------------------------------------------
ulong _call FastTrakNotification (WORD NotifyType,...)
{
   DEB(ft_printk("FastTrakNotification() : blPlug()\n"););
   return(0);
}

