/*+M*************************************************************************
 * Promise FastTrak device driver for Linux.
 *
 * Copyright (c) 2002  Promise Technology, Inc.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * --------------------------------------------------------------------------  
 * Copyright (c) 1999-2002 Promise Technology, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions, and the following disclaimer,
 *    without modification, immediately at the beginning of the file.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * Where this Software is combined with software released under the terms of 
 * the GNU Public License ("GPL") and the terms of the GPL would require the 
 * combined work to also be released under the terms of the GPL, the terms
 * and conditions of this License will apply in addition to those of the
 * GPL with the exception of any terms or conditions of this License that
 * conflict with, or are expressly prohibited by, the GPL.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *   		
 *-M*************************************************************************/
/***************************************************************************
 *
 * public_globals.h - contains structure definitions used in the FastTrak 
                      driver and the Promise Technology RAID engine.
 *
 ***************************************************************************/

#include "constant.h"

#pragma pack (1)        //12/16/97 S.Zhu
//#endif

#ifdef LINUX				/* 	Linus Chen 	*/
#define PACKED __attribute__((packed))
#endif

#define ACTUAL_ARRAY_SUPPORTED          0x04

#define MAX_CHANNEL_SUPPORTED           0x08
#define MAX_ID_PER_CHANNEL              0x02
#define MAX_CONCURRENT_THREAD            16
#define MAX_STREAM_PER_ARRAY            0x08 
#define SERIAL_SIZE                     0x08
#define RESOURCE_FREE_POOL_COUNT        0x05
#define bALL_CHANNELS			0xFF
#define blVALID_BLK                     bit(0)  /* blFlag */
#define blWORKING_BLK                   bit(1)
#define blASSIGNED_BLK                  bit(2)
#define blSPARE_BLK                     bit(3)
#define blDUPLICATE_BLK                 bit(4)
#define blREDIRECT_BLK                  bit(5)
#define blMARK_FOR_DOWN                 bit(6)
#define blNEED_HOT_SPARE                bit(6)
#define blMARK_FOR_UP                   bit(7)
#define blRECOVERING_LUN                bit(4)
#define blMARKED_LUN                    bit(5)


#define blMIRROR                        bit(0)  /* blRAID */
#define blRAID1                         blMIRROR


typedef struct q {                              /* sync with SBUF and BBUF */
  struct q  *p_next;
  ByTE   bReserved[3];
  ByTE   bMagic;
  ByTE   bReserved1[4];
  struct q  *p_link;
} PACKED Q, PACKED *bpQ, PACKED *spQ;

typedef struct block {
  BoOL blFlag;
  ByTE bStream;
  ByTE bCh;
  ByTE bID;
  ByTE bSerial[SERIAL_SIZE];
  WORD wStartLBA;
  WORD wSize;
  WORD wWatermark;
} PACKED BLOCK, PACKED *bpBLOCK, PACKED *dpBLOCK;

typedef struct seq {
  union {
    spQ     p_task;
    struct  req *p_rq;
    struct  seq *p_sq;
  } branch;
  BoOL blFlag;
  ByTE bType;
  ByTE bCount;
  ByTE bMagic;                                  /* signature byte to id struct */
  union {
    spQ     p_task;
    dpwBUF  p_w;
    dpBUF   p_b;
    BoOL    (* function)(void *);
  } action;
  union {
    struct  req *p_rq;                /* -> assoc. RAID cmd struct */
    WORD    w;
    ByTE    b;
    struct event *p_ev;
  } param;
} PACKED SEQ, PACKED *bpSEQ, PACKED *spSEQ;

typedef struct prd {
  WORD wStartAddr;                              /* physical addr */
  HALF hCount;                                  /* 16 bit count */
  ByTE bReserved;
  BoOL blFlag;
} PACKED PRD, PACKED *bpPRD;


typedef struct sg {
  WORD wStartAddr;                              /* physical addr */
  WORD wCount;                                  /* size of element in bytes */
} PACKED SG, PACKED *apSG, PACKED *spSG;


typedef struct strm {
  BoOL blFlag;                                  /* see BLOCK.blFlag */
  ByTE bIndex;
  ByTE bDuplicateIndex;
  ByTE bRedirectIndex;
} PACKED STRM, PACKED *dpSTRM;


typedef struct consolidate_task {
  spQ     p_next;                                 /* link pointer */
  ByTE    bReserved[3];
  ByTE    bTaskCount;
  BoOL    blOldFlag;
  ByTE    bOldCount;
  HALF    hOldSize;
  bpQ     p_link;
  bpPRD   p_prd;
  struct  dtask *p_tk[MAX_CONCURRENT_THREAD];
  PRD     prd;
} PACKED CTASK,  PACKED *bpCTASK;


typedef struct array {
  HALF hGeneration;
  BoOL blFlag;
  BoOL blRAID;
  ByTE bWidth;
  ByTE bStripeBlockPower;                       /* in power of 2 sctrs */
  ByTE bLogicalWidth;
  ByTE bID;
  WORD wSize;                                   /* # of sectors */
  HALF hCyl;                                    /* # of cyl */
  ByTE bHd;                                     /* # of hd per cyl */
  ByTE bSctr;                                   /* # of sctr per hd */
  ByTE bSerial[SERIAL_SIZE];
  WORD wWatermark;
  HALF hStripeBlockSize;                        /* in sctrs */
  HALF hStripeSize;                             /* in sctrs */
  STRM strm[MAX_STREAM_PER_ARRAY];              /* index into block database */
} PACKED ARY, PACKED *dpARY;

typedef struct req {
  spQ    p_next;                                /* -> next blocked req */
  ByTE   bType;                                 /* RAID Engine only */
  ByTE   bAdapterStatus;                        /* init to asINCOMPLETE_REQ */
  ByTE   bTargetStatus;                         /* init to ssINCOMPLETE_CMD */
  ByTE   bID;                                   /* array id */
  ByTE   bLUN;                                  /* must be 0 */
  ByTE   bTag;                                  /* RAID Engine only */
  ByTE   bSG_Count;                             /* 0 = no sg, 1 and up = sg */
  ByTE   bSenseSize;                            /* >= 14 */
  WORD   wSize;                                 /* total size of xfer */
  spBUF  p_buf;                                 /* virtual -> a buffer or */
  spSG   p_sg;                                  /* virtual -> sg entries */
  spBUF  p_sns;                                 /* virutal -> sense data buffer */
  ByTE   bCDB[12];                              /* actual scsi cmd */
  void  *p_DeviceExtension;                     /* for NT */
  struct map *p_map;                  /* RAID Engine Only */
  bpSEQ  p_sq;                                  /* RAID Engine Only */
  bpQ    p_head[RESOURCE_FREE_POOL_COUNT];      /* RAID Engine Only */
  WORD   wLBA;                                  /* RAID Engine Only */
  HALF   hCount;                                /* RAID Engine Only */
  ByTE   bTagAction;                            /* for NT */
  ByTE   bReserved;
} PACKED REQ, PACKED *bpREQ, PACKED *spREQ;

typedef struct dtask {
  spQ    p_next;                                /* link pointer */
  ByTE   bCh;                                   /* channel number of disk */
  ByTE   bID;                                   /* 0 = master, 1 = slave */
  ByTE   bStatus;                               /* Bus Master status */
  ByTE   bMagic;                                /* signature byte to id struct */
  spSEQ  p_sq;                                  /* RAID Engine only */
  bpQ    p_link;                                /* duplicated ptr for release */
  WORD   wPRD;                                  /* physical -> PRD table */
  bpPRD  p_PRD;                                 /* virtual -> PRD table */
  union {
    ByTE b1Fx[7];                               /* 1F1 to 1F7 */
    struct {
      #ifdef BIOS
      ByTE bFeature;                            /* ATA feature Register */
      ByTE bCount;                              /* # of sctr to process upto 128 */
      WORD wLBA;                                /* starting sector location */
      ByTE bCmd;                                /* ATA Command Register */
      ByTE bStatus;                             /* ATA Status Register */
      #else
      WORD wLBA;                                /* starting sector location */
      ByTE bFeature;                            /* ATA feature Register */
      ByTE bCount;                              /* # of sctr to process upto 128 */
      ByTE bCmd;                                /* ATA Command Register */
      ByTE bStatus;                             /* ATA Status Register */
      #endif
    } cmd;
  } ata_reg;
  HALF  hSize;                                  /* total xfer in half (2-byte) */
  BoOL  blCmdFlag;
  ByTE  bTimeout;                               /* in number of 128 msec */
  bpCTASK p_ct;
  ByTE  bRetry;
  BoOL  blFlag;
} PACKED DTASK, PACKED *bpDTASK;


typedef struct event {
  spQ  p_next;
  BoOL blFlag;
  ByTE bFunction;
  ByTE bReserved;
  ByTE bMagic;
  HALF hETA;
  HALF hInterval;
  bpDTASK p_tk;
  WORD wStartLBA;
  WORD wBlockEndLBA;
  union {
    struct {
      bpREQ   p_rq;
      ByTE    bAry;
      ByTE    bStrm;
      ByTE    bBlk;
    } recon;
    struct {
      bpREQ   p_rq;
      ByTE    bAry;
	  ByTE    bStrm;
	  ByTE    bBlk;
	} chkinte;
    struct {
      bpREQ   p_rq;
      ByTE    bCh;
      ByTE    bID;
    } heart;
    struct {
      bpREQ   p_rq;
      ByTE    bCh;
      ByTE    bID;
      ByTE    bBlk;
      ByTE    bOffset;
      ByTE    bCount;
      ByTE    bOriginalOffset;
      HALF    hSize;
      apBUF   p_src;
    } config;
    struct {
      bpREQ   p_rq;
      ByTE    b0;
      ByTE    b1;
      ByTE    b2;
      ByTE    b3;
    } peg;
  } param;
} PACKED EVENT, PACKED *bpEVENT;

extern void _call TimerReturn(bpEVENT p_ev);
extern void _call SubmitRAID(bpREQ p_rq);
extern BoOL _call RequestTimedOut(bpREQ p_rq);
extern BoOL StartRebuild(ByTE aryID,ByTE bCh,ByTE bID);
extern void _call ChannelReturn(bpDTASK p_tk);


extern bpEVENT p_evMaintenanceEvent[ACTUAL_ARRAY_SUPPORTED]; /* 2or4 * 4 */
extern HALF  hTicksTillEvent;                   /* 2 */
extern BoOL  bResetController;                   /* 05/13/97 */
extern HALF MaintenanceLength;
extern BLOCK blkDisk[MAX_CHANNEL_SUPPORTED*MAX_ID_PER_CHANNEL]; /* 24*16*2 */
extern ARY   aryDef[ACTUAL_ARRAY_SUPPORTED];    /* 64*4 */
extern BoOL  blDiskFlag[MAX_CHANNEL_SUPPORTED][MAX_ID_PER_CHANNEL]; /* 16*2 */
extern ByTE DisableHotSpare;
extern ByTE bLEDColor[4][2];
