/*+M*************************************************************************
 * Promise FastTrak device driver for Linux.
 *
 * Copyright (c) 2002  Promise Technology, Inc.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * --------------------------------------------------------------------------  
 * Copyright (c) 1999-2002 Promise Technology, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions, and the following disclaimer,
 *    without modification, immediately at the beginning of the file.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * Where this Software is combined with software released under the terms of 
 * the GNU Public License ("GPL") and the terms of the GPL would require the 
 * combined work to also be released under the terms of the GPL, the terms
 * and conditions of this License will apply in addition to those of the
 * GPL with the exception of any terms or conditions of this License that
 * conflict with, or are expressly prohibited by, the GPL.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *   		
 *-M*************************************************************************/
 /*M*************************************************************************
 *
 * scsi-3.h - contains scsi-3 interface definition and mappings
 *
 ***************************************************************************/

/*
 * scsi3 opcode and CDB bits
 */

#define svucSCSI_COMMAND                0x60

#define scsiINQUIRY			0x12
  #define blINQUIRY_EVPD		bit(0)
  #define bDISK_UNIT_INFO_PAGE          0xC0
  #define DISK_UNIT_INFO_PAGE_SIZE      0xF0
  #define STANDARD_INQUIRY_PAGE_SIZE    0x30
  #define INQUIRY_PRODUCT_NAME_OFFSET   0x10
#define scsiMODE_SELECT_6               0x15
#define scsiMODE_SELECT_10              0x55
#define scsiMODE_SENSE_6                0x1A
#define scsiMODE_SENSE_10               0x5A
  #define bCURRENT_PARAMETER            0x00
  #define bCHANGEABLE_PARAMETER         0x01
  #define bDEFAULT_PARAMETER            0x02
  #define bSAVED_PARAMETER              0x03
  #define bARRAY_MAPPING_MODE_PAGE      0x20
  #define bCONTROLLER_SPECIFIC_PAGE     0x21
  #define bMAINTENANCE_CONTROL_PAGE     0x22
  #define bALL_PAGES                    0x3F
#define scsiREAD_6                      0x08
#define scsiREAD_10			0x28
#define scsiREAD_CAPACITY		0x25
#define scsiREQUEST_SENSE               0x03
  #define REQUEST_SENSE_DATA_SIZE       0x20
#define scsiREZERO_UNIT			0x01
#define scsiSEEK_6                      0x0B
#define scsiSEEK_10			0x2B
#define scsiTEST_UNIT_READY		0x00
#define scsiVERIFY			0x2F
#define scsiWRITE_6                     0x0A
#define scsiWRITE_10			0x2A
#define scsiWRITE_VERIFY		0x2E


/*
 * SCSI 3 vendor unique opcodes
 */

#define svucARRAY_COMMAND               0xE0
  #define acmdMASK                      0x1F
  #define acmdREPORT_DEVICE_MAP         0x00
  #define acmdSAVE_CONFIG               0x01
  #define acmdREAD_PHYSICAL             0x02
  #define acmdWRITE_PHYSICAL            0x03
  #define acmdSCAN_CONFIG               0x04
  #define acmdFORCE_DISK_DOWN           0x05
  #define acmdFORCE_DISK_UP             0x06
  #define acmdFORCE_DISK_SPARE          0x07
  #define acmdREBUILD_DISK              0x08
  #define acmdCHECK_INTEGRITY           0x09
  #define acmdSYNC_REDUNDANCY           0x0A
  #define acmdMANUAL_THROTTLE           0x0B
  #define acmdHEART_BEAT                0x0C
  #define acmdCOMMAND                   0x0D    /* total number of 0xE0 cmds */
  #define acmdSMART_COMMAND				0x0E

#define svucCONTROLLER_COMMAND          0xE1
  #define ccmdCOMMAND                   0x02
  #define ccmdPING			0x00
  #define ccmdDEACTIVATE		0x01


/*
 * scsi status or bTargetStatus in REQ Structure
 */

#define ssGOOD				0x00    /* no error */
#define ssCHECK_CONDITION		0x02    /* check sense data for status */
#define ssCONDITION_MET			0x04    /* should not happen */
#define ssBUSY				0x08    /* RAID Engine is busy, retry after some time */
#define ssINTERMEDIATE_GOOD		0x10    /* should not happen */
#define ssINTERMEDIATE_MET		0x14    /* should not happen */
#define ssRESERVATION_CONFLICT		0x18    /* should not happen */
#define ssCOMMAND_TERMINATED		0x22    /* should not happen */
#define ssQUEUE_FULL			0x28    /* RAID Engine is busy, retry after some time */


/*
 * scsi sense key, ASC, and ASCQ
 */

#define skNO_SENSE                      0x00    /* no additional status */
#define skRECOVERED_ERROR               0x01    /* cmd completed using recovery */
#define skNOT_READY                     0x02    /* array is busy, retry after some time */
#define skMEDIUM_ERROR                  0x03    /* auto-reassign/recovery failed */
#define skHARDWARE_ERROR                0x04    /* disk hardware error */
#define skILLEGAL_REQUEST               0x05    /* invalid bits in CDB */
#define skUNIT_ATTENTION                0x06    /* special event just happened */
#define skDATA_PROTECT                  0x07    /* should not happen */
#define skBLANK_CHECK                   0x08    /* should not happen */
#define skCOPY_ABORTED                  0x0A    /* should not happen */
#define skABORTED_COMMAND               0x0B    /* should not happen */
#define skEQUAL                         0x0C    /* should not happen */
#define skVOLUME_OVERFLOW               0x0D    /* should not happen */
#define skMISCOMPARE                    0x0E    /* should not happen */

#define saNO_ASC                        0x00
  #define sqNO_ASCQ                       0x00
#define saLUN_NOT_READY                 0x04
  #define sqINTERVENTION_REQUIRED         0x03
  #define sqMAINTENANCE_IN_PROGRESS       0x80
#define saECC_ERROR                     0x10
#define saID_ADDR_MARK_NOT_FOUND        0x12
#define saINVALID_OPCODE                0x20
#define saLBA_OUT_OF_RANGE              0x21
#define saINVALID_FEILD_IN_CDB          0x24
#define saLOGICAL_UNIT_NOT_SUPPORTED    0x25
#define saINVALID_FIELD_IN_PARAMETER    0x26
#define saINTERNAL_TARGET_FAILURE       0x44
