/*+M*************************************************************************
 * Promise FastTrak device driver for Linux.
 *
 * Copyright (c) 2002  Promise Technology, Inc.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * --------------------------------------------------------------------------  
 * Copyright (c) 1999-2002 Promise Technology, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions, and the following disclaimer,
 *    without modification, immediately at the beginning of the file.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * Where this Software is combined with software released under the terms of 
 * the GNU Public License ("GPL") and the terms of the GPL would require the 
 * combined work to also be released under the terms of the GPL, the terms
 * and conditions of this License will apply in addition to those of the
 * GPL with the exception of any terms or conditions of this License that
 * conflict with, or are expressly prohibited by, the GPL.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *   		
 *-M*************************************************************************//*********************************************************************
* Filename : swapbox.c
* 10/30/00 Linus Chen  
*       Moved all monitoring SWAP-BOX and wrapper utility functions 
*	to this file .
* 11/08/00 Linus Chen
*	(1) Fix SwapBox power control 
*	(2) Not show SwapBox monitor information if box->MIC does
*		not equal to 0x5CA3 (Windbond IC).	
* 11/11/00 Jenny Liang
*	Fix SWAP_STATE to make sure in state 1.
* 11/24/00 Linus Chen
*	Adding output information about Adapter2 on RepArrInfo()
* 12/21/01 Hank Yang
*	Fix Adapter Model String for cat0
* 05/27/02 Myst Lin
*       Fix wrong channel description on second controller
*       add CtrlLED function
*       rewrite proc function
*********************************************************************/

#include "global.h"
#include "wrapper.h"
 
#include "promise_raid_engine/public_globals.h"
#include "promise_raid_engine/ata-3.h"

/*********************************************************************
 * Function:
 *   char DefDriveStruct(void)
 *
 * Description:
 *   Fill out data structure after RAID initialize
 *	1. define ft_drive[dev].arr
 *	2. define ft_drive[dev].strm
*********************************************************************/
char	DefDriveStruct (void) {		
   unsigned char	arr, strm, dev, blFlag, bIndex;  
   for( dev=0 ; dev < FT_Max_Dev ; dev++ ) {
   	ft_drive[dev].arr  = 0xFF ;
   	ft_drive[dev].strm = 0xFF ;
   }
   for( arr=0 ; arr < FT_Max_Arr ; arr++ ) {
	if(aryDef[arr].blFlag) {	// Array[arr] had been defined ??
   	   for( strm=0 ; strm < FT_Max_Dev ; strm++ ) {
		blFlag = aryDef[arr].strm[strm].blFlag ;
		if( (blFlag == 7) || (blFlag == 0x20) || (blFlag == 0x27) ) {
   	   	   bIndex = aryDef[arr].strm[strm].bIndex ;
		   dev = blkDisk[bIndex].bCh * 2 + blkDisk[bIndex].bID ;
   	   	   ft_drive[dev].arr  = arr ;
   	   	   ft_drive[dev].strm = strm ;
   	   	}
   	   }
   	}
   }
   return(0);
}
/*********************************************************************
 *
 * Swap Box Functions:
 *   void SWAP_STATE(unchar fun,unchar dev) 
 *   void SwapSubFun(unchar fun,ushort port)
 *   unchar RW_Byte(unchar fun,unchar value,ushort port)
 *   unchar RWWinReg(unchar fun,unchar dev,unchar index,unchar data)
 *   void GetSwapInfo (unchar *ptr)
 *
 * Clock:   
 *	200 ISA bus clock (0.125us) - udelay(25)
 *  	300 ISA bus clock (0.125us) - udelay(38)
 *  	500 ISA bus clock (0.125us) - udelay(64)
 *
 ********************************************************************/
unsigned char   WR  =	0x08 ;
unsigned char   SCL =	0x02 ;
unsigned char   SDA =	0x01 ;
// -------------------------------------------------------------------
void SWAP_STATE(unsigned char fun,unsigned char dev) {
   unsigned short	iobase = ft_drive[dev].pchannel->base;
   unsigned short	select = ft_drive[dev].select;

   switch (fun) {
	case 0 :// enter the Swap Box status
		ft_inb( iobase+6 );		ft_udelay(1); 
		ft_inb( iobase+6 );		ft_udelay(1); // for 277
		ft_inb( iobase+7 );		ft_udelay(1); 
		//make sure we are in STATE1
   		ft_outb( select , iobase+6 );	ft_udelay(1);
   		ft_outb( select , iobase+6 );	ft_udelay(1);
   		ft_outb( select , iobase+6 );	ft_udelay(1);
   		ft_inb( iobase+2 ); 		ft_udelay(1);
   		ft_inb( iobase+6 ); 		ft_udelay(1);
		break;
	case 1 :// leave the Swap Box status
   		ft_outb( select , iobase+6 );	ft_udelay(1);
		break;
	default : 
		ft_printk("FastTrak: SWAP function error , SWAP_STATE().\n");
   }
   return;
}
// -------------------------------------------------------------------
void SwapSubFun(unsigned char fun,unsigned short port) {

   switch (fun) {
	case 1:	// start sequence
		ft_outb( WR+SCL+SDA , port );		
   		ft_outb( WR+SCL , port );	ft_udelay(25);
   		ft_outb( WR , port );		
		break;
	case 2:	// stop  sequence
		ft_outb( WR , port );		ft_udelay(64);
   		ft_outb( WR+SCL , port );	ft_udelay(25);
   		ft_outb( WR+SCL+SDA , port );	ft_udelay(64);
		break;
	case 3:	// Set_SDA 0
		ft_outb( WR ,port );		ft_udelay(64);
		ft_outb( WR+SCL , port );	ft_udelay(64);
		break;
	case 4: // Set_SDA 1
		ft_outb( WR+SDA , port );	ft_udelay(64);
		ft_outb( WR+SCL+SDA , port );	ft_udelay(64);
		break;
	default : 
		ft_printk("FastTrak: SWAP subfunction error .\n");
   }
   return ;
}
// -------------------------------------------------------------------
unsigned char RW_Byte(unsigned char fun,unsigned char value,unsigned short port) {
   char		i;
   unsigned char	data = 0xFF ;

   switch (fun) {
	case 0:	// read byte
	   	for(i=0;i<8;i++) {
		   	data = data << 1;
			ft_outb( 0,port); 			ft_udelay(64);
			ft_outb( SCL ,port);			ft_udelay(25);
			data = data | (ft_inb(port) & 0x01) ;	ft_udelay(38);
	   	}
	   	break;
	case 1: // write byte
	   	for(i=0;i<8;i++) 
			(value & (0x80>>i)) ? SwapSubFun(4,port) :
                                              SwapSubFun(3,port) ;
	   	break;
	default :
		ft_printk("FastTrak: SWAP function error , RW_Byte().\n");
   }
   return(data);
}	
// -------------------------------------------------------------------
unsigned char RWWinReg(unsigned char fun,unsigned char dev,unsigned char index,unsigned char data) {
   // fun 0 : read  Winbond register
   // fun 1 : write Winbond register
   unsigned short 	port   = ft_drive[dev].pchannel->base+3 ;
   unsigned char	i ;
   unsigned char	frame1[10] = {3,4,3,4,4,3,4,3,3,0} ;
   unsigned char	frame3[10] = {3,4,3,4,4,3,4,4,3,0} ;

   SWAP_STATE(0,dev);			// enter state

   SwapSubFun(1,port);			// start sequence

   i = 0 ;
   while( frame1[i] )			// frame1
   	SwapSubFun(frame1[i++],port) ;

   RW_Byte(1,index,port);		// frame2: write byte
   SwapSubFun(3,port);			

   switch (fun) {
	case 0:	// read winbond register
   		SwapSubFun(2,port);	// stop  sequence
   		SwapSubFun(1,port);	// start sequence
   		i = 0 ;
   		while( frame3[i] )	// read  winbond register frame3
   			SwapSubFun(frame3[i++],port) ;
		data=RW_Byte(0,0,port);	// read  winbond register frame4
		SwapSubFun(3,port);
		break;
	case 1:	// write winbond register
		RW_Byte(1,data,port);	// write winbond register frame3
		SwapSubFun(3,port);
		break;
	default:
		ft_printk("FastTrak: SWAP function error , RWWinReg().\n");
   }

   SwapSubFun(2,port);			// stop  sequence
   SWAP_STATE(1,dev);			// leave sequence

   return (data) ;
}
/*********************************************************************
 * function :
 *	void GetSwapInfo (unsigned char dev) 
 * description :
 *	Get Swap Box Information for 
 *		. fasttrak_proc_info ()
 *		. fasttrak_ioctl ()
*********************************************************************/
void GetSwapInfo (unsigned char dev) {
   ft_swapbox_t 	*box = &ft_box[dev];
   unsigned short	base = ft_drive[dev].pchannel->base ;
   unsigned short	WinID ;
   char			fandiv = 0 ;
   unsigned char	Divisor = 1 ,Count;
   unsigned char	T1,T2,NegTmpFlag=0,TmpF;

   //	get Swap Box ID
   SWAP_STATE(0,dev);
   box->BoxID = ft_inb( base+6 );
   SWAP_STATE(1,dev);
   
   // box->boxid bits(MSB->LSB) mean : (RSVD)xx(ID2)xx(ID1)(ID0)
   box->present = 0 ;
   switch( box->BoxID & 0x93 ) {	
	case 0x01 : 
	case 0x11 : 
	case 0x10 : 
	   box->present = 1 ;
	   break;
   }

   //	get Winbond ID ( 0x5CA3 )
   RWWinReg(1,dev,0x4E,0) ;
   WinID = RWWinReg(0,dev,0x4F,0) ;	// get vendor ID low byte
   RWWinReg(1,dev,0x4E,0x80) ;
   WinID |= RWWinReg(0,dev,0x4F,0) << 8 ;	// get vendor ID
   box->MIC = WinID ;

   // 	get Fan RPM ( default 4400 )
   fandiv = ((RWWinReg(0,dev,0x5D,0) & 0x20) >> 3) +
	((RWWinReg(0,dev,0x47,0) & 0x30) >> 4);
   while( fandiv-- ) 
	Divisor = Divisor * 2 ;			// get FAN1 Divisor
   if( (Count=RWWinReg(0,dev,0x28,0)) == 0xFF )	// get FAN1 Count
   	box->RPM = 0 ;
   else
   	box->RPM = 1.35 * 1000000 / Count / Divisor ;// get FAN1 RPM

   // 	get Temperature ( TmpC , TmpF )
   RWWinReg(1,dev,0x4E,0x01) ;	// select Bank 1
   if( (T1 = RWWinReg(0,dev,0x50,0)) & 0x80 ) {
	T1 = ~T1 +1 ;
	NegTmpFlag = 1 ;	// negative Temperature flag
   }				// TmpC = T1 + 0.5 * (T2 & 0x80)
   T2 = RWWinReg(0,dev,0x51,0) ;
   box->TmpC = ( ( NegTmpFlag ) ? (-1)*T1 : T1 );
   TmpF = (T2 & 0x80)  ? (9 * T1 / 5 + 1) : (9 * T1 / 5) ;
   if( NegTmpFlag ) {
	TmpF = ( 32 > TmpF ) ? (32 - TmpF) : (TmpF - 32) ;
	NegTmpFlag = ( 32 > TmpF ) ? 0 : 1 ;
   }
   else
	TmpF =  32 + TmpF ;	// F = (9 * C) / 5 + 32
   box->TmpF = ( ( NegTmpFlag ) ? (-1)*TmpF : TmpF );

   // 	get Vcc(mV) V12(mV)
   box->Vcc = RWWinReg(0,dev,0x23,0) * 16 * (34+50) / 50 ;
   box->V12 = RWWinReg(0,dev,0x24,0) * 16 * (10+28) / 10 ;

   return ;
}
/*********************************************************************
 *	Control LED Light , if RAID 1 Critical : 
 *		Drive Fail		- Red LED 
 *		Initialize (Normal)	- Green LED
 *		Drive Critical		- Original LED 
********************************************************************/
void	InitLED(void) {
   unsigned char dev;
   for( dev=0 ; dev < FT_Max_Dev ; dev++ ) { 
   	ft_drive_t	*drive = &ft_drive[dev] ;
	unsigned char	arr    = drive->arr  ;
	unsigned char	strm   = drive->strm ;
	unsigned char	block ;

	CtrlLED(dev, LED_Green);	// initialize - Green LED

	if( !drive->present ) {
	   CtrlLED(dev, LED_Red);	// Drive fail - Red LED
	   continue;		
	}
	block  = aryDef[arr].strm[strm].bIndex ;	// Fix segment fault	v1.02.0.17
	if ( blkDisk[block].blFlag != 0x07 ) {	// drive status
	   CtrlLED(dev, LED_Amber);		// Drive critical - Orange LED
	}
   }
}
/*********************************************************************
 *	Control LED Light 
 *	argument:
 *		dev:	which box
 *		color:	LED_Off   = No light
 *			LED_Red   = Red
 *			LED_Green = Green
 *			LED_Amber = Amber
********************************************************************/
void CtrlLED(unchar dev, unchar color) {
       	ushort	base2	= ft_drive[dev].pchannel->base + 2;
	unchar	data;

	ft_drive[dev].light = color;

       	SWAP_STATE(0,dev);	// enter box

	data = ft_inb(base2);
	data &= ~LED_MASK;
	data |= color & LED_MASK;
	ft_outb(data, base2);

       	SWAP_STATE(0,dev);	// leave box 

	return;
}
/*********************************************************************
 * Function:
 *   	char GetAllSwapInfo (void)
 *
 * Description:
 *	Get SwapBox Information if system is not busy.
*********************************************************************/
char GetAllSwapInfo (void) {
   unchar  	valch[FT_Max_Ch+1] = {0,0,0,0,0} ;
   unchar	portvalue[6] ,ch ,dev ,i;
   unchar	portnum[3] = { 2,3,6 } ;	// 1F2,1F3,1F6
   ulong	flags;

   if( SCSIRequestFlag || StartRebJiffies ) {
	return(1);		// System/Drive is busy.
   }
   save_flags(flags);
   cli();
   for( ch=0 ; ch < FT_Max_Ch ; ch++ ) {	// Save 1F2,1F3,1F6 Data
	valch[ch] = ft_drive[ch*2].present | ft_drive[ch*2+1].present ;
	if( valch[ch] ) {
	   ushort	base = ft_channel[ch].base;
	   for( i=0 ; i<3 ; i++ ) 
		portvalue[ch*3+i]   = ft_inb(base+portnum[i]);
	}
   }
   for( dev=0 ; dev < FT_Max_Dev ; dev++ ) 
	GetSwapInfo(dev);
   for( ch=0 ; ch < FT_Max_Ch ; ch++ ) {	// Restore 1F2,1F3,1F6 Data
	if( valch[ch] ) {
	   ushort	base = ft_channel[ch].base;
	   for( i=0 ; i<3 ; i++ ) 
		ft_outb( portvalue[ch*3+i] , base+portnum[i]) ;
	}
   }
   restore_flags(flags);
   return(0);
}
static void ft_copy_mem_info(ft_proc_t *info, char *data, int len)
{
	if (info->position + len > info->length)
		len = info->length - info->position;

	if (info->position + len < info->offset) {
		info->position += len;
		return;
	}
	if (info->position < info->offset) {
		data += (info->offset - info->position);
		len  -= (info->offset - info->position);
	}
	if (len > 0) {
		memcpy(info->buffer + info->position, data, len);
		info->position += len;
	}
}
static int ft_copy_info(ft_proc_t *info, char *fmt, ...)
{
	va_list args;
	char buf[81];
	int len;

	va_start(args, fmt);
	len = vsprintf(buf, fmt, args);
	va_end(args);
	ft_copy_mem_info(info, buf, len);
	return len;
}
/*********************************************************************
 * Function:
 *   int fasttrak_proc_info(char *, char **, off_t, int, int, int)
 *
 * Description:
 *   in user space , user can view "/proc/scsi/fasttrsk/x" 
 *   to run fasttrak_proc_info() in kernel space.
 * PS: 
 *   the rebuilding information reported under the condition of
 *   one MIRROR array , not support two MIRROR array concurrently.
*********************************************************************/
int GetProcInfo( ft_proc_t *buff )
{
   unchar 	dev, mode ;
   unchar	arr, uc ;
   unchar	bWidth ;	
   char		name[20] ;
   int		count;
   uint		irq;
   ft_drive_t	*drive = NULL;
   extern	int GetATResult(ft_proc_t **buf);
   extern	int ProductName(char *name);
   ft_proc_t	*info = buff;

   // if AT has been defined , 
   // it will return some information for ConnectStor II
   GetATResult(&info);	

//  /*
//   *	below is general information about 
//   *		- Adapter
//   *		- Array
//   *		- Drive
//   *		- SwapBox
//   *		- Rebuilding
//   */

   ProductName(name) ;
   ft_copy_info(info,"PROMISE ");
   ft_copy_info(info,"%s ",name);
   ft_copy_info(info,"Series Linux Driver ");
   ft_copy_info(info,"%s ",VERSION);
   ft_copy_info(info,"%s \n",VERDATE);

   for( uc=0 ; uc < FT_Max_Ada ; uc++ ) {
	ushort	id;
	if( ft_adapter[uc].present != 1 )
	   break;
	id = ft_adapter[uc].deviceid ;
   	if( id ) {
   	   ft_copy_info(info,"Adapter%d - ",uc+1);
   	   switch( id ) {
   	   	case DID262 : 	
	 	   ft_copy_info(info,"FastTrak66\n");
	 	   break;
   	   	case DID265 : 	
		   ft_copy_info(info, "FastTrak100 Lite\n");
		   break;
   	   	case DID267 : 	
	 	   ft_copy_info(info,"FastTrak100\n");
	 	   break;
   	   	case DID270 : 	
	 	   ft_copy_info(info,"FastTrak LP/TX2/TX4\n");
	 	   break;
   	   	case DID271 : 	
	 	   ft_copy_info(info,"FastTrak TX2000\n");
	 	   break;
   	   	case DID276 : 	
	 	   ft_copy_info(info,"MBFastTrak133 Lite\n");
	 	   break;
   	   	case DID277 : 	
	 	   ft_copy_info(info,"SBFastTrak133 Lite\n");
	 	   break;
	 	default :
		   ft_copy_info(info,"FastTrak Series\n");
	   }   	   	
	}
   }

   //
   //   Monitor Array Status
   //
   ft_copy_info(info,"Array    ");
   arr = count = 0 ;
//   Rebuilding  = 0 ;
   do {
      unchar bIndex, bCH, bID;
      bIndex = aryDef[arr].strm[1].bIndex;
      bCH = blkDisk[bIndex].bCh;
      bID = blkDisk[bIndex].bID;
      drive = &ft_drive[bCH*2+bID];

   	if( aryDef[arr].blFlag ) {
	   if( count )	ft_copy_info(info,"         ");
   	   ft_copy_info(info,"- Array[%d] : ",arr+1);
	   bWidth = aryDef[arr].bWidth ;
	   switch ( aryDef[arr].blRAID ) {
		case 0:	ft_copy_info(info,"%d",bWidth);
			ft_copy_info(info,"+0 Stripe ");
			break;
		case 1: if( aryDef[arr].bWidth == 2 ) {
			   ft_copy_info(info,"%d",bWidth/2);
			   ft_copy_info(info,"X2 Mirror ");
			   break;
			}
			// aryDef[arr].bWidth == 4
			ft_copy_info(info,"%d",bWidth/2);
			ft_copy_info(info,"X2 Mirror/Stripe ");
			break;
		case 8:	ft_copy_info(info,"%d",bWidth);
			ft_copy_info(info,"+0 Span ");
			break;
		default:ft_copy_info(info,"Unknown");
	   }
	   if( aryDef[arr].blFlag == 0x0F )
		ft_copy_info(info,"(OK)\n");
	   else if( aryDef[arr].blFlag == 0x8F )
		ft_copy_info(info,"(OK-Gigabyte Boundary)\n");
	   else {
		if( p_evMaintenanceEvent[arr] ) {
		   if (drive->rebuilding == 2) 
		       ft_copy_info(info,"(Synchronzing)\n");
		   else 
		       ft_copy_info(info,"(Rebuilding)\n");
//		   Rebuilding = 1 ;
		}
		else {
		   ft_copy_info(info,"(Critical)");
		   if( DisableHotSpare )
			ft_copy_info(info,"..");
		   ft_copy_info(info,"\n");
		}
	   }
	   count++;	
	} 
   } while( ++arr < FT_Max_Arr ) ;

   //
   //   Monitor Drive Status
   //
   ft_copy_info(info,"Drive    - \n");
   for( dev=0 ; dev < FT_Max_Dev ; dev++ ) {
   	drive = &ft_drive[dev];
   	if( drive->present ) {
	   unchar	i,j,cmodel;
	   ulong	Msize;
	   struct ft_driveid	*id = drive->identify ;
	   if( (drive->rebuilding) && (drive->rebuilding != 2) )
	   	ft_copy_info(info,"* %d : ",dev+1);
	   else
	   	ft_copy_info(info,"  %d : ",dev+1) ; 
	   for(i=1;i<10;i=i+3) {
	   	cmodel = drive->identify->model[i] ;
		if( (cmodel <= 'Z' && cmodel >= 'A') ||
                    (cmodel <= 'z' && cmodel >= 'a') )
	   	   ft_copy_info(info,"%c",cmodel);
		else {
		   for(j=i;j<10;j++)
			ft_copy_info(info," ");
		   break;
		}
		i--;
	   	cmodel = drive->identify->model[i] ;
		if( (cmodel <= 'Z' && cmodel >= 'A') ||
                    (cmodel <= 'z' && cmodel >= 'a') )
	   	   ft_copy_info(info,"%c",cmodel);
		else {
		   for(j=i;j<8;j++)
			ft_copy_info(info," ");
		   break;
		}
	   }
	   irq = drive->pchannel->padapter->irq;
	   ft_copy_info(info,"%s/",((dev/2)%2)?"Secondary":"Primary  ");
	   ft_copy_info(info,"%s ",(dev%2)?"Slave ":"Master");
	   Msize = ( !drive->LBA_48bit ) ? (id->lba_capacity/1000) : 
	   	(((id->max_48LBA[1]) << 16)+(id->max_48LBA[0]))/1000 ;
	   Msize = Msize * 512 /1000 ;
	   ft_copy_info(info,"%6dMB ",Msize);
	   ft_copy_info(info,"BASE(%#x) ",drive->pchannel->base);
	   ft_copy_info(info,"BM(%#x) ",drive->pchannel->bm);
	   ft_copy_info(info,"IRQ(%d) ",irq);
	   mode = drive->xfermode & 0x07;
	   switch( drive->xfermode >> 3 ) {
		case 1:  ft_copy_info(info,"PIO%d",mode);  break;
		case 4:  ft_copy_info(info,"DMA%d",mode);  break;
		case 8:  ft_copy_info(info,"UDMA%d",mode); break;
	   }
	   ft_copy_info(info,"\n");
	}
   }

   //
   //   Monitor SwapBox Status
   //		: Get SwapBox Information if system is not busy.
   //   
   if( !GetAllSwapInfo() ) { 
	unchar	uc = 0 ;
	for( dev=0 ; dev < FT_Max_Dev ; dev++ ) {
	   ft_swapbox_t	*box = &ft_box[dev];
	   if( (!box->present) || (box->MIC!=0x5CA3) )	// Winbond IC 
		continue;
	   if( !uc ) {
   		ft_copy_info(info,"SwapBox  - \n");
		uc++;
	   }
	   ft_copy_info(info,"  %d : ",dev+1);
	   switch (box->BoxID & 0x93) {
	   case 0x01 : 	
		ft_copy_info(info,"FastSwap66Box -");	break;
	   case 0x11 : 	
		ft_copy_info(info,"SuperSwapBox  -");	break;
	   case 0x10 : 	
		ft_copy_info(info,"SuperBox      -");	break;
	   default: 	
		ft_copy_info(info,"none\n",+1);		continue;
	   }
	   ft_copy_info(info," RPM(%4ld)",box->RPM);
	   ft_copy_info(info," TmpC(%d",box->TmpC);
	   ft_copy_info(info,"%cC)",0xB0);
	   ft_copy_info(info," TmpF(%d",box->TmpF);
	   ft_copy_info(info,"%cF)",0xB0);
	   ft_copy_info(info," Vcc(%d.",box->Vcc/1000);
	   ft_copy_info(info,"%dV)",box->Vcc/100%10);
	   ft_copy_info(info," V12(%d.",box->V12/1000);
	   ft_copy_info(info,"%dV)",box->V12/100%10);
	   ft_copy_info(info,"\n");
	}
   }

   //
   //   Monitor Rebuilding Array/Drive 
   //
   for( arr=0 ; arr < FT_Max_Arr ; arr++ ) {
   	if( p_evMaintenanceEvent[arr] ) {
	   ulong	wSize  = aryDef[arr].wSize / (2*1024);
	   ulong	wWater = aryDef[arr].wWatermark / (2*1024);
	   ulong	spenttime ;
	   unchar	bIndex,bCH,bID;

	   bIndex = aryDef[arr].strm[1].bIndex;
	   bCH = blkDisk[bIndex].bCh;
	   bID = blkDisk[bIndex].bID;
	   drive = &ft_drive[bCH*2+bID];

	   if (drive->rebuilding == 2)
	      ft_copy_info(info,"Synchronizing Array%d - percentage : ",arr+1);
	   else
	      ft_copy_info(info,"Rebuilding Array%d - percentage : ",arr+1);
	   ft_copy_info(info,"%3d ",(wWater*100/wSize));
	   ft_copy_info(info,"%c \n",0x25);
	   ft_copy_info(info,"                  - spent time : ");
	   spenttime = (StartRebJiffies) ? ((ft_get_jiffies() - StartRebJiffies) / 100) : 0 ;
	   ft_copy_info(info,"%3ld (minutes)\n",spenttime / 60);
	   if (drive->rebuilding != 2)
	      ft_copy_info(info,"                  - rebuilding drive : *\n");
	}
   }
//   ft_copy_info(info,"\n"); 

   return(0);
}

/*********************************************************************
 * Function:
 *   char RepArrInfo (void *buffer)
 *
 * Description:
 *   Report current Array Information for Ioctl function 0x3801
*********************************************************************/		
int RepArrInfo (void *buffer) {
   ft_info_t 	*buf;
   unchar	arr, dev, rebdev, *buf0, uc ;

   buf0 = (unchar *) buffer ;
   buf  = (ft_info_t *) buffer ;

   ft_sprintf(buf0,"%s",VERSION);
   uc = 0 ;
   do {
   	ft_adapter_t	*adapter = &ft_adapter[uc] ;
   	ft_ladainfo_t	*ap = &buf->adainfo[uc] ;
   	ap->deviceid = adapter->deviceid;
   	ap->irq	     = adapter->irq;
   	ap->base[0]  = adapter->pchannel[0]->base;
   	ap->base[1]  = adapter->pchannel[0]->ctrl;
   	ap->base[2]  = adapter->pchannel[1]->base;
   	ap->base[3]  = adapter->pchannel[1]->ctrl;
   	ap->base[4]  = adapter->pchannel[0]->bm;
   	ap->rom      = adapter->rom_base;
   } while ( ++uc < FT_Max_Ada ) ;
   for( arr=0 ; arr < FT_Max_Arr ; arr++ ) {
   	ARY		*p  = &aryDef[arr] ;
   	ft_larrinfo_t	*ap = &buf->arrinfo[arr] ;
	unchar	strm ;
	ap->ID   = arr+1 ;
	ap->Flag = p->blFlag ;
	ap->RAID = p->blRAID ;
	ap->StripeBlockPower = p->bStripeBlockPower-1 ;
	ap->Cyl  = p->hCyl+1 ;
	ap->Hd   = p->bHd+1 ;
	ap->Sctr = p->bSctr ;
	ap->Size = p->wSize/2048 ;
	ap->Watermark = 
		((p->wWatermark == 0xFFFFFFFF) ? 0 : p->wWatermark/2048);
	strm = 0 ;
   	for( dev=0 ; dev < FT_Max_Dev ; dev++ ) {
   	   ft_drive_t	*drive = &ft_drive[dev] ;
	   if( drive->present && (drive->arr == arr) ) {
	   	strm |= (1 << dev) ;	
	   }
   	}
	ap->strm = strm ;
   }
   rebdev = 0;
   for( dev=0 ; dev < FT_Max_Dev ; dev++ ) {
	if( ft_drive[dev].rebuilding ) 
	   rebdev |= (1 << dev) ;
   }
   buf->rebdev = rebdev ;
   return(0);
}
/*********************************************************************
 * Function:
 *   int Check1F7(unchar dev,ulong msec,unchar good,unchar bad)
 *
 * Description:	
 *   good - the bits of 1F7 status must be 1
 *   bad  - the bits of 1F7 status must be 0
 *
 * Return Value:
 *   0 - OK
 *   1 - Fail
*********************************************************************/		
int Check1F7(unchar dev,ulong msec,unchar good,unchar bad) {
   ft_drive_t	*drive = &ft_drive[dev];
   ulong	ul;
   ushort	base = drive->pchannel->base ;

   for( ul=0 ; ul < msec ; ul++ ) {
	if( (ft_inb(base+7) & (good | bad)) == good )
	   return(0);
	ft_udelay(1000);
   }
   return(1);
}
/*********************************************************************
*  Functions:
*	int CheckChannelBus(unchar ch)   	   
* 
*  Description: (Fujihsu request)
*    	Checking Data Bit(0-7) of channel bus by base+2 if they are OK
*********************************************************************/
int CheckChannelBus(unchar ch) {

   ft_channel_t	*p_ch = &ft_channel[ch];
   ushort	base = p_ch->base ;
   int		value, bit;
   
   for( bit=0 ; bit<8 ; bit++ ) {
	value = 1 << bit ;
   	ft_outb( value , base+2 );
	if( ft_inb(base+2) != value ) {
	   DEB(printk("CheckChannelBus: Channel%d/Bit%d\n",ch+1,bit););
	   p_ch->abnormality = 1 ;
	   return(-1);
	}
   }
	   
   p_ch->abnormality = 0 ;   
   return(0);
}
/*********************************************************************
 * Function:
 *   int InitForSpecificOEM(void)
 *
 * Description:	
 *   Init Engine global variables for specific OEM
*********************************************************************/		
int InitForSpecificOEM(void) {
   #ifdef Linux_Fujitsu_0060
	DisableHotSpare = 1 ;		// Disable Auto Rebuild
//	FirstCheckRebuild = 0 ;		// Disable Auto Rebuild
   #endif
	
   MaintenanceLength = 4 ;
   return(0);
}
/*********************************************************************
 * Function:
 *   int Ioctl3804(void)
 *
 * Description:	
 *   Enable(1)/Disable(0) Rebuild Option with args[0]
*********************************************************************/		
int Ioctl3804(unchar *args) {

   switch( args[0] ) {
	case 0 : // Disable Auto Rebuild
	   DisableHotSpare = 1 ;
//	   FirstCheckRebuild = 0 ;
	   break;
	case 1 : // Enable Auto Rebuild
	   DisableHotSpare = 0 ;
//	   FirstCheckRebuild = 1 ;
	   break;
	default :
	   return(1) ;
   }
   return(0);
}
/*********************************************************************
 * Function:
 *   int Ioctl3807(unchar *)
 *
 * Description:	
 *   RAID1 Synchronization
 *   argument - submit one array that you want to synchronize (1-4)
 *              default is auto select
 *   return - represent request is submitted successfully or not
 *        0 - request is submitted successfully
 *       -1 - submit request failed
 *       -2 - can't rebuild two array simultaneously
*********************************************************************/		
int Ioctl3807(unchar *args) {
   unchar arr, SyncArr;
   unchar rebFlag=0, synFlag=0;
   unchar uc;
   unchar blFlag, bIndex;
   unchar bCH, bID;
   ft_drive_t *dp;

   switch( args[0] ) {
      case 1:
	 SyncArr = 1; break;
      case 2:
	 SyncArr = 2; break;
      case 3:
	 SyncArr = 3; break;
      case 4:
	 SyncArr = 4; break;
      default:
	 SyncArr = 0; break;
   }

   //engine can rebuild only 1 array at the same time
   for (arr=0; arr < FT_Max_Arr; arr++) {
      blFlag = aryDef[arr].blFlag;

      /* there is another raid1  rebuilding */
      if (blFlag & blMARKED_LUN) {
	 return(-2);
      }
   }

   // scan array
   for (arr=0; arr < FT_Max_Arr; arr++) {
      blFlag = aryDef[arr].blFlag;

      if ( (SyncArr > 0 ) && (arr != SyncArr-1) ) {
	 continue;
      }

      /* not RAID1 */
      if (!(aryDef[arr].blRAID & blRAID1)) {
	 continue;
      }

      //array is rebuilding
      if (blFlag & blMARKED_LUN) {
	 return(-2);
      }
      //array is down or critical
      else if (blFlag & blRECOVERING_LUN) {
	 //scan every strm in array
	 for ( uc=0; uc < FT_Max_Dev; uc++) {
	    bIndex = aryDef[arr].strm[uc].bIndex;
	    bCH = blkDisk[bIndex].bCh;
	    bID = blkDisk[bIndex].bID;
	    // this disk is present and strm is normal 
	    if ( (blDiskFlag[bCH][bID] == 1) && (aryDef[arr].strm[uc].blFlag & 0x07) ) {
	       continue;
	    }
	    // this disk is present but strm is abnormal (blREDIRECT_BLK)
	    else if ( (blDiskFlag[bCH][bID] == 1) && (aryDef[arr].strm[uc].blFlag & blREDIRECT_BLK) ) { 

	       if (SyncArr == 0 && (rebFlag != 0 || synFlag != 0) )
		  return(-2);

	       rebFlag = 1;
	       if (!StartRebuild(arr,bCH,bID))
		  return(-1);
	       else return (0);
	       
	    }
	    // this disk is offline, skip this array
	    else if ( blDiskFlag[bCH][bID] == 0 ) {
	       continue;
	    }
	 }
      }
      //array is normal
      //default sync from disk0 to disk1
      else if (blFlag & 0x0f) {
#if 0
	 unchar todo=1;
	 for ( uc=0; uc < FT_Max_Dev; uc++) {
	    bIndex = aryDef[arr].strm[uc].bIndex;
	    if (blkDisk[bIndex].blFlag & blASSIGNED_BLK)
	       todo = uc;
	 }
#endif
	 bIndex = aryDef[arr].strm[1].bIndex;
	 bCH = blkDisk[bIndex].bCh;
	 bID = blkDisk[bIndex].bID;
	 dp = &ft_drive[bCH*2+bID];

	 dp->rebuilding=2;

	 //check if there is another rebuilding array
	 if (SyncArr == 0 && (rebFlag != 0 || synFlag != 0) )
	    return(-2);

	 synFlag = 1;
	 if (!StartRebuild(arr,bCH,bID))
	 //if (!StartCheckIntegrity(arr))
	    return(-1);
	 else return(0);
      }
   }
   return(-1);
}
/*********************************************************************
 * Function:
 *   int Iocal3808(void)	By Hank Yang 2002.01.25
 *   
 * Description:	
 *   Control disk write cache for Fujitsu only
 ********************************************************************/
int Ioctl3808(unchar *args) {
	ft_drive_t *drive;
	struct ft_driveid *id;
	ulong	flags;
	ushort 	base;
	int	DiskWriteCache;
	int	dev, i;

   switch( args[0] ) {
	case 0 :
	   DiskWriteCache = 0x82;	//Disable
	   break;
	case 1 :
	   DiskWriteCache = 0x02;	//Enable
	   break;
	default :
	   DiskWriteCache = 0x82;
   }
   
	for (dev=0 ; dev < FT_Max_Dev ; dev++)
	{
		drive = &ft_drive[dev];
		id = drive->identify;
		if (!drive->present)
			continue;
		for (i=0 ; i<10; i++)	//wait 10 ms
		{
		if (drive->lbusy)
			ft_udelay(1000);
		else break;
		}
		if (drive->lbusy)	//busy timeout
			continue;
		if (!(id->cmd_set[0] & (1<<5)))	//write cache not support
		{
			printk("FastTrak: Drive %d not support Write Cache Command.\n", drive->pchannel->id+1);
			continue;
		}
		save_flags(flags);
		cli();
		base = drive->pchannel->base;
		ft_outb( DiskWriteCache, base+1 );
		ft_outb( drive->select, base+6 );
		ft_outb( drive->select, base+6 );		// for 277
		ft_outb( 0xEF, base+7 );
		if( Check1F7(dev,3*1000,DRDY,BSY|DF|DRQ|ERR) )
			return(-1);
		identify_device(dev, id);	//restore identify info
		restore_flags(flags);
	}
	return(0);
}
/*********************************************************************
 * Function:
 *   int Iocal3809(void)	By Hank Yang 2002.01.25
 *   
 * Description:	
 *   Flush disk cache for Fujitsu only
 ********************************************************************/
int Ioctl3809(unchar *args) {
	ft_drive_t *drive;
	struct ft_driveid *id;
	ushort 	base;
	int	FlushCache;
	int	dev, i;

   switch( args[0] ) {
	case 0 :
	   FlushCache = 0;	//Disable
	   break;
	case 1 :
	   FlushCache = 1;	//Enable
	   break;
	default :
	   FlushCache = 0;
   }
   if (FlushCache)
   {
	for (dev=0 ; dev < FT_Max_Dev ; dev++)
	{
		drive = &ft_drive[dev];
		id = drive->identify;
		if (!drive->present || !(id->cmd_set[0] & (1<<5)) ) //write chche not support
			continue;
		for (i=0 ; i<10; i++)	//wait 10 ms
		{
		if (drive->lbusy)
			ft_udelay(1000);
		else break;
		}
		if (drive->lbusy)	//busy timeout
			continue;
		/*if (!(id->cmd_set[1] & (1<<12)) )	//flush cache not support
		{
			printk("FastTrak: Drive %d not support Flush Cache Command.\n", drive->pchannel->id+1);
			continue;
		}*/
		base = drive->pchannel->base;
		// Flush Cache
		ft_outb( drive->select, base+6 );
		ft_outb( drive->select, base+6 );		// for 277
		ft_outb( 0xE7, base+7 );
		if( Check1F7(dev,3*1000,DRDY,BSY|DF|DRQ|ERR) )
			return(-1) ;
	}
   }
   return(0);
}
