/*+M*************************************************************************
 * Promise FastTrak device driver for Linux.
 *
 * Copyright (c) 2002  Promise Technology, Inc.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * --------------------------------------------------------------------------  
 * Copyright (c) 1999-2002 Promise Technology, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions, and the following disclaimer,
 *    without modification, immediately at the beginning of the file.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * Where this Software is combined with software released under the terms of 
 * the GNU Public License ("GPL") and the terms of the GPL would require the 
 * combined work to also be released under the terms of the GPL, the terms
 * and conditions of this License will apply in addition to those of the
 * GPL with the exception of any terms or conditions of this License that
 * conflict with, or are expressly prohibited by, the GPL.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *   		
 *-M*************************************************************************//*********************************************************************
* channel_wrapper.c
* 08/23/00 Frank Tiernan
* 	This file wraps the linux kernel calls used by the lower 
* channel driver of the RAID engine. This allows the pre-compiled 
* channel driver to be compiled under various flavors of the linux kernel 
* 	The channel driver also uses the following structures:
* timer_list in /usr/src/linux/include/linux/timer.h
* Functions inside the channel driver that use these structures have
* been given wrappers so that in the worst case scenario that these
* structures change and no new engine binaries are released it
* may be possible to convert the new structures into the old structures. 
* 10/30/00 Linus Chen
*	Adding some wrapper functions for os driver.
*********************************************************************/
 
#include <asm-i386/io.h>
#include <asm-i386/ptrace.h>
#include <asm-i386/delay.h>
#include <linux/sched.h>
#include <linux/types.h>
//#include <linux/malloc.h>
#include <linux/blk.h>
#include <linux/pci.h>

/* used same as list_head in linux kernel */
struct 	ft_list_head {
	struct ft_list_head *next, *prev;
};

/* used same as timer_list in linux kernel */
struct 	ft_timer_list {
	struct ft_list_head list;
	unsigned long expires;
	unsigned long data;
	void (*function)(unsigned long);
};

extern void ft_handler(int irq, void *dev_id);

unchar ft_inb (ushort port) {
  	return inb(port);
}

unsigned ft_inw (ushort port) {
	return inw(port);
}

void ft_outb (unchar value, ushort port) {
  	outb(value, port);
}

void ft_outw (ushort value, ushort port) {
	outw(value, port);
}

void ft_outl (ulong value, ushort port) {
  	outl(value, port);
}

//int ft_printk(const char *fmt, ...) {
int ft_printk(const char *fmt, va_list *argp) {
 	return printk(fmt, argp);
}

//extern int sprintf(char * buf, const char * fmt, ...);
int ft_sprintf(char * buf, const char * fmt, va_list *argp) {
	return sprintf(buf, fmt, argp);
}

void ft_init_timer(struct ft_timer_list * timer) {
/* If necessary convert ft_timer_list into timer_list */
  	init_timer((struct timer_list *)timer);
}

void ft_add_timer(struct ft_timer_list * timer) {
/* If necessary convert ft_timer_list into timer_list */
  	add_timer((struct timer_list *)timer);
}

void ft_del_timer(struct ft_timer_list * timer) {
/* If necessary convert ft_timer_list into timer_list */
  	del_timer((struct timer_list *)timer);
}
  
void ft_udelay(unsigned long usecs) {
  	udelay(usecs);
}

unsigned long volatile ft_get_jiffies(void) {
/* jiffies is the number of clock ticks since the OS was booted */
/* it is exported by the linux kernel */
  	return jiffies;
}

void ft_handler_wrapper(int irq, void *dev_id, struct pt_regs *regs) {
   	/* channel driver is system independent so it does not
      	   need pt_regs structure */
	ulong	smpflag;
   	spin_lock_irqsave(&io_request_lock,smpflag);
   	ft_handler(irq, dev_id);
   	spin_unlock_irqrestore(&io_request_lock,smpflag);
}

void ft_do_gettimeofday(struct timeval *tv) {
	do_gettimeofday(tv);
}

//void *ft_kmalloc(unsigned int size,int priority) {
//	return(kmalloc(size,priority));
//}

//void ft_kfree(void *obj) {
//	kfree(obj);
//}

