/*
 * linux/kernel/lids_logs.c
 *
 * Author : Philippe Biondi, (pbi@cartel-info.fr)
 *
 * Description : This file contains the LIDS logging functions
 *
 * Changes :
 *
 *  - Nov 06, 2001. Philippe Biondi : initial creation
 *
 */

#include <linux/sched.h>
#include <linux/mm.h>
#include <linux/slab.h>
#include <linux/lids.h>

#define STR2(x) #x
#define STR(x) STR2(x)




#ifdef CONFIG_LIDS_HANGUP
void lids_hangup_console() 
{
	if(current->tty != NULL && !lids_first_time ) {
		if (!test_bit(TTY_LIDS_NO_HANGUP,&current->tty->flags) )
                	tty_vhangup(current->tty);
        }
}	
#endif


#ifdef CONFIG_LIDS_SA_THROUGH_NET
#define lids_print(message, args...) sprintf(buf, message , ## args)
#else
#define lids_print(message, args...) printk(KERN_ALERT message , ## args)
#endif

void lids_log(int flood, char *message, ...)
{
	va_list args;
	struct vm_area_struct * vma;
	char ttyname[64];
	char progname[64];
	char proginfo[64+10+10+20+24]; /* %s+%d+%d+%ld+le texte avec un peu de marge = 128 */
	char msgstr[256];
	       
	struct dentry *f_dentry;



#ifdef CONFIG_LIDS_SA_THROUGH_NET
	char *buf;
	buf=(char *)kmalloc(2048,GFP_KERNEL);
#endif

	/* Get args on the stack */
	va_start(args, message);

	/* Get dentry of current process, if any */
	f_dentry = NULL;
	if (current->mm) {
		vma = current->mm->mmap;
		while (vma) {
			if ((vma->vm_flags & VM_EXECUTABLE) &&
			    vma->vm_file) {
				f_dentry=vma->vm_file->f_dentry;
				break;
			}
			vma = vma->vm_next;
		}
	}

	/* Get the tty name, if any */
	if (current->tty)
		tty_name(current->tty,ttyname);		
	else
		strncpy(ttyname,"null tty",10);

	/* Make the proginfo string */       
	if (f_dentry && f_dentry->d_inode) {
		strncpy(progname,f_dentry->d_iname,63);
		snprintf(proginfo,127,"%s (dev %d:%d inode %ld)",
			 progname,
			 MAJOR(f_dentry->d_inode->i_dev),
			 MINOR(f_dentry->d_inode->i_dev),
			 f_dentry->d_inode->i_ino);
	}
	else {
		strncpy(proginfo,"(undetermined program)",63);
	}

	/* Make the message string */
	vsnprintf(msgstr,255,message,args);


	/* Make the log string */
	lids_print("LIDS: %s pid %d ppid %d uid/gid (%d/%d) on (%s) : %s %s\n", 
		   proginfo,
		   current->pid,
		   current->p_pptr->pid,			
		   current->uid,
		   current->gid,
		   ttyname,
		   msgstr,
		   flood ? " - logging disabled for " STR(CONFIG_LIDS_TIMEOUT_AFTER_FLOOD) "s" : "");

		   

#ifdef CONFIG_LIDS_SA_THROUGH_NET
	printk(KERN_ALERT "%s",buf);
	lids_send_message(buf,strlen(buf));
#endif

	/* deal with args on the stack */
	va_end(args);

}

